/*  Screem:  screem-application.c
 *
 *  The main screem application code, manages startup and generally looks
 *  after things
 *
 *  Copyright (C) 2001  David A Knight
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  For contact information with the author of this source code please see
 *  the AUTHORS file.  If there is no AUTHORS file present then check the
 *  about box under the help menu for a contact address
 */
#include <config.h>

#include <errno.h>

#include <gmodule.h>

#include <gtk/gtkcombo.h>
#include <gtk/gtkimage.h>
#include <gtk/gtkvbox.h>
#include <gtk/gtkmessagedialog.h>
#include <gtk/gtkiconfactory.h>
#include <gtk/gtkstock.h>
#include <gdk-pixbuf/gdk-pixbuf.h>

#include <libgnome/gnome-i18n.h>

#include <libgnomevfs/gnome-vfs-mime-handlers.h>
#include <libgnomevfs/gnome-vfs-mime-utils.h>
#include <libgnomevfs/gnome-vfs-file-info.h>
#include <libgnomevfs/gnome-vfs-ops.h>
#include <libgnomevfs/gnome-vfs-utils.h>

#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <dirent.h>
#include <sys/types.h>
#include <unistd.h>

#include <pcreposix.h>

#include "pageUI.h"

#include "screem-application.h"
#include "screem-hint.h"
#include "screem-window.h"

#include "screem-editor.h"

#include "screem-window-private.h" /* YUK! */

#include "screem-plugin.h"  /* YUK */


#include "screem-site-ui.h" /* YUK again */

#include "screem-helper.h"

#include "screem-tagtree.h"

#include "support.h"
#include "fileops.h"
#include "php3.h"
#include "xml.h"


#include "rb-sidebar.h"
#include "rb-sidebar-button.h"

#include "libegg/menu/egg-menu.h"

static void screem_application_init( ScreemApplication *application );
static void screem_application_class_init( ScreemApplicationClass *application );
static void screem_application_finalize( GObject *object );

static void screem_application_destroyed_window( GObject *object,
						 ScreemApplication *application );
static gboolean screem_window_delete_event_callback( GtkWidget *widget,
						     GdkEvent *event,
						     gpointer user_data );

static void dtds_insert( gpointer key, gpointer value, GtkWidget *widget );

static void screem_application_add_icons( ScreemApplication *application );

static gboolean screem_application_load( ScreemApplication *application );

static void screem_application_load_helpers( ScreemApplication *application );
static void screem_application_helper_failed( ScreemHelper *helper,
					      ScreemApplication *application );
static void screem_application_helper_error( ScreemHelper *helper,
					     gint status,
					     ScreemApplication *application );

static void screem_application_load_macros( ScreemApplication *application );

#define SPLASH_TEXT( w, t ) gtk_label_set_text( GTK_LABEL( w ),t ); \
while( g_main_pending() ) g_main_iteration(TRUE);

static void screem_application_add_sub_dtd( ScreemDTD *dtd, 
					    const gchar *element,
					    gpointer *module,
					    ScreemApplication *application );


static void screem_application_offline_notify( GConfClient *client,
					       guint cnxn_id,
					       GConfEntry *entry,
					       gpointer data );

static void screem_application_copy_tag_table( GtkTextTag *tag,
					       gpointer data );
static void screem_application_copy_sub_tags( GtkTextTag *tag,
					      gpointer data );


GList *screem_application_get_window_list( ScreemApplication *app )
{
	GList *list;

	list = app->window_list;

	return list;
}

void screem_application_close_all_windows( ScreemApplication *app )
{
	while( app->window_list ) {
		screem_window_close( SCREEM_WINDOW( app->window_list->data ) );
	}
}

static gboolean screem_window_delete_event_callback( GtkWidget *widget,
						     GdkEvent *event,
						     gpointer user_data )
{
	ScreemWindow *window;
	GList *list;
	gboolean close;

	window = SCREEM_WINDOW( widget );

	close = TRUE;
	
	/* if this is the last window we should confirm saving */
	list = screem_application_get_window_list( window->application );
	if( g_list_length( list ) == 1 ) {
		close = screem_application_save_sites( window->application );
	} 
	
	if( close ) {
		screem_window_close( window );
	}
		
	return TRUE;
}

static void
screem_application_destroyed_window( GObject *object, ScreemApplication *app )
{
	GList *list;
	ScreemPlugin *plugin;

	app->window_list = g_list_remove( app->window_list, object );

	for( list = app->plugins; list; list = list->next ) {
		plugin = (ScreemPlugin*)list->data;
		plugin->remove_ui( GTK_WIDGET( object ), NULL, NULL, NULL );
	}

	/* if all windows are closed then exit */
	if( ! app->window_list && ! app->finalizing ) {
		gtk_main_quit();
	}
}

ScreemWindow *screem_application_create_window( ScreemApplication *app )
{
	ScreemWindow *window;
	GList *list;

	window = SCREEM_WINDOW( g_object_new( screem_window_get_type(),
					      "app",
					      G_OBJECT( app ),
					      "app_id", "screem", NULL ) );
	
	g_signal_connect( G_OBJECT( window ), "delete_event",
			  G_CALLBACK( screem_window_delete_event_callback ),
			  NULL );
	g_signal_connect( G_OBJECT( window ), "destroy",
			  G_CALLBACK( screem_application_destroyed_window ),
			  app );

	app->window_list = g_list_prepend( app->window_list, window );

	for( list = app->loaded_sites; list; list = list->next ) {
		screem_window_add_site( window, SCREEM_SITE( list->data ) );
	}

	for( list = app->plugins; list; list = list->next ) {
		ScreemPlugin *plugin = (ScreemPlugin*)list->data;

		plugin->add_ui( GTK_WIDGET( window ), 
				window->details->editor,
				window->details->preview, 
				window->details->link_view );
	}
	for( list = app->helpers; list; list = list->next ) {
		ScreemHelper *helper = (ScreemHelper*)list->data;
		screem_helper_add_to_window( helper, window );
	}

	return window;
}

ScreemApplication *screem_application_new()
{
	ScreemApplication *application;
	GType type;

	type = screem_application_get_type();

	application = SCREEM_APPLICATION( g_object_new( type, NULL ) );

	return application;
}

static GtkWidget *screem_application_splash( ScreemApplication *application,
					     gboolean close)
{
       	GtkWidget *label;
	static GtkWidget *window;
	GtkWidget *image;
	GtkWidget *box;

	if( ! close ) {
		window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
		gtk_window_set_title( GTK_WINDOW( window ),
				      _( "Loading - Screem" ) );
		gtk_window_set_position( GTK_WINDOW( window ), 
					 GTK_WIN_POS_CENTER_ALWAYS );
       
		image = gtk_image_new_from_file( SPLASH_DIR"/splash.png" );
		gtk_widget_show( image );		

		label = gtk_label_new( _( "Starting..." ) );
		gtk_widget_show( label );
		
		box = gtk_vbox_new( FALSE, 0 );
		gtk_box_pack_start( GTK_BOX( box ), image, TRUE, TRUE, 0 );
		gtk_box_pack_start( GTK_BOX( box ), label, TRUE, TRUE, 0 );

		gtk_widget_show( box );
		
		gtk_container_add( GTK_CONTAINER( window ), box );
		
		gtk_widget_show_all( window );
		
		return label;
	} else {
		gtk_widget_destroy( window );
		return NULL;
	}
}

void screem_application_startup( ScreemApplication *application,
				 const gchar **start_files )
{
	ScreemWindow *window;
	ScreemSite *site;
	ScreemPage *page;
	
	GtkWidget *splash_text;
	gchar *glade_path;

	application->start_files = start_files;

	application->client = gconf_client_get_default();
	
	gconf_client_add_dir( application->client, "/apps/screem",
			      GCONF_CLIENT_PRELOAD_ONELEVEL, NULL );
	gconf_client_add_dir( application->client, "/apps/screem/editor",
			      GCONF_CLIENT_PRELOAD_ONELEVEL, NULL );
	gconf_client_add_dir( application->client, "/apps/screem/editor/print",
			      GCONF_CLIENT_PRELOAD_ONELEVEL, NULL );
	
	glade_path = gconf_client_get_string(application->client,
					     "/apps/screem/general/glade_path",
					     NULL );
	if( ! glade_path ) {
		glade_path = g_strconcat( GLADE_PATH, "/screem.glade",
					  NULL );
		gconf_client_set_string( application->client,
					 "/apps/screem/general/glade_path",
					 glade_path, NULL );
	}
	g_free( glade_path );

	/* create splash screen */
	splash_text = screem_application_splash( application, FALSE );
	SPLASH_TEXT( splash_text, _( "Initialising..." ) );

	/* load syntax files */
	SPLASH_TEXT( splash_text, _( "Loading Syntax files..." ) );
	screem_application_load_syntax_tables();

	/* load php files */
	SPLASH_TEXT( splash_text, _( "Loading PHP files..." ) );
	screem_php_build_function_hash_table();

	/* setup offline notification / status */
	application->offline = gconf_client_get_bool( application->client,
						      "/apps/screem/general/work_offline",
						      NULL );
	application->offline_notify = 
		gconf_client_notify_add( application->client,
					 "/apps/screem/general/work_offline", 
					 screem_application_offline_notify,
					 application, NULL, NULL );
	
	application->hint = GTK_WIDGET( screem_hint_new() );

	/* add icons */
	screem_application_add_icons( application );

	/* load plugins */
	if( g_module_supported() ) {
		SPLASH_TEXT( splash_text, _( "Loading Plugins..." ) );
		scan_plugins( application, splash_text );
	}

	/* load macros */
	SPLASH_TEXT( splash_text, _( "Loading Macros..." ) );
	screem_application_load_macros( application );

	/* load helpers */
	SPLASH_TEXT( splash_text, _( "Loading Helpers..." ) );
	screem_application_load_helpers( application );

	/* load DTDs */
	SPLASH_TEXT( splash_text, _( "Loading DTDs..." ) );
	screem_dtd_load_catalogs( application );

	/* create the fake site for editing individual files */
	site = screem_site_new();
	screem_application_add_site( application, site );

	/* create a dummy page for the fake site */
	page = screem_page_new();
	screem_page_set_data( page, NULL );

	screem_site_add_page( site, page );

	/* create initial window */
	SPLASH_TEXT( splash_text, _( "Creating Interface..." ) );

	window = screem_application_create_window( application );

	screem_window_set_current( window, site );

	screem_window_set_document( window, page );

	screem_application_splash( application, TRUE );

	gtk_widget_show( GTK_WIDGET( window ) );

	/* display hints if necessary */
	{
		gboolean show_hints;

		show_hints = gconf_client_get_bool( application->client,
						    "/apps/screem/general/show_tips",
						    NULL );
		if( show_hints ) {
			gtk_window_set_transient_for(GTK_WINDOW(application->hint),
						     GTK_WINDOW(window));
								  
			gtk_widget_show( application->hint );
		}
	}

	/* delay loading of tag trees so we can get the main window up
	   quicker */
	g_idle_add( (GSourceFunc)screem_tag_tree_load_trees, NULL );
	
	/* delay loading sites / files until the main loop has started up */
	g_idle_add( (GSourceFunc)screem_application_load, application );
}

static gboolean screem_application_load( ScreemApplication *application )
{
	const gchar **start_files = application->start_files;
	const gchar *mime_type;
	gchar *file;
	ScreemWindow *window;
	gint i;

       	window = application->window_list->data;

	if( gconf_client_get_bool( application->client,
				   "/apps/screem/autoload",
				   NULL ) ) {
		GSList *sites;
		GSList *tmp;
		sites = gconf_client_get_list( application->client,
					       "/apps/screem/sites",
					       GCONF_VALUE_STRING, NULL );
		for( tmp = sites; tmp; tmp = tmp->next ) {
			screem_site_open_with_filename(window,
						       (const gchar*)tmp->data);
			g_free( (gchar*)tmp->data );
		}
		if( sites ) {
			g_slist_free( sites );
		}
	}

	/* process command line files / sites */
	for( i = 0; start_files && start_files[ i ]; i ++ ) {
		GnomeVFSURI *uri;
		gboolean isdir;

		uri = gnome_vfs_uri_new( start_files[ i ] );
		
		/* is the filename local? */
		if( gnome_vfs_uri_is_local( uri ) ) {
			/* yes */
			/* HACK */
			if( *start_files[ i ] == G_DIR_SEPARATOR ) {
				file = g_strconcat( "file://", 
						    start_files[ i ], NULL );
			} else {
				file = relative_to_full( start_files[ i ],
							 NULL );
			}
		} else {
			file = g_strdup( start_files[ i ] );
		}
		isdir = screem_uri_is_dir( file );
		gnome_vfs_uri_unref( uri );

		/* is it a project file or an html file? */
		mime_type = gnome_vfs_get_mime_type( file );
		gdk_threads_enter();
		if( isdir || 
		    ( mime_type &&
		      ! strcmp( mime_type, "application/x-screem" ) ) ) {
			screem_site_open_with_filename( window, file );
		} else {
			ScreemSite *site;
			site =screem_application_get_default_site(application);
			screem_page_open_with_filename( site, window, file );
		}
	      	gdk_threads_leave();
		g_free( file );
	}

	return FALSE;
}

void screem_application_add_site( ScreemApplication *app, ScreemSite *site )
{
	GList *list;
	ScreemWindow *window;

	/* FIXME: should be a property of ScreemSite */
	g_object_set_data( G_OBJECT( site ), "application", app );

	/* add to list of loaded sites */
	app->loaded_sites = g_list_append( app->loaded_sites, site );

	/* add to recent sites */
	if( ! screem_site_get_fake_flag( site ) ) {
		/* work on the first window, others will pick up
		   the changes via gconf */
		window = app->window_list->data;
		gedit_recent_add( window->details->recent_sites,
				  screem_site_get_pathname( site ) );
	}

	for( list = screem_application_get_window_list( app ); list;
	     list = list->next ) {
		screem_window_add_site( SCREEM_WINDOW( list->data ), site );
	}
}

ScreemSite *screem_application_get_site( ScreemApplication *app,
					 const gchar *uri )
{
	ScreemSite *site;
	GList *list;

	/* loaded_sites->next as the first one is guarenteed
	   to be the fake site */
	site = NULL;
	for( list = app->loaded_sites->next; list; list = list->next ) {
		const gchar *pathname;

		site = SCREEM_SITE( list->data );
		pathname = screem_site_get_pathname( site );
		if( ! strcmp( uri, pathname ) ) {
			break;
		}
		site = NULL;
	}

	return site;
}

ScreemSite *screem_application_get_default_site( ScreemApplication *app )
{
	/* default site is always first in the list */
	return SCREEM_SITE( app->loaded_sites->data );
}

void screem_application_close_site( ScreemApplication *app, ScreemSite *site )
{
	gint item;
	const gchar *name;
	gchar *verb;
	GList *list;

	item = g_list_index( app->loaded_sites, site );

	app->loaded_sites = g_list_remove( app->loaded_sites, site );

	/* remove verb for switch site menus and remove button from site
	   sidebars */
	name = screem_site_get_name( site );

	verb = g_strconcat( "SwitchSite_", name, NULL );
	
	for( list = screem_application_get_window_list( app ); list;
	     list = list->next ) {
		ScreemWindow *window;
		GSList *sites;
		RBSidebarButton *button;
		EggAction *action;
		guint mergeid;
		gpointer sd;
		     
		window = SCREEM_WINDOW( list->data );

		button = NULL;
		for( sites = window->details->sitebargroup; sites;
		     sites = sites->next ) {
			button = RB_SIDEBAR_BUTTON( sites->data );
			if( ! strcmp( name, button->unique_id ) ) {
				break;
			}
			button = NULL;
		}

		if( button ) {
			g_assert( site ==
				  g_object_get_data( G_OBJECT( button ),
						     "site" ) );
			window->details->sitebargroup = 
				g_slist_remove( window->details->sitebargroup,
						button );
			rb_sidebar_remove(RB_SIDEBAR(window->details->sitebar),
					  button);
		}

		action = egg_action_group_get_action( EGG_ACTION_GROUP( window->action_group ),
						      verb );
		mergeid = GPOINTER_TO_UINT( g_object_get_data( G_OBJECT( action ),
								"mergeid" ) );
		sd = g_object_get_data( G_OBJECT( action ), "sitedata" );
		egg_action_group_remove_action( EGG_ACTION_GROUP( window->action_group ),
						action );
		egg_menu_merge_remove_ui( EGG_MENU_MERGE( window->merge ),
					  mergeid );
		g_free( sd );
	}
	g_free( verb );
}

void screem_application_close_all_sites( ScreemApplication *app )
{
	ScreemSite *site;

	GSList *sites;

	sites = NULL;

	while( app->loaded_sites ) {
		const gchar *pathname;

		site = SCREEM_SITE( app->loaded_sites->data );
		if( ! screem_site_get_fake_flag( site ) ) {
			pathname = screem_site_get_pathname( site );

			sites = g_slist_prepend( sites,
						 g_strdup( pathname ) );
		}
		screem_application_close_site( app, site );
		g_object_unref( site );
	}

	gconf_client_set_list( app->client, "/apps/screem/sites",
			       GCONF_VALUE_STRING, sites, NULL );
	g_slist_foreach( sites, (GFunc)g_free, NULL );
	g_slist_free( sites );
}

gboolean screem_application_save_sites( ScreemApplication *app )
{
	GList *list;
	ScreemSite *site;

        gint button;
	gboolean ret;

	ret = TRUE;

	/* request confirmation to save any opened sites / pages */
	for( list = app->loaded_sites; ret && list; list = list->next ) {
		site = SCREEM_SITE( list->data );
		if( ! screem_site_get_fake_flag( site ) ) {
			screem_site_save( site );
			button = GTK_RESPONSE_OK;
		} else {
			/* this is the fake site, so ask about saving the
			   individual pages in it */
			GList *list2;

			button = GTK_RESPONSE_YES;
			for( list2 = screem_site_get_pages( site ); 
			     list2 && button != GTK_RESPONSE_CANCEL ;
			     list2 = list2->next ) {
				button = screem_page_save_confirmation( SCREEM_PAGE(list2->data) );
			}
		}
		
		if( button == GTK_RESPONSE_CANCEL ||
		    button == GTK_RESPONSE_NONE ||
		    button == GTK_RESPONSE_DELETE_EVENT ){
			ret = FALSE;
		}
	}

	return ret;
}

ScreemDTD *screem_application_add_dtd( ScreemApplication *application,
				       const gchar *element,
				       ScreemDTD *parent )
{
	ScreemDTD *dtd;
	gchar *id;
	gchar *path;
	gchar *dtdpath;

	g_return_val_if_fail( element != NULL, NULL );

	dtd = NULL;

     	if( screem_dtd_parse_doctype( element, &id, &path ) ) {
		dtdpath = g_hash_table_lookup( application->dtd_types, id );

		if( dtdpath ) {
			dtd = g_hash_table_lookup( application->loaded_dtds,
						   dtdpath );
		} else {
			gchar *temp;

			path = g_strstrip( path );
			temp = path;
			if( *temp == '"' ) {
				gchar *end;

				temp ++;
				end = strchr( temp, '"' );
				if( end ) {
					*end = '\0';
				}
			}
			if( temp[ strlen( temp ) - 1 ] == '>' ) {
				temp[ strlen( temp ) - 1 ] = '\0';
			}

			dtdpath = temp;
		}
		
		if( ! dtd ) {
			dtd = screem_dtd_new_with_id( id );
			g_signal_connect( G_OBJECT( dtd ), "load_module",
					  G_CALLBACK( screem_application_add_sub_dtd ), 
					  application );
			g_object_set_data( G_OBJECT( dtd ), "type",
					   g_strdup( id ) );
			g_hash_table_insert( application->dtd_types,
					     g_strdup( id ), 
					     g_strdup( dtdpath ) );
			g_hash_table_insert( application->loaded_dtds,
					     g_strdup( dtdpath ),
					     dtd );
		}
		if( ! screem_dtd_load( dtd, dtdpath, parent,
				       ! application->offline ) ) {
			/* load failed, remove from hash table */
			gchar *id;
			
			id = g_object_get_data( G_OBJECT( dtd ),
						"type" );
			g_hash_table_remove( application->dtd_types,
					     id );
			g_hash_table_remove( application->loaded_dtds,
					     dtdpath );
			
			g_print( "FAILED TO LOAD DTD: %s\nID: %s\n",
				 dtdpath, id );
			g_free( id );
			g_object_unref( dtd );
			dtd = NULL;
		}
	}

	g_free( id );
	g_free( path );

	return dtd;
}

static void screem_application_add_sub_dtd( ScreemDTD *dtd, 
					    const gchar *element,
					    gpointer *module,
					    ScreemApplication *application )
{
	*module = screem_application_add_dtd( application, element, dtd );
}

/* DTD loading */
void screem_dtd_fill_combo( ScreemApplication *application, GtkWidget *combo )
{
	GList *list;

	g_object_set_data( G_OBJECT( combo ), "list", NULL );
	g_hash_table_foreach( application->dtd_types, 
			      (GHFunc)dtds_insert, combo );
	list = g_object_get_data( G_OBJECT( combo ), "list" );
	list = g_list_sort( list, (GCompareFunc)strcmp );
	gtk_combo_set_popdown_strings( GTK_COMBO( combo ), list );
	g_list_free( list );
}

static void dtds_insert( gpointer key, gpointer value, GtkWidget *widget )
{
	if( GTK_IS_COMBO( widget ) ) {
		GList *list;

		list = (GList*)g_object_get_data( G_OBJECT( widget ), "list" );

		list = g_list_prepend( list, key );

		g_object_set_data( G_OBJECT( widget ), "list", list );
	}
}

const gchar* screem_dtd_get_system_id( ScreemApplication *application,
				       const gchar *publicid )
{
	const gchar *systemid;

	systemid = g_hash_table_lookup( application->dtd_types, publicid ); 

	return systemid;
}

void screem_dtd_load_catalog( ScreemApplication *application, GString *pathname )
{
	gchar *buffer;
	FILE *file;

	gchar *id;
	gchar *path;

	gchar *tid;
	gchar *tpath;
	gchar *temp;
	gint ret;
	GString *tpathname;

	buffer = g_new0( gchar, BUFSIZ );
	id = g_new0( gchar, BUFSIZ );
	path = g_new0( gchar, BUFSIZ );

	/* we have a catalog file */
	file = fopen( pathname->str, "r" );
	if( ! file )
		return;

	while( fgets( buffer, BUFSIZ, file ) ) {
		buffer = g_strchug( buffer );
		if( ! strncmp( "CATALOG ", buffer, strlen( "CATALOG " ) ) ) {
			/* another catalog file to check */
			buffer[ strlen( buffer) - 1 ] = '\0';
			tpathname = g_string_new( buffer + strlen( "CATALOG " ) );
			screem_dtd_load_catalog( application, tpathname );
			g_string_free( tpathname, TRUE );
			continue;
		}
		if( strncmp( "PUBLIC ", buffer, strlen( "PUBLIC " ) ) )
			continue;
		ret = sscanf( buffer, "PUBLIC \"%[^\"]\" \"%[^\"]\"", 
			      id, path );
		if( ! ret )
			break;
		else if( ret == 1 ) {
			/* ok try a different match */
			if( sscanf( buffer, "PUBLIC \"%[^\"]\" %s", 
				    id, path ) != 2 )
				break;
		}
		
		/* is id already in the hash table? */
		if( ! g_hash_table_lookup( application->dtd_types, id ) ) {
			/* nope */
			tid = g_strdup( id );
			if( ! g_path_is_absolute( path ) ) {
				tpath = g_dirname( pathname->str );
				temp = g_strconcat( "file://",
						    tpath, G_DIR_SEPARATOR_S,
						    path, NULL );
				g_free( tpath );
				tpath = temp;
			}
			else
				tpath = g_strconcat( "file://", path, NULL );
			
			g_hash_table_insert( application->dtd_types, tid, tpath );
		}
	}
	fclose( file );

	g_free( buffer );
	g_free( id );
	g_free( path );
}


void screem_dtd_load_catalogs( ScreemApplication *application )
{
	gchar *paths[] = {
		SCREEM_DTD_PATH,
		"/etc/sgml",
		"/usr/lib/sgml",                      /* Debian, RedHat */
		"/usr/share/sgml",                    /* SuSE */
		"/usr/share/lib/sgml/locale/C/dtds",  /* Solaris, use the 
							 current locale at 
							 runtime? */
		NULL,                                 /* replaced with the
							 users .screem dir at
							 runtime */
		NULL
	};

	gint i;
	GString *file;

	DIR *dir;
	struct dirent *entry;

	paths[ 4 ] = g_strconcat( g_get_home_dir(), G_DIR_SEPARATOR_S,
				  ".screem", G_DIR_SEPARATOR_S,
				  "dtds", NULL );

	/* make dtds dir if needed */
	mkdir_recursive( paths[ 4 ], GNOME_VFS_PERM_USER_ALL );

	application->dtd_types = g_hash_table_new( g_str_hash, g_str_equal );
	application->loaded_dtds = g_hash_table_new( g_str_hash, g_str_equal );

	for( i = 0; paths[ i ]; ++ i ) {
		if( ! ( dir = opendir( paths[ i ] ) ) ) {
			continue;
		}

		/* get next file */
		while( ( entry = readdir( dir ) ) ) {
			if( ( ! g_strcasecmp( "catalog", entry->d_name ) ) ||
			    ( ! g_strncasecmp( "catalog.", 
					       entry->d_name, 
					       strlen( "catalog." ) ) ) ) {
				file = g_string_new( paths[ i ] );
				g_string_append_c( file, G_DIR_SEPARATOR );
				g_string_append( file, entry->d_name );
				
				screem_dtd_load_catalog( application, file );
				
				g_string_free( file, TRUE );
			}
		}
		closedir( dir );
	}
	
	g_free( paths[ 4 ] );
}

void screem_application_register_plugin( ScreemApplication *application, 
					 ScreemPlugin *plugin )
{
	application->plugins = g_list_prepend( application->plugins, plugin );
}

static GObject* screem_application_create_tag_table( GHashTable *tables,
						     const gchar *filename )
{
	GtkTextTagTable *table;
	GtkTextTag *tag;
	gchar *str;
	GList *types;
	GList *subs;
	GList *tags;   /* for easier access than the hash table
			  in the GtkTextTagTable */
	GList *tmp;

	xmlNodePtr root;
	xmlNodePtr node;
	xmlNodePtr subnode;
	gchar *tagname;
	gchar *val;
	gboolean use;

	GString *start_patterns;
	GString *sub_patterns;
	regex_t *sregex;
	xmlDocPtr doc;

	gint i;
	static const gchar *features[] = {
		"markup", "feature_markup",
		"doctype", "feature_doctype",
		"charset", "feature_charset",
		"dynamic", "feature_dynamic",
		NULL
	};

	doc = xmlParseFile( filename );

	if( doc ) {
		root = xmlDocGetRootElement( doc );
		
		g_assert( root );

		val = xmlGetProp( root, "version" );
		if( val && strcmp( "2", val ) ) {
			g_free( val );
			val = NULL;
		}
		if( ! val ) {
			xmlFreeDoc( doc );
			doc = NULL;
		} else {
			g_free( val );
		}
	}
	if( ! doc ) {
		return NULL;
	}
	
	table = gtk_text_tag_table_new();
	tags = NULL;
	
	/* get mime types */
	types = NULL;
	node = xml_search_child( root, "type" );
	while( node ) {
		gchar *type;
		
		if( node->name && ! strcmp( "type", node->name ) ) {
			type = xmlNodeGetContent( node );
			types = g_list_append( types, type );
		}
		node = node->next;
	}
	g_object_set_data( G_OBJECT( table ), "types", types );
	
	start_patterns = g_string_new( "" );

	/* check for includes, e.g. PHP and ASP will want to
	   include an HTML syntax file */
	node = xml_search_child( root, "include" );
	if( node && ( val = xmlGetProp( node, "filename" ) ) ) {
		GtkTextTagTable *inctable;
		gchar *dirname;
		gchar *temp;

		dirname = g_dirname( filename );
		temp = g_strconcat( dirname, G_DIR_SEPARATOR_S, val, NULL );

		inctable = g_hash_table_lookup( tables, temp );
		if( ! inctable ) {
			inctable = GTK_TEXT_TAG_TABLE( screem_application_create_tag_table( tables, temp ) );
			if( inctable ) {
				g_hash_table_insert( tables, temp, inctable );
			} else {
				g_free( temp );
			}
		} else {
			g_free( temp );
		}
		g_free( dirname );
		if( inctable ) {
			/* we have the syntax file loaded for the one we
			   want to include */

			/* copy features */
			for( i = 0; features[ i ]; i += 2 ) {
				if( g_object_get_data( G_OBJECT( inctable ), 
						       features[ i + 1 ] ) ) {
					g_object_set_data( G_OBJECT( table ),
							   features[ i + 1 ],
							   GUINT_TO_POINTER( 1 ) );
				}
			}

			/* copy tags */
			g_object_set_data( G_OBJECT( table ), "start_patterns",
					   start_patterns );
			gtk_text_tag_table_foreach( inctable,
						    screem_application_copy_tag_table, table );
			gtk_text_tag_table_foreach( inctable,
						    screem_application_copy_sub_tags, table );
			tags = g_object_get_data( G_OBJECT( table ), "tags" );
		}
		g_free( val );
	}

	/* get requested features */
	node = xml_search_child( root, "features" );
	if( node ) {
		for( i = 0; features[ i ]; i += 2 ) {
			val = xmlGetProp( node, features[ i ] );
			if( val ) {
				if( ! strcmp( "1", val ) ) {
					g_object_set_data( G_OBJECT( table ),
							   features[ i + 1 ],
							   GUINT_TO_POINTER( 1 ) );
				}
				g_free( val );
			}
		}
	}

	/* get highlighting patterns / colours */
	node = xml_search_child( root, "highlighting" );
	for( node = node->xmlChildrenNode; node; node = node->next ) {
		gboolean isgroup;

		subs = NULL;
		
		if( node->name && 
		    ( ! strcmp( node->name, "group" ) ||
		      ! strcmp( node->name, "subgroup" ) ) ) {
			tagname = xmlGetProp( node, "name" );
		} else {
			tagname = NULL;
		}

		if( ! tagname ) {
			continue;
		}

		isgroup = ! strcmp( node->name, "group" );

		tag = gtk_text_tag_table_lookup( table, tagname );

		if( tag ) {
			g_warning( "Duplicate group %s in syntax highlighting: %s\nOld group will be ignored",
				   tagname, filename );
		}

		tag = gtk_text_tag_new( tagname );
		
		/* set foreground colour */
		subnode = xml_search_child( node, "fore" );
		if( subnode ) {
			/* add a tag with name == tagname_fore,
			   we then put that in the table without any pattern,
			   it is then set as a data item on the tag with
			   name == tagname, so we can support enable/disable
			   of fore/back tags */
			gchar *foretagname = g_strconcat( tagname, "_fore",
							  NULL );
			GtkTextTag *ftag;
			ftag = gtk_text_tag_new(foretagname);
			g_free( foretagname );

			val = xmlGetProp( subnode, "use" );
			use = (atoi( val ) == 1);
			g_free( val );

			str = xmlGetProp( subnode, "color" );
			if( str ) {
				g_object_set( G_OBJECT( ftag ), "foreground",
					      str, NULL );
				g_free( str );
			}

			g_object_set_data( G_OBJECT( ftag ), "usefore", 
					   GINT_TO_POINTER( use ) );
			g_object_set_data( G_OBJECT( tag ), "fore_tag", ftag );
			g_object_set_data( G_OBJECT( ftag ), "sub",
					   GINT_TO_POINTER( ! isgroup ) );
			gtk_text_tag_table_add( table, ftag );
		}
		/* set background colour */
		subnode = xml_search_child( node, "back" );
		if( subnode ) {
			gchar *backtagname = g_strconcat( tagname, "_back",
							  NULL );
			GtkTextTag *btag;
			btag = gtk_text_tag_new(backtagname);
			g_free( backtagname );

			val = xmlGetProp( subnode, "use" );
			use = (atoi( val ) == 1);
			g_free( val );

			str = xmlGetProp( subnode, "color" );
			if( str ) {
				g_object_set( G_OBJECT( btag ),
					      "background", str, NULL );
				g_free( str );
			}
			g_object_set_data( G_OBJECT( btag ), "useback", 
					   GINT_TO_POINTER( use ) );
			g_object_set_data( G_OBJECT( tag ), "back_tag", btag );
			g_object_set_data( G_OBJECT( btag ), "sub",
					   GINT_TO_POINTER( ! isgroup ) );
			gtk_text_tag_table_add( table, btag );
		}
		/* set regexps */
		subnode = xml_search_child( node, "start" );
		if( subnode ) {
			regex_t *regex;

			val = xmlNodeGetContent( subnode );
			val = g_strstrip( val );
			
			regex = g_new0( regex_t, 1 );
			regcomp( regex, val, REG_EXTENDED | REG_NEWLINE );

			g_object_set_data( G_OBJECT( tag ), "start", regex );
			if( isgroup ) {
				if( *start_patterns->str != '\0' ) {
					g_string_append_c( start_patterns, '|' );
				}
				g_string_append( start_patterns, val );
			}
			g_object_set_data( G_OBJECT( tag ),
					   "pattern", val );
		}

		sub_patterns = g_string_new( "" );

		subnode = xml_search_child( node, "end" );
		if( subnode ) {
			regex_t *regex;

			val = xmlNodeGetContent( subnode );
			val = g_strstrip( val );
			
			regex = g_new0( regex_t, 1 );
			regcomp( regex, val, REG_EXTENDED | REG_NEWLINE );

			g_object_set_data( G_OBJECT( tag ), "end", regex );
		
			g_string_append( sub_patterns, val );

			subs = g_list_prepend( subs, tag );

			g_free( val );
		}
		if( isgroup && 
		    ( subnode = xml_search_child( node, "sub" ) ) ) {
			/* group has subgroups */
			GtkTextTag *sub;
	
			do {
				if( strcmp( "sub", subnode->name ) ) {
					subnode = subnode->next;
					continue;
				}
				val = xmlNodeGetContent( subnode );
				val = g_strstrip( val );
				
				sub = gtk_text_tag_table_lookup( table, val );
				
				if( sub ) {
					GtkTextTag *stag;

					if( *sub_patterns->str != '\0' ) {
						g_string_append_c( sub_patterns, '|' );
					}
					g_string_append( sub_patterns,
							 g_object_get_data( G_OBJECT( sub ),
									    "pattern" ) );
					
					stag = g_object_get_data( G_OBJECT( sub ),
								  "fore_tag" );
					if( stag ) {
						gtk_text_tag_table_remove( table,
									   stag );
						gtk_text_tag_table_add( table, 
									stag );
					}
					
					stag = g_object_get_data( G_OBJECT( sub ),
								  "back_tag" );
					if( stag ) {
						gtk_text_tag_table_remove( table,
									   stag );
						gtk_text_tag_table_add( table, 
									stag );
					}
					
					subs = g_list_prepend( subs, sub );
				} else {
					g_warning( "Missing <subgroup name=\"%s\">\n", val );
				}
				g_free( val );
				subnode = subnode->next;
			} while( subnode );
		}
		
		if( sub_patterns->len > 0 ) {
			sregex = g_new0( regex_t, 1 );
			regcomp( sregex, 
				 sub_patterns->str, 
				 REG_EXTENDED | REG_NEWLINE );
			g_object_set_data( G_OBJECT( tag ),
					   "sub_patterns", 
					   sregex );
			g_string_free( sub_patterns, TRUE );
		}

		gtk_text_tag_table_add( table, tag );
		if( isgroup ) {
			if( subs ) {
				subs = g_list_reverse( subs );
				g_object_set_data( G_OBJECT( tag ),
						   "subs", subs );
				subs = NULL;
			}
			tags = g_list_prepend( tags, tag );
		}
		/* keep a ref on a tag so its not destroyed
		   when / if we remove it from the table */
		g_object_ref( G_OBJECT( tag ) );
		
		g_free( tagname );
	}

	{
		regex_t *sregex;
		
		sregex = g_new0( regex_t, 1 );
		
		regcomp( sregex, start_patterns->str, REG_EXTENDED | REG_NEWLINE );
		g_object_set_data( G_OBJECT( table ), "start_patterns", 
				   sregex );
	}
	g_assert( tags );

	/* ensure tag rule is last */
	tag = gtk_text_tag_table_lookup( table, "default" );
	if( tag ) {
		for( tmp = tags; tmp; tmp = tmp->next ) {
			if( tmp->data == tag ) {
				tags = g_list_remove( tags, tag );
			}
		}
		tags = g_list_prepend( tags, tag );
	}
	
	/* fix up all sub tags so they have a greater priority than
	   non sub tags */
	for( tmp = tags; tmp; tmp = tmp->next ) {
		GtkTextTag *fore;
		GtkTextTag *back;
		GList *subs;
		
		tag = GTK_TEXT_TAG( tmp->data );
		
		subs = g_object_get_data( G_OBJECT( tag ), "subs" );
		for( ; subs; subs = subs->next ) {
			tag = GTK_TEXT_TAG( subs->data );
			fore = g_object_get_data( G_OBJECT( tag ),
						  "fore_tag" );
			back = g_object_get_data( G_OBJECT( tag ),
						  "back_tag" );
			
			if( fore ) {
				gtk_text_tag_table_remove( table, fore );
				gtk_text_tag_table_add( table, fore );
			}
			if( back ) {
				gtk_text_tag_table_remove( table, back );
				gtk_text_tag_table_add( table, back );
			}
		}
	}

	tags = g_list_reverse( tags );
	/* to improve the speed of hi_trigger_fine_match()
	   we need to prepend a fake item to the list */
	tags = g_list_prepend( tags, NULL );
	g_object_set_data( G_OBJECT( table ), "tags", tags );
	g_string_free( start_patterns, TRUE );


	g_object_set_data( G_OBJECT( table ), "doc", doc );
	g_object_set_data( G_OBJECT( table ), "filename", (gchar*)filename );

	return G_OBJECT( table );
}

static void screem_application_copy_tag_table( GtkTextTag *tag,
					       gpointer data )
{
	GtkTextTag *inctag;
	GtkTextTag *incftag;
	GtkTextTag *incbtag;
	GtkTextTag *ftag;
	GtkTextTag *btag;
	GdkColor *color;
	regex_t *regex;
	GtkTextTagTable *table;
	GString *string;
	gchar *value;

	gboolean isgroup;
	GList *tags;

	table = GTK_TEXT_TAG_TABLE( data );
	tags = g_object_get_data( G_OBJECT( table ), "tags" );

	inctag = tag;

	if( gtk_text_tag_table_lookup( table, inctag->name ) ) {
		return;
	}

	tag = gtk_text_tag_new( inctag->name );
	g_object_set_data( G_OBJECT( tag ), "included",
			   GUINT_TO_POINTER( 1 ) );
	
	g_object_ref( G_OBJECT( tag ) );
	
	/* copy fore tag */
	incftag = g_object_get_data( G_OBJECT( inctag ),
				     "fore_tag" );
	isgroup = TRUE;
	ftag = NULL;
	btag = NULL;
	if( incftag ) {
		ftag = gtk_text_tag_new( incftag->name );
		g_object_get( G_OBJECT( incftag ),
			      "foreground_gdk",
			      &color, NULL );
		g_object_set( G_OBJECT( ftag ),
			      "foreground_gdk",
			      color, NULL );
		if( g_object_get_data( G_OBJECT( incftag ),
				       "usefore" ) ) {
			g_object_set_data( G_OBJECT( ftag ),
					   "usefore",
					   GINT_TO_POINTER(1));
		}
		if( g_object_get_data( G_OBJECT( incftag ),
				       "sub" ) ) {
			g_object_set_data( G_OBJECT( ftag ),
					   "sub",
					   GINT_TO_POINTER(1));
			isgroup = FALSE;
		}
		gtk_text_tag_table_add( table, ftag );
		g_object_set_data( G_OBJECT( tag ), "fore_tag", ftag );
	}
	/* copy back tag */
	incbtag = g_object_get_data( G_OBJECT( inctag ),
				     "back_tag" );
	if( incbtag ) {
		btag = gtk_text_tag_new( incbtag->name );
		g_object_get( G_OBJECT( incbtag ),
			      "background_gdk",
			      &color, NULL );
		g_object_set( G_OBJECT( btag ),
			      "background_gdk",
			      color, NULL );
		if( g_object_get_data( G_OBJECT( incbtag ),
				       "useback" ) ) {
			g_object_set_data( G_OBJECT( btag ),
					   "useback",
					   GINT_TO_POINTER(1));
		}
		if( g_object_get_data( G_OBJECT( incbtag ),
				       "sub" ) ) {
			g_object_set_data( G_OBJECT( btag ),
					   "sub",
					   GINT_TO_POINTER(1));
			isgroup = FALSE;
		}
		gtk_text_tag_table_add( table, btag );
		g_object_set_data( G_OBJECT( tag ), "back_tag", btag );
	}

	if( ftag || btag ) {
		gtk_text_tag_table_add( table, tag );
	}
	
	if( isgroup && ( ftag || btag ) ) {
		tags = g_list_prepend( tags, tag );
		g_object_set_data( G_OBJECT( table ), "tags", tags );
	}

	/* copy regexps */
	regex = g_object_get_data( G_OBJECT( inctag ),
				   "start" );
	g_object_set_data( G_OBJECT( tag ), "start",
			   regex );
	regex = g_object_get_data( G_OBJECT( inctag ),
				   "end" );
	g_object_set_data( G_OBJECT( tag ), "end",
			   regex );
	
	regex = g_object_get_data( G_OBJECT( inctag ),
				   "sub_patterns" );
	g_object_set_data( G_OBJECT( tag ), 
			   "sub_patterns",
			   regex );


	string = (GString*)g_object_get_data( G_OBJECT( table ),
					      "start_patterns" );
	value = g_object_get_data( G_OBJECT( inctag ), "pattern" );
	g_object_set_data( G_OBJECT( tag ), "pattern", value );

	if( isgroup && value ) {
		if( *string->str != '\0' ) {
			g_string_append_c( string, '|' );
		}
		g_string_append( string, value );
	}
}

static void screem_application_copy_sub_tags( GtkTextTag *tag,
					      gpointer data )
{
	GtkTextTagTable *table;
	GtkTextTag *tg;
	GList *list;
	GList *subs;

	table = GTK_TEXT_TAG_TABLE( data );
	
	list = (GList*)g_object_get_data(G_OBJECT(tag),
					    "subs" );

	subs = NULL;
	for( ; list; list = list->next ) {
		tg = GTK_TEXT_TAG( list->data );
		tg = gtk_text_tag_table_lookup( table,
						tg->name );
		g_assert( tg );
		subs = g_list_prepend( subs, tg );
	}
	subs = g_list_reverse( subs );

	tag = gtk_text_tag_table_lookup( table, tag->name );
	g_assert( tag );

	g_object_set_data( G_OBJECT( tag ), "subs", subs );
}

GHashTable* screem_application_load_syntax_tables()
{
	static GHashTable *tables = NULL;
	static gboolean loaded = FALSE;

	gchar *syntax_path;
	gchar *version_check;

	GObject *table;
	gchar *filename;
	struct dirent **namelist;
	gint num;

	if( loaded ) {
		return tables;
	}

	loaded = TRUE;

	syntax_path = g_strconcat( g_get_home_dir(), G_DIR_SEPARATOR_S,
				   ".screem", G_DIR_SEPARATOR_S,
				   "syntax", G_DIR_SEPARATOR_S, NULL );

	if( ! tables ) {
		gint number;

		version_check = g_strconcat( g_get_home_dir(), G_DIR_SEPARATOR_S,
					     ".screem", G_DIR_SEPARATOR_S,
					     ".syntaxv2.2", NULL );
		
		gdk_threads_enter();
		if( ! uri_exists( syntax_path, NULL ) ||
		    ! uri_exists( version_check, NULL ) ) {
			if( ! copy_dir( SYNTAX_PATH, syntax_path, FALSE ) ) {
				g_free( syntax_path );
				g_free( version_check );
				gdk_threads_leave();
				return NULL;
			}
			save_file( version_check, "", GNOME_VFS_PERM_USER_READ |
				   GNOME_VFS_PERM_USER_WRITE );
		}
		gdk_threads_leave();
		g_free( version_check );


		tables = g_hash_table_new( g_str_hash, g_str_equal );

		num = scandir( syntax_path, &namelist, 0, alphasort );

		for( number = 0; number < num; ++ number ) {
			if(strcmp("..", namelist[ number ]->d_name ) &&
			   strcmp(".", namelist[ number ]->d_name) ) {
				filename = g_strconcat( syntax_path,
							namelist[ number ]->d_name,
							NULL );
				table = g_hash_table_lookup( tables, filename );
				if( ! table ) {
					table = screem_application_create_tag_table(tables,
										    filename );
				}
				if( table ) {
					g_hash_table_insert( tables,
							     filename,
							     table );
				} else {
					g_free( filename );
				}
			}
			free( namelist[ number ] );
		}
		if( num > 0 ) {
			free( namelist );
		}
	}

	g_free( syntax_path );

	return tables;
}

static void screem_application_add_icons( ScreemApplication *application )
{
	static struct ScreemStockPixmap {
		guchar const *stock_id;
		guchar const *filename;
		guchar const *ext;
	} const icons[] = {
		{ "Screem_Button", "button", ".xpm" },
		{ "Screem_Caption", "caption", ".xpm" },
		{ "Screem_Checkbutton", "checkbutton", ".xpm" },
		{ "Screem_Entry", "entry", ".xpm" },
		{ "Screem_FileEntry", "gnome-fileentry", ".xpm" },
		{ "Screem_Image", "stock_insert_image", ".png" },
		{ "Screem_Entity", "insert-symbol", ".png" },
		{ "Screem_Link", "add-link", ".png" },
		{ "Screem_Optionmenu", "optionmenu", ".xpm" },
		{ "Screem_Paragraph", "paragraphs", ".png" },
		{ "Screem_Pre", "pre", ".xpm" },
		{ "Screem_Radiobutton", "radiobutton", ".xpm" },
		{ "Screem_Sub", "sub", ".xpm" },
		{ "Screem_Sup", "sup", ".xpm" },
		{ "Screem_Table", "stock_insert_table", ".png" },
		{ "Screem_Td", "add_column", ".png" },
		{ "Screem_Text", "text", ".xpm" },
		{ "Screem_Th", "th", ".xpm" },
		{ "Screem_Tr", "add_row", ".png" },

		{ "Screem_Browser1", "browser", ".xpm" },
		{ "Screem_Browser2", "browser", ".xpm" },
		{ "Screem_Browser3", "browser", ".xpm" },

		{ "Screem_Site", "screem_site", ".png" },
		{ "Screem_Preview", "site_preview", ".png" },
		{ "Screem_LinkView", "site_structure", ".png" },
		{ "Screem_Resources", "resources", ".png" },

		{ "Screem_Todo", "todo", ".png" },

		{ "Screem_Bookmarks", "bookmarks-preferences", ".png" },
		{ "Screem_Bookmark", "bookmarks-open", ".png" },

		{ "Screem_Online", "stock_connect", ".png" },
		{ "Screem_Offline", "stock_disconnect", ".png" },

		{ "Screem_CVSCheckout", "cvs-checkout", ".png" },
		{ "Screem_CVSUpdate", "cvs-update", ".png" },
		{ "Screem_CVSAdd", "cvs-add", ".png" },
		{ "Screem_CVSRemove", "cvs-remove", ".png" },

		{ NULL, NULL, NULL }
	};
	static const GtkStockItem stockItems[] = {
		{ "Screem_Button", "<button>", 0, 0, 0 },
		{ "Screem_Caption", "<caption>", 0, 0, 0 },
		{ "Screem_Checkbutton", "Checkbox", 0, 0, 0 },
		{ "Screem_Entry", "text entry", 0, 0, 0 },
		{ "Screem_FileEntry", "file entry", 0, 0, 0 },
		{ "Screem_Image", "<img>", 0, 0, 0 },
		{ "Screem_Entity", "Entity", 0, 0, 0 },
		{ "Screem_Link", "<a>", 0, 0, 0 },
		{ "Screem_Optionmenu", "<option>", 0, 0, 0 },
		{ "Screem_Paragraph", "<p>", 0, 0, 0 },
		{ "Screem_Pre", "<pre>", 0, 0, 0 },
		{ "Screem_Radiobutton", "Radiobutton", 0, 0, 0 },
		{ "Screem_Sub", "<sub>", 0, 0, 0 },
		{ "Screem_Sup", "<sup>", 0, 0, 0 },
		{ "Screem_Table", "<table>", 0, 0, 0 },
		{ "Screem_Td", "<td>", 0, 0, 0 },
		{ "Screem_Text", "<textarea>", 0, 0, 0 },
		{ "Screem_Th", "<th>", 0, 0, 0 },
		{ "Screem_Tr", "<tr>", 0, 0, 0 },

		{ "Screem_Browser1", "Browser 1", 0, 0, 0 },
		{ "Screem_Browser2", "Browser 2", 0, 0, 0 },
		{ "Screem_Browser3", "Browser 3", 0, 0, 0 },

		{ "Screem_Site", "", 0, 0, 0 },

		{ "Screem_Preview", "Preview", 0, 0, 0 },
		{ "Screem_LinkView", "Link View", 0, 0, 0 },
		{ "Screem_Resources", "Resources", 0, 0, 0 },

		{ "Screem_Todo", "Task List", 0, 0, 0 },

		{ "Screem_Bookmarks", "Edit Bookmarks...", 0, 0, 0 },
		{ "Screem_Bookmark", "", 0, 0, 0 },

		{ "Screem_Online", "", 0, 0, 0 },
		{ "Screem_Offline", "", 0, 0, 0 },

		{ "Screem_CVSCheckout", "", 0, 0, 0 },
		{ "Screem_CVSUpdate", "", 0, 0, 0 },
		{ "Screem_CVSAdd", "", 0, 0, 0 },
		{ "Screem_CVSRemove", "", 0, 0, 0 }

	};

	GtkIconFactory *factory;
	gint i;

	factory = gtk_icon_factory_new();

	for( i = 0; icons[ i ].stock_id; ++ i ) {
		GtkIconSet *set;
		GtkIconSource *src;
		GdkPixbuf *pixbuf;
		gchar *filename;
		gint j;
		gboolean wildcarded;
		static const gchar *sizes[] = {
			"", "-48", "-24", "-16", NULL
		};

		set = gtk_icon_set_new();
		wildcarded = FALSE;

		for( j = 0; sizes[ j ]; ++ j ) {
			filename = g_strconcat( PIXMAP_PATH, G_DIR_SEPARATOR_S,
						icons[ i ].filename,
						sizes[ j ],
						icons[ i ].ext,
						NULL );
			
			pixbuf = gdk_pixbuf_new_from_file( filename, NULL );
			if( pixbuf ) {
				src = gtk_icon_source_new();
				gtk_icon_source_set_pixbuf( src, pixbuf );
				

				if( ! wildcarded ) {
					wildcarded = TRUE;
				} else {
					gtk_icon_source_set_size_wildcarded( src, FALSE );
				}
				
				if( j == 1 ) {
					gtk_icon_source_set_size( src,
								  GTK_ICON_SIZE_DIALOG );
				} else if( j == 2 ) {
					gtk_icon_source_set_size( src,
								  GTK_ICON_SIZE_BUTTON );
				} else if( j == 3 ) {
					gtk_icon_source_set_size( src,
								  GTK_ICON_SIZE_MENU );
				}
				gtk_icon_set_add_source( set, src );
				g_object_unref( G_OBJECT( pixbuf ) );
				gtk_icon_source_free( src );
			}
			g_free( filename );
		}
		gtk_icon_factory_add( factory, icons[ i ].stock_id, set );
		
		gtk_icon_set_unref( set );
	}

	gtk_icon_factory_add_default( factory );

	gtk_stock_add( stockItems, G_N_ELEMENTS( stockItems ) );

	g_object_unref( G_OBJECT( factory ) );
}



static void screem_application_load_macros( ScreemApplication *application )
{
	GSList *macro_list;
	GSList *macro_action_list;
	GSList *list;
	GSList *alist;

	macro_list = gconf_client_get_list( application->client,
					    "/apps/screem/editor/macros",
					    GCONF_VALUE_STRING, NULL );
	macro_action_list = 
		gconf_client_get_list(application->client,
				      "/apps/screem/editor/macro_actions",
				      GCONF_VALUE_STRING, NULL );
	
	for( alist = macro_action_list, list = macro_list; list && alist; 
	     list = list->next, alist = alist->next ) {
		ScreemEditorMacro *macro = g_new0( ScreemEditorMacro, 1 );
	
		macro->key_name = g_strdup( list->data );
		macro->text = g_strdup( alist->data );
		screem_application_add_macro( application, macro );
	}
	if( macro_list )
		g_slist_free( macro_list );

	if( macro_action_list )
		g_slist_free( macro_action_list );
}

void screem_application_add_macro( ScreemApplication *application,
				   ScreemEditorMacro *macro )
{
	GList *macros;
	GSList *mlist;
	GSList *malist;
	GSList *list;
	GSList *alist;

	/* insert into macro list */
	for( macros = application->macros; macros; macros = macros->next ) {
		ScreemEditorMacro *mac = (ScreemEditorMacro*)macros->data;

		if( ! strcmp( mac->key_name, macro->key_name ) )
			break;
	}
	if( macros ) {
		macros->data = macro;
	} else {
		application->macros = g_list_append( application->macros,
						     macro );
	}

	/* store via gconf */
	mlist = gconf_client_get_list( application->client,
				       "/apps/screem/editor/macros",
				       GCONF_VALUE_STRING, NULL );
	malist = gconf_client_get_list( application->client,
					"/apps/screem/editor/macro_actions",
				       GCONF_VALUE_STRING, NULL );

	for( list = mlist, alist = malist; list && alist; 
	     list = list->next, alist = alist->next ) {
		const gchar *mname = (const gchar*)mlist->data;
		if( mname && ! strcmp( mname, macro->key_name ) )
			break;
	}
	if( list && alist ) {
		alist->data = g_strdup( macro->text );
	} else {
		mlist = g_slist_append( mlist, g_strdup( macro->key_name ) );
		malist = g_slist_append( malist, g_strdup( macro->text ) );
	}

	gconf_client_set_list( application->client,
			       "/apps/screem/editor/macros",
			       GCONF_VALUE_STRING, mlist, NULL );
	gconf_client_set_list( application->client,
			       "/apps/screem/editor/macro_actions",
			       GCONF_VALUE_STRING, malist, NULL );

	gconf_client_suggest_sync( application->client, NULL );

	/* cleanup */
	if( mlist ) {
		g_slist_free( mlist );
	}
	if( malist ) {
		g_slist_free( malist );
	}
}

void screem_application_remove_macro( ScreemApplication *application,
				      ScreemEditorMacro *macro )
{
	GList *macros;

	for( macros = application->macros; macros; macros = macros->next ) {
		ScreemEditorMacro *mac = (ScreemEditorMacro*)macros->data;

		if( ! strcmp( mac->key_name, macro->key_name ) )
			break;
	}
	if( macros ) {
		ScreemEditorMacro *mac = (ScreemEditorMacro*)macros->data;

		GSList *macros;
		GSList *actions;
		GSList *list;
		GSList *alist;

		g_free( mac->key_name );
		g_free( mac->text );
		g_free( mac );
		application->macros = g_list_remove( application->macros,
						     mac );

	
		/* now remove from gconf */
		macros = gconf_client_get_list( application->client,
						"/apps/screem/editor/macros",
						GCONF_VALUE_STRING, NULL );
		actions = gconf_client_get_list( application->client,
						 "/apps/screem/editor/macro_actions",
						 GCONF_VALUE_STRING, NULL );
		for(list = macros, alist = actions; list && alist;
		    list = list->next, alist = alist->next ) {
			const gchar *mname = (const gchar*)list->data;
			if( mname && ! strcmp( mname, macro->key_name ) )
				break;
		}
		if( list ) {
			g_free( list->data );
			macros = g_slist_remove( macros, list->data );
		}
		if( alist ) {
			g_free( alist->data );
			actions = g_slist_remove( actions, alist->data );
		}
		gconf_client_set_list( application->client,
				       "/apps/screem/editor/macros",
				       GCONF_VALUE_STRING, macros, NULL );
		gconf_client_set_list( application->client,
				       "/apps/screem/editor/macro_actions",
				       GCONF_VALUE_STRING, actions, NULL );

		gconf_client_suggest_sync( application->client, NULL );

		/* cleanup */
		if( macros ) {
			g_slist_free( macros );
		}
		if( actions ) {
			g_slist_free( actions );
		}
	}
}

static void screem_application_load_helpers( ScreemApplication *application )
{
	GSList *helper_list;
	GSList *list;

	helper_list = gconf_client_get_list( application->client,
					     "/apps/screem/helpers",
					     GCONF_VALUE_STRING, NULL );
	for( list = helper_list; list; list = list->next ) {
		ScreemHelper *helper;
		gchar *name = NULL;
		gchar *pathname = NULL;
		gchar *iconpath = NULL;

		GSList *data_list;
		GSList *dlist;
		gchar *base_path;
	     
		helper = screem_helper_new( list->data, "", "",
					    SCREEM_HELPER_STDIN_PAGE,
					    SCREEM_HELPER_STDOUT_PAGE );
		base_path = screem_helper_get_gconf_path( helper );

		data_list = gconf_client_get_list( application->client,
						   base_path,
						   GCONF_VALUE_STRING,
						   NULL );
		dlist = data_list;

		g_free( base_path );
		g_object_unref( helper );


		if( data_list ) {
			name = data_list->data;
			data_list = data_list->next;
		}
		if( data_list ) {
			pathname = data_list->data;
			data_list = data_list->next;
		}
		if( data_list ) {
			iconpath = data_list->data;
			data_list = data_list->next;
		}
		if( name && pathname ) {
			helper = screem_helper_new( name, pathname,
						    iconpath,
						    SCREEM_HELPER_STDIN_PAGE,
						    SCREEM_HELPER_STDOUT_PAGE );
			screem_application_add_helper( application,
						       helper );
		}
		if( name ) {
			g_free( name );
		}
		if( pathname ) {
			g_free( pathname );
		}
		if( iconpath ) {
			g_free( iconpath );
		}
		g_slist_free( dlist );
	}
	if( helper_list ) {
		g_slist_free( helper_list );
	}
}

static void screem_application_helper_failed( ScreemHelper *helper,
					      ScreemApplication *application )
{
	GtkWidget *widget;
	const gchar *name;

	g_object_get( G_OBJECT( helper ), "name", &name, NULL );

	widget = gtk_message_dialog_new( NULL, 
					 GTK_DIALOG_MODAL,
					 GTK_MESSAGE_ERROR,
					 GTK_BUTTONS_CLOSE,
					 _( "Helper %s encountered a fatal error"),
					 name );
	gtk_window_set_wmclass( GTK_WINDOW( widget ),
				"Screem", "helper_failed" );
	gtk_dialog_run( GTK_DIALOG( widget ) );
	gtk_widget_destroy( widget );
}

static void screem_application_helper_error( ScreemHelper *helper,
					     gint status,
					     ScreemApplication *application )
{

	GtkWidget *widget;
	const gchar *name;

	g_object_get( G_OBJECT( helper ), "name", &name, NULL );

	widget = gtk_message_dialog_new( NULL, 
					 GTK_DIALOG_MODAL,
					 GTK_MESSAGE_ERROR,
					 GTK_BUTTONS_CLOSE,
					 _( "Helper %s returned errors"),
					 name );
	gtk_window_set_wmclass( GTK_WINDOW( widget ),
				"Screem", "helper_error" );
	gtk_dialog_run( GTK_DIALOG( widget ) );
	gtk_widget_destroy( widget );
}

void screem_application_add_helper( ScreemApplication *application,
				    ScreemHelper *helper )
{
	const gchar *name;
	const gchar *path;
	const gchar *icon;
	
	gchar *base_path;
	GSList *list;
	GSList *helpers;
	GSList *dlist;
	GSList *data_list;
	GList *windows;

	GConfClient *client;

	/* add to all open windows */
	for( windows = application->window_list; windows;
	     windows = windows->next ) {
		ScreemWindow *window;
		
		window = SCREEM_WINDOW( windows->data );
		screem_helper_add_to_window( helper, window );
	}

	client = application->client;

	base_path = screem_helper_get_gconf_path( helper );

	application->helpers = g_list_append( application->helpers, helper );

	/* hookup callbacks */
	g_signal_connect( G_OBJECT( helper ), "helper_failed",
			  G_CALLBACK( screem_application_helper_failed ),
			  application );
	g_signal_connect( G_OBJECT( helper ), "helper_error",
			  G_CALLBACK( screem_application_helper_error ),
			  application );

	g_object_get( G_OBJECT( helper ), 
		      "name", &name, "pathname", &path,
		      "iconpath", &icon, NULL );

	/* store via gconf */
	client = application->client;
	helpers = gconf_client_get_list( client, "/apps/screem/helpers",
					 GCONF_VALUE_STRING, NULL );

	/* we don't want to add one we already have so check for it,
	   based on name */
	for( list = helpers; list; list = list->next ) {
		const gchar *hname = (const gchar *)list->data;
		if( hname && ! strcmp( name, hname ) ) {
			/* found it */
			break;
		}
	}
	/* add to the list / update the list */
	if( list ) {
		dlist = gconf_client_get_list( client, base_path,
					       GCONF_VALUE_STRING, NULL );
		data_list = dlist;

		if( dlist ) {
			dlist = dlist->next;
		} else {
			dlist = g_slist_append( NULL, g_strdup( name ) );
			data_list = dlist;
		}
		if( dlist ) {
			g_free( dlist->data );
			dlist->data = g_strdup( path );
			dlist = dlist->next;
		} else {
			dlist = g_slist_append( dlist, g_strdup( path ) );
			dlist = g_slist_last( dlist );
		}
		if( dlist ) {
			g_free( dlist->data );
			dlist->data = g_strdup( icon );
			dlist = dlist->next;
		} else {
			dlist = g_slist_append( dlist, g_strdup( icon ) );
			dlist = g_slist_last( dlist );
		}
	} else {
		data_list = dlist = g_slist_append( NULL, g_strdup( name ) );
		dlist = g_slist_append( dlist, g_strdup( path ) );
		dlist = g_slist_append( dlist, g_strdup( icon ) );

		helpers = g_slist_append( helpers, g_strdup( name ) );
	}

	/* store helper */
	gconf_client_set_list( client, base_path, GCONF_VALUE_STRING,
			       data_list, NULL );

	/* store in list of helper names */
	gconf_client_set_list( client, "/apps/screem/helpers", 
			       GCONF_VALUE_STRING, helpers, NULL );

	gconf_client_suggest_sync( client, NULL );

	/* clean up */
	if( helpers ) {
		g_slist_free( helpers );
	}
	if( data_list ) {
		g_slist_free( data_list );
	}

	g_free( base_path );
}

void screem_application_remove_helper( ScreemApplication *application,
				       ScreemHelper *helper )
{
	GList *hlist;
	const gchar *name;

	g_object_get( G_OBJECT( helper ), "name", &name, NULL );

	/* first try and find a matching helper in application->helpers */
	for( hlist = application->helpers; hlist; hlist = hlist->next ) {
		const gchar *dname;

		g_object_get( G_OBJECT( hlist->data ), "name", &dname, NULL );

		if( ! strcmp( name, dname ) ) {
			break;
		}
	}
	if( hlist ) {
		GSList *helpers;
		GSList *list;
		GList *windows;

		/* found it */
		application->helpers = g_list_remove( application->helpers,
						      hlist->data );


		/* remove from all open windows */
		for( windows = application->window_list; windows;
		     windows = windows->next ) {
			ScreemWindow *window;

			window = SCREEM_WINDOW( windows->data );
			screem_helper_remove_from_window( helper, window );
		}

		/* now remove from gconf */
		helpers = gconf_client_get_list( application->client, 
						 "/apps/screem/helpers",
						 GCONF_VALUE_STRING, NULL );

		for( list = helpers; list; list = list->next ) {
			const gchar *hname = (const gchar *)list->data;
			if( hname && ! strcmp( name, hname ) ) {
				/* found it */
				break;
			}
		}
		if( list ) {
			gchar *base_path;
	
			helpers = g_slist_remove( helpers, list->data );
			g_free( list->data );

			gconf_client_set_list( application->client,
					       "/apps/screem/helpers",
					       GCONF_VALUE_STRING,
					       helpers, NULL );

			/* now remove from helpers_data */
			base_path = screem_helper_get_gconf_path( helper );
			
			gconf_client_unset( application->client,
					    base_path, NULL );

			gconf_client_suggest_sync( application->client, NULL );

			g_free( base_path );
		}
		/* clear up helpers list */
		g_slist_free( helpers );
	}
}

static void screem_application_offline_notify( GConfClient *client,
					       guint cnxn_id,
					       GConfEntry *entry,
					       gpointer data )
{
	ScreemApplication *application;

	application = SCREEM_APPLICATION( data );

	if( entry->value && entry->value->type == GCONF_VALUE_BOOL ) {
		gboolean state;

		state = gconf_value_get_bool( entry->value );

		if( application->offline != state ) {
			GList *list;

			application->offline = state;
	
			list = screem_application_get_window_list( application );
			g_list_foreach( list, 
					(GFunc)screem_window_toggle_online_status,
					NULL );
		}
	}
}
     
/* G Object stuff */

#define PARENT_TYPE G_TYPE_OBJECT

static gpointer parent_class;

static void screem_application_class_init( ScreemApplicationClass *klass )
{
	GObjectClass *object_class;

	object_class = G_OBJECT_CLASS( klass );
	parent_class = g_type_class_peek_parent( klass );

	object_class->finalize = screem_application_finalize;
}

static void screem_application_init( ScreemApplication *application )
{
	/* we properly init later by calling screem_application_startup */

}

static void screem_application_finalize( GObject *object )
{
	ScreemApplication *application;

	application = SCREEM_APPLICATION( object );

	application->finalizing = TRUE;

	screem_application_close_all_windows( application );

	screem_application_close_all_sites( application );

	screem_php_destroy_function_hash_table();

	gconf_client_notify_remove( application->client, application->offline_notify );

	g_object_unref( application->client );

	G_OBJECT_CLASS( parent_class )->finalize( object );
}

GType screem_application_get_type()
{

	static GType type = 0;

	if( ! type ) {
		static const GTypeInfo info = {
			sizeof( ScreemApplicationClass ),
			NULL, /* base init */
			NULL, /* base finalise */
			(GClassInitFunc)screem_application_class_init,
			NULL, /* class finalise */
			NULL, /* class data */
			sizeof( ScreemApplication ),
			0, /* n_preallocs */
			(GInstanceInitFunc)screem_application_init
		};

		type = g_type_register_static( PARENT_TYPE,
					       "ScreemApplication",
					       &info, 0 );
	}

	return type;
}
