/*  Screem:  screem-dtd.c
 *
 *  DOCTYPE parsing / validity checking functions
 *
 *  Copyright (C) 2002 David A Knight
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  For contact information with the author of this source code please see
 *  the AUTHORS file.  If there is no AUTHORS file present then check the
 *  about box under the help menu for a contact address
 */

#include <config.h>

#include <ctype.h>

#include <gmodule.h>

#include <glib/gstrfuncs.h>

#include <libgnomevfs/gnome-vfs.h>

#include <libgnome/gnome-i18n.h>

#include <gtk/gtkmain.h>
#include <gtk/gtkmenu.h>
#include <gtk/gtkmenuitem.h>

#include <string.h>

#include "screemmarshal.h"
#include "screem-dtd.h"

/* nasty hack */
typedef void (*InsertAttrCb)(GtkWidget *item, const gchar *attribute);

static void screem_dtd_initialise_class( ScreemDTDClass *klass );
static void screem_dtd_initialise( ScreemDTD *dtd );
static void screem_dtd_finalize( GObject *object );

static void screem_dtd_parse( ScreemDTD *dtd, GString *dtd_data, 
			      ScreemDTD *parent );

static void screem_dtd_parse_entity( ScreemDTD *dtd, gchar *tag );


static void screem_dtd_parse_element( ScreemDTD *dtd, gchar *tag );
static void screem_dtd_parse_allowed_elements( ScreemDTD *dtd,
					       ScreemDTDElement *element,
					       const gchar *tag );
static const gchar *screem_dtd_allowed( ScreemDTD *dtd, const gchar *tag,
					ScreemDTDElement *element,
					GSList *block,
					gint level );
static void screem_dtd_remove_allowed( ScreemDTDElement *element,
				       GString *tag );


static void screem_dtd_parse_attlist( ScreemDTD *dtd, gchar *tag );
static GSList* screem_dtd_parse_attribute( ScreemDTD *dtd, 
					   const gchar *tag,
					   const gchar *element );

static gchar* screem_dtd_parse_name( ScreemDTD *dtd, 
				     const gchar *tag, 
				     GString *full );

static gchar* screem_dtd_get_entity_value( ScreemDTD *dtd, const gchar *name,
					   gboolean isent );

static const gchar *screem_dtd_lookup_entity( ScreemDTD *dtd, 
					      const gchar *name );
static ScreemDTDElement *screem_dtd_lookup_element( ScreemDTD *dtd, 
						    const gchar *name,
						    gboolean casesens );


static void screem_dtd_set_root_element( ScreemDTD *dtd, const gchar *name );

/* TEMPORARY */
static gchar* next_tag_( const gchar *text, gint *pos, gchar **name );

static void screem_dtd_free_data( gpointer data, GClosure *closure );

static gboolean screem_dtd_copy_entities( gpointer key, gpointer value,
					  gpointer data );
static gboolean screem_dtd_copy_elements( gpointer key, gpointer value,
					  gpointer data );
static gboolean screem_dtd_copy_attlist( gpointer key, gpointer value,
					 gpointer data );
static gboolean screem_dtd_check_root( gpointer key, gpointer value,
				       gpointer data );
static gboolean screem_dtd_sort_elements( gpointer key, gpointer value,
					  gpointer data );
static gboolean screem_dtd_merge_attlist( gpointer key, gpointer value,
					  gpointer data );
static gint screem_dtd_compare_allowed( gconstpointer a, gconstpointer b );
static gint screem_dtd_compare_attributes( gconstpointer a, gconstpointer b );

static gboolean g_casestr_func( gpointer key, gpointer value, gpointer data );

static ScreemDTDElement *screem_dtd_element_copy( const ScreemDTDElement *element );
static ScreemDTDAttribute *
screem_dtd_copy_attribute( const ScreemDTDAttribute *attr );

enum { ANAME, AVAL, ADEF, ANEXT };
enum { NAME, ENTITY, EXPR };

enum {
	LOAD_MODULE,
	LAST_SIGNAL
};
static guint screem_dtd_signals[LAST_SIGNAL] = { 0 };


struct ScreemDTDElement {
	gchar *name;

	ScreemDTDClose close;

	GSList *allowed;
	GSList *attributes;
};


typedef struct ScreemDTDAllowed {
	gchar *name;

	gint min;
	gint max;

	gboolean set;
} ScreemDTDAllowed;


struct ScreemDTDAttribute {
	gchar *name;
	gchar *default_value;

	GSList *values;

	gboolean isuri;
};


struct ScreemDTDPrivate {
	gchar *baseuri;

	/* this is the base uri of a dtd, if a dtd has a "parent"
	   then it will be the uri of the parent 
	   (ie if a dtd includes others)*/
	gchar *dtduri;

	gchar *id;

	GHashTable *entities;
	GHashTable *elements;

	GHashTable *attlists;

	GSList *sorted_elements;

	gboolean loaded;

	const gchar *root;

	ScreemDTD *parent;
	ScreemDTD *self;
};

#define PARENT_TYPE G_TYPE_OBJECT

static gpointer parent_class;


static void screem_dtd_initialise_class( ScreemDTDClass *klass )
{
	GObjectClass *object_class;
	GtkWidgetClass *widget_class;

	object_class = G_OBJECT_CLASS( klass );
	parent_class = g_type_class_peek_parent( klass );
	widget_class = (GtkWidgetClass *)klass;

	object_class->finalize = screem_dtd_finalize;

	screem_dtd_signals[ LOAD_MODULE ] = 
		g_signal_new( "load_module",
			      G_OBJECT_CLASS_TYPE( object_class ),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET( ScreemDTDClass, load_module ),
			      NULL, NULL,
			      screem_marshal_VOID__STRING_POINTER,
			      G_TYPE_NONE, 2,
			      G_TYPE_STRING,
			      G_TYPE_POINTER );
}

static void screem_dtd_initialise( ScreemDTD *dtd )
{
	ScreemDTDPrivate *priv;

	priv = dtd->private = g_new0( ScreemDTDPrivate, 1 );

	priv->self = dtd;

	priv->entities = g_hash_table_new( g_str_hash, g_str_equal );
	priv->elements = g_hash_table_new( g_str_hash, g_str_equal );
	priv->attlists = g_hash_table_new( g_str_hash, g_str_equal );
}

static void screem_dtd_finalize( GObject *object )
{
	ScreemDTD *dtd;
	ScreemDTDPrivate *priv;

	dtd = SCREEM_DTD( object );

	priv = dtd->private;

	if( priv->baseuri ) {
		g_free( priv->baseuri );
	}
	if( priv->dtduri ) {
		g_free( priv->dtduri );
	}
	if( priv->id ) {
		g_free( priv->id );
	}

	g_free( priv );

	G_OBJECT_CLASS( parent_class )->finalize( object );
}


GType screem_dtd_get_type()
{
	static GType type = 0;

	if( ! type ) {
		static const GTypeInfo info = {
			sizeof( ScreemDTDClass ),
			NULL, /* base init */
			NULL, /* base finalise */
			(GClassInitFunc)screem_dtd_initialise_class,
			NULL, /* class finalise */
			NULL, /* class data */
			sizeof( ScreemDTD ),
			0, /* n_preallocs */
			(GInstanceInitFunc)screem_dtd_initialise
		};

		type = g_type_register_static( PARENT_TYPE,
					       "ScreemDTD",
					       &info, 0 );
	}

	return type;
}

ScreemDTD *screem_dtd_new()
{
	return SCREEM_DTD( g_object_new( screem_dtd_get_type(), NULL ) );
}


ScreemDTD *screem_dtd_new_with_id( const gchar *id )
{
	ScreemDTD *dtd;

	dtd = screem_dtd_new();
	dtd->private->id = g_strdup( id );

	return dtd;
}

gboolean screem_dtd_load( ScreemDTD *dtd, const gchar *uri, ScreemDTD *parent,
			  gboolean online )
{
	GnomeVFSHandle *handle;
	GnomeVFSResult result;
	GnomeVFSURI *vfsuri;
	GnomeVFSFileSize red;
	GString *string;
	gchar *buffer;
	gboolean ret;

	ScreemDTDPrivate *priv;

	priv = dtd->private;

	ret = FALSE;

	if( priv->loaded ) {
		return TRUE;
	}

	vfsuri = gnome_vfs_uri_new( uri );
	if( ! vfsuri ) {
		return ret;
	}
	if( ! online && ! gnome_vfs_uri_is_local( vfsuri ) ) {
		gnome_vfs_uri_unref( vfsuri );
		return FALSE;
	}
	
	priv->baseuri = g_dirname( uri );

	priv->parent = parent;

	if( ! parent ) {
		priv->dtduri = g_strdup( priv->baseuri );
	} else {
		priv->dtduri = g_strdup( parent->private->dtduri );
	}

	handle = NULL;
	result = gnome_vfs_open_uri( &handle, vfsuri, GNOME_VFS_OPEN_READ );

	if( result != GNOME_VFS_OK ) {
		/* its possible that the uri given was a relative one,
		   if we have a parent we can try prepending it's baseuri
		   to uri and then try to open that */
		gchar *tmp;

		if( handle ) {
			gnome_vfs_close( handle );
			handle = NULL;
		}

		gnome_vfs_uri_unref( vfsuri );

		if( parent ) {
			tmp = g_strconcat( parent->private->baseuri, "/", uri, NULL );
			vfsuri = gnome_vfs_uri_new( tmp );
			g_free( tmp );
		} else {
			vfsuri = NULL;
		}

		if( vfsuri ) {
			if( ! online && ! gnome_vfs_uri_is_local( vfsuri ) ) {
				gnome_vfs_uri_unref( vfsuri );
				return FALSE;
			}
       			result = gnome_vfs_open_uri( &handle, vfsuri,
						     GNOME_VFS_OPEN_READ );
		}
	}
	if( result != GNOME_VFS_OK ) {
		/* still failed, check the dtduri */
		gchar *tmp;

		if( handle ) {
			gnome_vfs_close( handle );
			handle = NULL;
		}

		gnome_vfs_uri_unref( vfsuri );

		tmp = g_strconcat( priv->dtduri, "/", uri, NULL );
		vfsuri = gnome_vfs_uri_new( tmp );
		g_free( tmp );

		if( vfsuri ) {
			if( ! online && ! gnome_vfs_uri_is_local( vfsuri ) ) {
				gnome_vfs_uri_unref( vfsuri );
				return FALSE;
			}
			result = gnome_vfs_open_uri( &handle, vfsuri,
						     GNOME_VFS_OPEN_READ );
		}
	}
	if( result != GNOME_VFS_OK ) {
		if( handle ) {
			gnome_vfs_close( handle );
			handle = NULL;
		}
	} else {
		buffer = g_new0( gchar, BUFSIZ + 1 );
		string = g_string_new( "" );
		while( gnome_vfs_read( handle, buffer,
				       BUFSIZ, &red ) == GNOME_VFS_OK ) {
			buffer[ red ] = '\0';
			g_string_append( string, buffer );
		}
		buffer[ red ] = '\0';
		g_string_append( string, buffer );
		g_free( buffer );
		
		gnome_vfs_close( handle );
		
		/* save dtd if vfsuri is not local */
		if( ( ! gnome_vfs_uri_is_local( vfsuri ) ) && priv->id ) {
			gchar *path;
			gchar *filename;
			gchar *catalog;
			FILE *out;
			struct stat s;
			gint i;

			path = g_strconcat( g_get_home_dir(), 
					    G_DIR_SEPARATOR_S,
					    ".screem",
					    G_DIR_SEPARATOR_S,
					    "dtds", 
					    G_DIR_SEPARATOR_S,
					    NULL );
			i = 0;
			do {
				filename = g_strdup_printf( "%s%i.dtd",
							    path, i ++ );
			} while( ! stat( filename, &s ) );

			out = fopen( filename, "w" );
			if( out ) {
				fprintf( out, "%s", string->str );
				fclose( out );

				catalog = g_strconcat( path, "catalog", NULL );
				
				out = fopen( catalog, "a" );
				if( out ) {
					fprintf( out,
						 "PUBLIC \"%s\" \"%s\"\n", priv->id,
						 filename );
					fclose( out );
				}
				g_free( catalog );
			}
			g_free( filename );
			g_free( path );
		}

		screem_dtd_parse( dtd, string, parent );
		
		g_string_free( string, TRUE );

		ret = priv->loaded;
	}
	gnome_vfs_uri_unref( vfsuri );

	priv->parent = NULL;

	return ret;
}


ScreemDTDElement *screem_dtd_valid_element( ScreemDTD *dtd, const gchar *element )
{
	const gchar *id;
	ScreemDTDElement *el;

	if( ! dtd ) {
		return NULL;
	}

	id = (const gchar*)g_object_get_data( G_OBJECT( dtd ), "type" );

	if( element ) {
		el = screem_dtd_lookup_element( dtd, element, TRUE );
	} else {
		el = NULL;
	}
	
	/* the problem is HTML allows tags of any case to be valid,
	   where as xml doesn't.   lets cheat a little, and see
	   if the doctype contains " HTML "  */
	if( (! el) && id ) {
		if( strstr( id, " HTML " ) || strstr( id, " html " ) ||
		    strstr( id, " HTML/" ) || strstr( id, " html/" ) ) {
			el = screem_dtd_lookup_element( dtd, element,
							FALSE );
		}
	}

	return el;
}

ScreemDTDAttribute *screem_dtd_valid_attr( ScreemDTD *dtd, const gchar *element,
					   const gchar *attr )
{
	const gchar *id;

	ScreemDTDElement *el;
	ScreemDTDAttribute *sattr;
	GSList *list;

	id = (const gchar*)g_object_get_data( G_OBJECT( dtd ), "type" );

	el = screem_dtd_valid_element( dtd, element );
	
	/* if the element isn't valid then the attribute cant be */
	if( ! el ) {
		return NULL;
	}

	sattr = NULL;
	for( list = el->attributes; list; list = list->next ) {
		sattr = (ScreemDTDAttribute*)list->data;

		if( ! strcmp( attr, sattr->name ) ) {
			break;
		}
		sattr = NULL;
	}
	/* the problem is HTML allows tags of any case to be valid,
	   where as xml doesn't.   lets cheat a little, and see
	   if the doctype contains " HTML "  */
	if( (! sattr) && id ) {
		if( strstr( id, " HTML " ) || strstr( id, " html " ) ||
		    strstr( id, " HTML/" ) || strstr( id, " html/" ) ) {
			for( list = el->attributes; list; list = list->next ) {
				sattr = (ScreemDTDAttribute*)list->data;
				
				if( ! g_strcasecmp( attr, sattr->name ) ) {
					break;
				}
				sattr = NULL;
			}
		}
	}

	return sattr;
}

gboolean screem_dtd_valid_child( ScreemDTD *dtd, const gchar *element,
				 const gchar *child )
{
	const gchar *id;

	ScreemDTDElement *el;
	ScreemDTDAllowed *schild;
	GSList *list;
	
	if( ! dtd ) {
		return FALSE;
	}

	id = (const gchar*)g_object_get_data( G_OBJECT( dtd ), "type" );

	if( element ) {
		el = screem_dtd_valid_element( dtd, element );
	} else { 
		el = NULL;
	}
	
	/* if the element isn't valid then the attribute cant be */
	if( ! el ) {
		return FALSE;
	}

	schild = NULL;
	for( list = el->allowed; list; list = list->next ) {
		schild = (ScreemDTDAllowed*)list->data;

		if( ! strcmp( child, schild->name ) ) {
			break;
		}
		schild = NULL;
	}
	/* the problem is HTML allows tags of any case to be valid,
	   where as xml doesn't.   lets cheat a little, and see
	   if the doctype contains " HTML "  */
	if( (! schild) && id ) {
		if( strstr( id, " HTML " ) || strstr( id, " html " ) ||
		    strstr( id, " HTML/" ) || strstr( id, " html/" ) ) {
			for( list = el->allowed; list; list = list->next ) {
				schild = (ScreemDTDAllowed*)list->data;
				
				if( ! g_strcasecmp( child, schild->name ) ) {
					break;
				}
				schild = NULL;
			}
		}
	}

	return ( schild != NULL );
}

GtkWidget *screem_dtd_attribute_menu( ScreemDTD *dtd, const gchar *element,
				      gpointer data )
{
	GtkWidget *menu;
	GtkWidget *item;
	GtkWidget *subitem;
	GtkWidget *submenu;

	gpointer editor;
	InsertAttrCb attr_cb;
	GModule *self;

	ScreemDTDElement *el;
	GSList *list;
	GSList *values;

	if( ! dtd ) {
		return NULL;
	}

	editor = data;

	el = screem_dtd_valid_element( dtd, element );

	if( ! el ) {
		return NULL;
	}

	item = gtk_menu_item_new_with_label( element );
	g_object_set_data( G_OBJECT( item ), "editor", editor );
	gtk_widget_show( item );

	menu = gtk_menu_new();
	gtk_menu_item_set_submenu( GTK_MENU_ITEM( item ), menu );

	/* avoid including screem-editor.h,
	   dynamically lookup the cb */
	self = g_module_open( NULL, G_MODULE_BIND_LAZY );
	if( ! g_module_symbol( self,
			       "screem_editor_insert_attr_cb",
			       (gpointer*) &attr_cb ) ) {
		return NULL;
	}

	for( list = el->attributes; list; list = list->next ) {
		ScreemDTDAttribute *attr;

		attr = (ScreemDTDAttribute*)list->data;

		subitem = gtk_menu_item_new_with_label( attr->name );
		g_object_set_data( G_OBJECT( subitem ), "editor", editor );
		gtk_widget_show( subitem );
		gtk_menu_shell_append( GTK_MENU_SHELL( menu ), subitem );

		submenu = NULL;
		if( attr->values ) {
			submenu = gtk_menu_new();
			gtk_menu_item_set_submenu( GTK_MENU_ITEM( subitem ),
						   submenu );
		}
		for( values = attr->values; values; values = values->next ) {
			gchar *val;
			gchar *data;

			data = g_strdup( values->data );
			val = screem_dtd_get_entity_value( dtd,
							   data, 
							   FALSE );
			if( ! val ) {
				val = g_strdup( data );
			}
			g_free( data );
			
			subitem = gtk_menu_item_new_with_label( val );
			g_object_set_data( G_OBJECT( subitem ), 
					   "editor", editor );
			gtk_widget_show( subitem );
			
			data = g_strconcat( attr->name, "=\"", val, "\"",
					    NULL );
			g_free( val );
			g_signal_connect_data( G_OBJECT( subitem ),
					       "activate",
					       G_CALLBACK(attr_cb),
					       data,
					       screem_dtd_free_data,
					       0 );
			
			gtk_menu_shell_append( GTK_MENU_SHELL(submenu),
					       subitem );
		}
	}

	return item;
}

GtkWidget *screem_dtd_allowed_menu( ScreemDTD *dtd, const gchar *element )
{
	GtkWidget *menu;
	GtkWidget *item;
	GtkWidget *subitem;
	ScreemDTDElement *el;
	
	item = NULL;

	if( dtd ) {
		el = screem_dtd_valid_element( dtd, element );
		
		if( el ) {
			GSList *list;
			
			item = gtk_menu_item_new_with_label( _( "Valid Element") );
			gtk_widget_show( item );

			menu = NULL;

			if( el->allowed ) {
				menu = gtk_menu_new();
				gtk_menu_item_set_submenu( GTK_MENU_ITEM(item),
							   menu );
			}
			for( list = el->allowed; list; list = list->next ) {
				ScreemDTDAllowed *allow;

				allow = (ScreemDTDAllowed*)list->data;

				subitem = gtk_menu_item_new_with_label( allow->name );
				gtk_widget_show( subitem );
				gtk_menu_append( GTK_MENU( menu ), subitem );
			}
		}
	}

	return item;
}

void screem_dtd_screen_dump( ScreemDTD *dtd )
{
	GSList *list;

	g_print( "ASSUMED ROOT ELEMENT FOR %p: %s\n\n", 
		 dtd,
		 dtd->private->root );

	for( list = dtd->private->sorted_elements; list; list = list->next ) {
		ScreemDTDElement *element;
		GSList *attrs;
		GSList *sub;

		element = (ScreemDTDElement*)list->data;

		g_print( "ELEMENT: %s\n", element->name );
		for( attrs = element->attributes; attrs; attrs = attrs->next ) {
			ScreemDTDAttribute *attr;

			attr = (ScreemDTDAttribute*)attrs->data;
			g_print( "\tATTRIBUTE: %s\n", attr->name );
		}
		for( sub = element->allowed; sub; sub = sub->next ) {
			ScreemDTDAllowed *allow;

			allow = (ScreemDTDAllowed*)sub->data;
			g_print( "\tALLOWED: %s\n", allow->name );
		}
		g_print( "\n" );
	}
}

static void screem_dtd_parse( ScreemDTD *dtd, GString *dtd_data, 
			      ScreemDTD *parent )
{
	gchar *data;
	gchar *current;
	gchar *name;
	GString *full;
	gint pos;
	gint pos2;
	gboolean dtd_switch;

	ScreemDTDPrivate *priv;

	priv = dtd->private;

	data = dtd_data->str;
	pos = 0;


#define DTD_STAGE1(s)  !g_strncasecmp( s, current, strlen( s ) )

	priv->loaded = TRUE;

	while( data && data[ pos ] != '\0' ) {
		switch( data[ pos ] ) {
		case '%':
			pos ++;
			full = g_string_new( "" );
			while( data[ pos ] != ';' && data[ pos ] != '\0' ) {
				g_string_append_c( full, data[ pos ] );
				pos ++;
			}
			name = screem_dtd_get_entity_value( dtd, full->str,
							    TRUE );
			
			if( name ) {
				ScreemDTD *moddtd;

				moddtd = NULL;
				g_signal_emit( G_OBJECT( dtd ),
					       screem_dtd_signals[LOAD_MODULE],
					       0, name, &moddtd );
					       
				if( moddtd ) {
					/* got it, now copy the elements /
					   attributes / entities into this
					   ScreemDTD */
					g_hash_table_foreach( moddtd->private->elements, 
							      (GHFunc)screem_dtd_copy_elements, dtd );
					g_hash_table_foreach( moddtd->private->entities, 
							      (GHFunc)screem_dtd_copy_entities, dtd );

				
					g_hash_table_foreach( moddtd->private->attlists,
							      (GHFunc)screem_dtd_copy_attlist, dtd );
			
				}

				g_free( name );
			}
			g_string_free( full, TRUE );
			break;
		case '<':
			pos2 = pos;
			current = next_tag_( data + pos, &pos, NULL );

			g_assert( current != NULL );
			
			if( DTD_STAGE1( "<!--" ) ) {
				/* a comment, we ignore this */
			} else if( DTD_STAGE1( "<!ENTITY" ) ) {
				/* its an ENTITY */
				screem_dtd_parse_entity( dtd, current );
			} else if( DTD_STAGE1( "<!ELEMENT" ) ) {
				/* its an ELEMENT */
				screem_dtd_parse_element( dtd, current );
			} else if( DTD_STAGE1( "<!ATTLIST" ) ) {
				screem_dtd_parse_attlist( dtd, current );
			} else if( DTD_STAGE1( "<![" ) ) {
				/* its a feature switch */
				name = current + strlen( "<![" );
				full = g_string_new( "" );
				while( ! isalpha( *name ) && *name != '\0' ) {
					name ++;
				}
				while( !isspace( *name ) &&
				       *name != ';' && *name != '\0' ) {
					g_string_append_c( full, *name );
					name ++;
				}

				dtd_switch = FALSE;

				name = screem_dtd_get_entity_value( dtd,
								    full->str,
								    TRUE );

				g_string_free( full, TRUE );

				if( name ) {
					dtd_switch = ! strcmp( "INCLUDE", 
							       name );

					g_free( name );
				}
			
				if( ! dtd_switch ) {
					/* not switched on */
					gchar *end;

					end = strstr( &data[ pos ], "]]>" );
					if( end ) {
						pos = ( end - data );
					} else {
						pos = strlen( data );
					}
				} else {
					/* is switched on */
					pos = pos2 + 1;
					while( data[ pos ] != '<' && 
					       data[ pos ] != '\0' ) {
						pos ++;
					}
				}
			}

			g_free( current );

			break;
		default:
			pos ++;
			break;
		}
	}

	/* merge attlists */
	g_hash_table_foreach( priv->attlists,
			      (GHFunc)screem_dtd_merge_attlist,
			      dtd );

	/* sort elements, element allowed, and element attributes */
      	g_hash_table_foreach( priv->elements,
			      (GHFunc)screem_dtd_sort_elements,
			      priv );
}

static void screem_dtd_parse_entity( ScreemDTD *dtd, gchar *tag )
{
	GString *name;
	GString *value;
	gchar term;

	ScreemDTDPrivate *priv;

	priv = dtd->private;

	tag += strlen( "<!ENTITY " );

	if( *tag == '%' ) {
		tag ++;
		if( *tag == ' ' ) {
			tag ++;
		}
	}

	/* get entity name */
	name = g_string_new( "" );
	while( *tag != '\0' && ! isspace( *tag ) ) {
		g_string_append_c( name, *tag );
		tag ++;
	}

	term = '\0';
      	while( *tag != '\0' ) {
		term = *tag;
		if( ! strncmp( "--", tag, strlen( "--" ) ) ) {
			tag = strstr( tag + 2, "--" );
			if( ! tag ) {
				break;
			}
			tag += 2;
			term = *tag;
		}
		if( term == '"' || term == '\'' ) {
			break;
		} else if( ! isspace( term ) ) {
			tag --;
			term = '>';
			break;
		}
		tag ++;
	}

	if( tag ) {
		/* get value */
		value = g_string_new( "" );
		tag ++;
		while( *tag != '\0' && *tag != term ) {
			g_string_append_c( value, *tag );
			tag ++;
		}

		if( ! g_hash_table_lookup( priv->entities, name->str ) ) {
			g_hash_table_insert( priv->entities,
					     name->str,
					     g_strchug( value->str ) );
			g_string_free( name, FALSE );
			g_string_free( value, FALSE );
		} else {
			g_string_free( name, TRUE );
			g_string_free( value, TRUE );
		}
	} else {
		g_string_free( name, TRUE );
	}
}

static void screem_dtd_parse_element( ScreemDTD *dtd, gchar *tag )
{
	ScreemDTDPrivate *priv;
	GString *name;
	gchar **split;
	gint i;

	ScreemDTDElement *element;
	GSList *allowed;
	ScreemDTDClose close_state;

	priv = dtd->private;

	tag += strlen( "<!ELEMENT " );

	name = g_string_new( "" );

	tag = screem_dtd_parse_name( dtd, tag, name );

	if( ! name->len ) {
		/* discard this element, it has no name! */
		g_string_free( name, TRUE );
		return;
	}

	/* split the name up into its individual element names */
	split = g_strsplit( name->str, "|", -1 );
	g_string_free( name, TRUE );
	element = NULL;
	allowed = NULL;

	close_state = MUST;

	for( i = 0; split[ i ]; i ++ ) {
		gchar *sname;
		sname = g_strstrip( split[ i ] );

		if( *sname != '\0' &&
		    ! g_hash_table_lookup( priv->elements, sname ) ) {

			element = g_new0( ScreemDTDElement, 1 );
			element->name = g_strdup( sname );
			
			g_hash_table_insert( priv->elements,
					     element->name,
					     element );
		
			if( ! allowed ) {
				/* first one so get allowed sub elements */
				screem_dtd_parse_allowed_elements( dtd, 
								   element, 
								   tag );
				allowed = element->allowed;
				close_state = element->close;
			} else {
				/* copy allowed elements */
				element->allowed = g_slist_copy( allowed );
				element->close = close_state;
			}
		}
	}
	g_strfreev( split );

}

static void screem_dtd_parse_allowed_elements( ScreemDTD *dtd,
					       ScreemDTDElement *element,
					       const gchar *tag )
{
	gint offset;

	while( isspace( *tag ) ) {
		tag ++;
	}

	if( ! strncmp( "O O ", tag, strlen( "O O " ) ) ) {
		element->close = SHOULD;
		offset = strlen( "O O " );
	} else if( ! strncmp( "- O EMPTY", tag, strlen( "- O EMPTY" ) ) ) {
		element->close = MUSTNT;
		offset = strlen( "- O EMPTY" );
	} else if( ! strncmp( "- O ", tag, strlen( "- O " ) ) ) {
		element->close = SHOULD;
		offset = strlen( "- O " );
	} else if( ! strncmp( "EMPTY", tag, strlen( "EMPTY" ) ) ) {
		element->close = MUSTNT;
		offset = strlen( "EMPTY" );
	} else if( ! strncmp( "- - ", tag, strlen( "- - " ) ) ) {
		element->close = MUST;
		offset = strlen( "- - " );
	} else {
		element->close = MUST;
		offset = 0;
	}

	tag += offset;
	screem_dtd_allowed( dtd, tag, element, element->allowed, 0 );
}

static const gchar *screem_dtd_allowed( ScreemDTD *dtd, const gchar *tag,
					ScreemDTDElement *element,
					GSList *block,
					gint level )
{
	GString *name;
	static gboolean no;
	gboolean inent;
	ScreemDTDAllowed *allow;
	gchar *value;

	gboolean endBlock = FALSE;
	gchar minmax = '\0';

	name = g_string_new( "" );
	inent = FALSE;

	if( ! element )
		return NULL;

	if( level == 0 )
		no = FALSE;

	allow = NULL;

	while( tag && *tag != '>'  ) {

		/* bit of a hack */
		if( inent && *tag != ';' &&
		    ( ! isalnum( *tag ) && *tag != '.' &&
		      *tag != '_' && *tag != '-' ) ) {
			/* end entity, expand it */
			value = (gchar*)screem_dtd_lookup_entity( dtd, name->str );
			if( value ) {
				screem_dtd_allowed( dtd, value, 
						    element, block, level );
			} else {
				g_warning( "IN ALLOWED MISSING ENT: %s\n", name->str );
			}
			g_string_assign( name, "" );
			inent = FALSE;
		}

		switch( *tag ) {
		case '-':
			if( *(tag  + 1) == '-' ) {
				tag = strstr( tag + 2, "--" );
				if( tag ) {
					tag += 2;
				}
			} else if( name->len == 0 ) {
				/* don't add the next ones to the list of
				   allowed elements. (but only if the - isn't
				   occuring as part of an element name)*/
				no = TRUE;
			} else {
				g_string_append_c( name, '-' );
			}
			break;
		case '(':
			if( block ) {
				block = g_slist_last( block );
			}
			tag = screem_dtd_allowed( dtd, tag + 1, 
						  element,
						  block,
						  level + 1 );
			break;
		case '%':
			inent = TRUE;
			break;
		case ';':
			/* if we aren't in an entity, ie there is a rogue ';' 
			   then do nothing */
			if( ! inent )
				break;
			
			/* end entity, expand it */
			value = (gchar*)screem_dtd_lookup_entity( dtd, name->str );
			if( value ) {
				screem_dtd_allowed( dtd, value, 
						    element, block, level );
			} else {
				g_warning( "IN ALLOWED MISSING ENT: %s\n", name->str );
			}
			g_string_assign( name, "" );
			inent = FALSE;
			break;
		case '*':
		case '+':
		case '?':
			if( endBlock ) {
				gint min;
				gint max;
				if( *tag == '*' || *tag == '?' ) {
					min = 0;
				} else {
					min = 1;
				}
				if( *tag == '?' ) {
					max = 1;
				} else {
					max = -1;
				}

				if( ! block ) {
					block = element->allowed;
				}
				while( block ) {
					/* if min/max are already
					   set we don't reset */
					allow = (ScreemDTDAllowed*)block->data;
					if( ! allow->set ) {
						allow->min = min;
						allow->max = max;
					}
					if( ! block->next ) {
						break;
					}
					block = block->next;
				}

			} else {
				/* flag we need to set when we
				   create the node */
				minmax = *tag;
			}
			
			if( endBlock ) {
				block = NULL;
				endBlock = FALSE;
			}

			/* as this ends an expression block we return */
			if( level == 0 ) {
				return tag + 1;
			}
			break;
		case '\0':
		case ')':
			/* end of an expression */
			endBlock = TRUE;
		case ' ':
		case '\n':
		case '\r':
		case '|':
		case '&':   /* seen in IE 2.0 dtd */
		case ',':
			if( no && ( name->len > 0 ) ) {
				/* remove this word */
				screem_dtd_remove_allowed( element, name );
				g_string_assign( name, "" );
			} else if( name->len > 0 ) {
				/* end of element name */
				name->str = g_strstrip( name->str );
				g_string_truncate( name, strlen( name->str ) );
				if( strcmp( "EMPTY", name->str ) &&
				    name->len ) {
					allow = g_new0( ScreemDTDAllowed, 1 );
					allow->name = g_strdup( name->str );
					
					element->allowed = 
						g_slist_prepend(element->allowed,
								allow);

					if( minmax != '\0' ) {
						gint min;
						gint max;
						if( minmax == '*' || 
						    minmax == '?' ) {
							min = 0;
						} else {
							min = 1;
						}
						if( minmax == '?' ) {
							max = 1;
						} else {
							max = -1;
						}
						allow->min = min;
						allow->max = max;
					}
					minmax = '\0';
				}
				g_string_assign( name, "" );
			}
			if( level == 0 )
				no = FALSE;
			break;
		default:
			if( ! isspace( *tag ) ) {
				g_string_append_c( name, *tag );
			}
			break;
		}
		if( endBlock && *tag != ')' ) {
			block = NULL;
			endBlock = FALSE;
		}
		if( *tag == '\0' )
			break;
		if( tag )
			tag ++;
	}

	g_string_free( name, TRUE );

	return tag;
}

static void screem_dtd_remove_allowed( ScreemDTDElement *element,
				       GString *tag )
{
	GSList *list;
	ScreemDTDAllowed *allow;
	
	if( ! element || ! tag )
		return;

	for( list = element->allowed; list; list = list->next ) {
		
		allow = (ScreemDTDAllowed*)list->data;
		
		if( ! strcmp( allow->name, tag->str ) ) {
			g_free( allow->name );
			g_free( allow );
			break;
		}
	}
	if( list ) {
		element->allowed = g_slist_remove( element->allowed,
						   list->data );
	}
}

static void screem_dtd_parse_attlist( ScreemDTD *dtd, gchar *tag )
{
	GString *name;
	gchar **split;
	gint i;
	gchar *aname;
	ScreemDTDElement *element;
	GSList *attr;

	tag += strlen( "<!ATTLIST " );
	name = g_string_new( "" );
	tag = screem_dtd_parse_name( dtd, tag, name );

	if( ! name->len ) {
		/* discard this element as we didn't find a name */
		g_string_free( name, TRUE );
		return;
	}

	attr = screem_dtd_parse_attribute( dtd, tag, name->str );

	split = g_strsplit( name->str, "|", -1 );
	
	for( i = 0; split[ i ]; i ++ ) {
		aname = g_strstrip( split[ i ] );
		if( *aname != '\0' ) {
			element = screem_dtd_lookup_element( dtd, aname, TRUE );
			if( ! element ) {
				/* unknown element, possibly an error in
				   the DTD, or more likely a modular DTD
				   that is just appending attributes to
				   those allowed by an element */
				element = g_new0( ScreemDTDElement, 1 );
				element->name = g_strdup( aname );

				g_hash_table_insert( dtd->private->attlists,
						     element->name, element );
			}
			/* append attr to element */
			element->attributes =
				g_slist_concat( element->attributes,
						g_slist_copy( attr ) );
		}
	}
	g_strfreev( split );

	g_slist_free( attr );

	g_string_free( name, TRUE );
}

static GSList *screem_dtd_parse_attribute( ScreemDTD *dtd, 
					   const gchar *tag,
					   const gchar *element )
{
	GString *att_name;
	GString *att_value;
	GString *defval;

	gint state;
	gboolean isent;
	gboolean start;

	ScreemDTDAttribute *attr;
	GSList *attrs;
	GSList *new_attr;
	gchar **values;
	gchar *temp;
	gchar term;
	gint i;

	gboolean isuri;

	GString *temptag;

	attr = NULL;
	attrs = NULL;

	att_name = g_string_new( "" );
	att_value = g_string_new( "" );
	defval = g_string_new( "" );

	start = TRUE;
	state = ANEXT;
	isent = FALSE;
	isuri = FALSE;
	g_assert( tag );
     	tag = g_strchug( (gchar*)tag );

	temptag = g_string_new( "" );

	while( tag && ( *tag != '\0' || state == ANEXT ) ) {
		if( *tag == '>' ) {
			tag ++;
			state = ANEXT;
		}
		switch( state ) {
		case ANAME:
			if( isent && *tag != ';' ) {
				g_string_append_c( att_name, *tag );
				tag ++;
				break;
			} else if( ! isent && ! isspace( *tag ) ) {
				g_string_append_c( att_name, *tag );
				tag ++;
				break;
			} else if( ( ! isent ) && 
				   ( *att_name->str == '"' ||
				     *att_name->str == '\'' ) &&
				   ( *tag != *att_name->str ) ) {
				g_string_append_c( att_name, *tag );
				tag ++;
				break;
			}
			/* end of name */
			state = AVAL;
			
			/* is the name an entity? */
			if( isent ) {
				tag ++;
				temp = (gchar*)screem_dtd_lookup_entity( dtd,
									 att_name->str);
				state = ANEXT;
				if( temp ) {
					/* entity found, need to strdup as we g_strchug
					   it */
					GString *tmp;

					tmp = g_string_new( temp );
					g_string_append( tmp, tag );
					g_string_free( temptag, TRUE );
					temptag = tmp;
					tag = temptag->str;
				} else {
					g_print("couldn't find entity for %s\n",
						att_name->str );
				}
			}
			g_assert( tag );
			tag = g_strchug( (gchar*)tag );
			break;
		case AVAL:
			/* bit of a hack, determine if an attribute holds a URI
			   based on the entity name, HTML uses %URI; XHTML 1.1
			   uses %URI.datatype */
			isuri = ( ! strncmp( "%URI;", tag, strlen( "%URI;" ) ) ||
				  ! strncmp( "%URL;", tag, strlen( "%URL;" ) ) ||
				  ! strncmp( "%URL ", tag, strlen( "%URL " ) ) ||
				  ! strncmp( "%URI.datatype;", tag, 
					     strlen( "%URI.datatype;" ) ) );
			tag = screem_dtd_parse_name( dtd, tag, att_value );
			state = ADEF;
			if( tag ) {
				tag = g_strchug( (gchar*)tag );
			}
			break;
		case ADEF:
			if( ! isspace( *tag ) ) {
				gchar c;
				gchar *end;

				c = *tag;
				tag ++;
				if( c == '"' || c == '\'' ) {
					end = strchr( tag, c );
					if( ! end ) {
						end = tag + strlen( tag ) - 1;
					}
					g_string_append_len( defval,
							     tag,
							     end - tag );
					tag = end + 1;
				} else {
					g_string_append_c( defval, c );
				}
				
				if( *tag == '\0' )
					state = ANEXT;
				break;
			} else if( ! strcmp( "#FIXED", defval->str ) ) {
				tag = g_strchug( (gchar*)tag );
				term = *tag;
				tag ++;
				g_string_assign( att_value, "" );
				g_string_assign( defval, "" );
				while( *tag != '\0' && *tag != term ) {
					g_string_append_c( defval, *tag );
					tag ++;	
				}
				if( *tag == term )
					tag ++;
			}
			/* end of default, move to next */
			state = ANEXT;
			g_assert( tag );
			tag = g_strchug( (gchar*)tag );
			break;
		case ANEXT:
			if( ! start && ! isent && 
			    ( att_value->len != 0  || defval->len != 0 ) ) {
				/* build the attribute now */
				attr = g_new0( ScreemDTDAttribute, 1 );
				attr->name = g_strdup( att_name->str );
				attr->default_value = g_strdup( defval->str );
				attr->isuri = isuri;

				values = g_strsplit( att_value->str, "|", -1 );
				for( i = 0; values[ i ]; i ++ ) {
					values[ i ] = g_strstrip( values[i] );
					if( *values[i] != '\0' &&
					    strcmp( values[ i ], 
						    defval->str ) ) {
						attr->values = g_slist_prepend( attr->values, g_strdup( values[ i ] ) );
					}
				}
				if( defval->len != 0 && *defval->str != '#' ) {
					attr->values = g_slist_prepend( attr->values,
									g_strdup( defval->str ) );
				}
				g_strfreev( values );

				attrs = g_slist_prepend( attrs, attr );

				start = TRUE;
			}
			if( *tag == '-' && *(tag + 1) =='-' ) {
				tag = strstr( tag + 2, "--" );
				if( ! tag )
					break;
				tag += 2;
				g_assert( tag );
				tag = g_strchug( (gchar*)tag );
			} else {
				start = FALSE;
				state = ANAME;
				g_string_assign( att_name, "" );
				g_string_assign( att_value, "" );
				g_string_assign( defval, "" );
				if( ( isent = ( *tag == '%' ) ) ) {
					tag ++;
				}
			}
			isuri = FALSE;
			break;
		}
	}

	g_string_free( temptag, TRUE );

	g_string_free( att_name, TRUE );
	g_string_free( att_value, TRUE );
	g_string_free( defval, TRUE );

	return attrs;
}

static gchar* screem_dtd_parse_name( ScreemDTD *dtd, const gchar *tag,
				     GString *full )
{
	gchar *temp;
	gint state;
	gboolean stop;
	GString *sub;
	gchar *temp2;

	/* skip white space at start */
	g_assert( tag );
	temp = g_strchug( (gchar*)tag );

     	sub = NULL;

	switch( *temp ) {
	case '(':
		state = EXPR;
		temp ++;
		break;
	case '%':
		state = ENTITY;
		temp ++;
		sub = g_string_new( "" );
		break;
	default:
		state = NAME;
		break;
	}

	stop = FALSE;
	sub = g_string_new( "" );
	while( ! stop && *temp != '\0' && *temp != '>' ) {
		/* bypass comments */
		if( *temp == '-' && *(temp + 1) == '-' ) {
			temp += 2;
			temp = strstr( temp, "--" );
			if( ! temp ) {
				break;
			} else {
				temp += strlen( "--" );
			}
		}
		switch( state ) {
		case NAME:
			if( *temp == '(' ) {
				sub = g_string_assign( sub, "" );
				temp = screem_dtd_parse_name( dtd, temp + 1, sub );
				g_string_append( full, sub->str );
			} else if( *temp == '%' ) {
				sub = g_string_new( "" );
				temp = screem_dtd_parse_name( dtd, temp + 1, sub );
				g_string_append( full, sub->str );
			} else if( ( ! isspace( *temp ) ) && *temp != '>' ) {
				g_string_append_c( full, *temp );
			} else {
				stop = TRUE;
			}
			break;
		case ENTITY:
			if( isalnum(*temp) || *temp == '_' || *temp == '-' ||
			    *temp == '.' ) {
				g_string_append_c( sub, *temp );
			} else {
				/* do lookup */
				temp2 = (gchar*)screem_dtd_lookup_entity( dtd,
									  sub->str );
				if( ! temp2 ) {
					g_warning( "IN PARSE NAME MISSING ENT: %s\n",
						   sub->str );
				}

				g_string_assign( sub, "" );
				/* entity expansion may contain other
				   entities */
				if( temp2 && *temp2 != '(' ) {
					tag = g_strconcat( "(", temp2, 
							   ")",NULL );
				} else if( temp2 ) {
					tag = g_strdup( temp2 );
				} else {
					tag = NULL;
				}
				if( tag ) {
					temp2 = (gchar*)tag;
					screem_dtd_parse_name( dtd, temp2, 
							       sub );
					g_free( temp2 );
				}
				if( full->len > 0 ) {
					g_string_append_c( full, '|' );
				}
				g_string_append( full, sub->str );
				g_string_assign( sub, "" );
				stop = TRUE;
			}
			break;
		case EXPR:
			switch( *temp ) {
			case ')':
				stop = TRUE;
				break;
			case '%':
				temp = screem_dtd_parse_name( dtd,temp, full );
				/* bit of a hack
				   <!ELEMENT (%font) .......
				   so the ) closes the entity off,
				   in which case we want to - 2
				*/
				if( *(temp-1) != ';' ) {
					temp --;	
				}
				temp --;
				break;
			default:
				g_string_append_c( full, *temp );
				break;
			}
			break;
		}
		temp ++;
	}

	g_string_free( sub, TRUE );
	
	return temp;
}

static gchar* screem_dtd_get_entity_value( ScreemDTD *dtd, const gchar *name,
					   gboolean isent )
{
	ScreemDTDPrivate *priv;
	const gchar *entval;

	gchar *ret;

	ret = NULL;

	priv = dtd->private;

	entval = screem_dtd_lookup_entity( dtd, name );
	if( ! entval && 
	    *name == '%' && *(name + strlen( name ) - 1) == ';' ) {
		entval = name;
	}

	if( entval ) {
		/* value may contain other entities, parse it */
		const gchar *temp;
		const gchar *ent;
		GString *str;
		
		str = g_string_new( "" );
		
		for( ent = NULL, temp = entval; temp && *temp; ++ temp ) {
			if( *temp == '%' ) {
				ent = temp + 1;
			} else if( ent && *temp == ';' ) {
				gchar *tmp;
				gchar *tmpval;

				tmp = g_strndup( ent, temp - ent );
				tmpval = screem_dtd_get_entity_value( dtd, tmp,
								      TRUE );
				if( tmpval ) {
					g_string_append( str, tmpval );
					g_free( tmpval );
				}
				g_free( tmp );
				
				ent = NULL;
			} else if( ! ent ) {
				g_string_append_c( str, *temp );
			}
		}
		ret = str->str;
		g_string_free( str, FALSE );
	}
	
	return ret;
}

static const gchar *screem_dtd_lookup_entity( ScreemDTD *dtd, const gchar *name )
{
	ScreemDTDPrivate *priv;
	const gchar *value;

	priv = dtd->private;

	value = g_hash_table_lookup( priv->entities, name );
	if( ( ! value ) && priv->parent ) {
		value = screem_dtd_lookup_entity( priv->parent, name );
	}

	return value;
}

static ScreemDTDElement *screem_dtd_lookup_element( ScreemDTD *dtd, const gchar *name,
						    gboolean casesens )
{
	ScreemDTDElement *element;

	if( casesens ) {
		element = g_hash_table_lookup( dtd->private->elements, name );
	} else {
		/* grrr */
		g_object_set_data( G_OBJECT( dtd ), "lookup_element", (gpointer)name );
		g_hash_table_foreach( dtd->private->elements,
				      (GHFunc)g_casestr_func,
				      dtd );
		element = g_object_get_data( G_OBJECT( dtd ),
					     "lookedup_element" );
		g_object_set_data( G_OBJECT( dtd ), "lookedup_element", NULL );
		g_object_set_data( G_OBJECT( dtd ), "lookup_element", NULL );
	}
	return element;
}

gboolean screem_dtd_is_root_element( ScreemDTD *dtd, const gchar *name )
{
	gboolean ret ;

	const gchar *id;
	ScreemDTDElement *element;

	if( ! dtd ) {
		return FALSE;
	}

	id = (const gchar*)g_object_get_data( G_OBJECT( dtd ), "type" );

	element = screem_dtd_lookup_element( dtd, name, TRUE );
	
	/* the problem is HTML allows tags of any case to be valid,
	   where as xml doesn't.   lets cheat a little, and see
	   if the doctype contains " HTML "  */
	if( (! element) && id ) {
		g_assert( id != NULL );
		if( strstr( id, " HTML " ) || strstr( id, " html " ) ||
		    strstr( id, " HTML/" ) || strstr( id, " html/" ) ) {
			element = screem_dtd_lookup_element( dtd, name,
							     FALSE );
		}
	}

	ret = ( dtd->private->root &&
		element && ( ! strcmp( element->name, dtd->private->root ) ) );

	return ret;
}

gboolean screem_dtd_parse_doctype( const gchar *element,
				   gchar **id, gchar **system )
{
	gint ret;
	gchar *public;

	g_return_val_if_fail( element != NULL, FALSE );
	g_return_val_if_fail( id != NULL, FALSE );
	g_return_val_if_fail( system != NULL, FALSE );

	public = g_new0( gchar, BUFSIZ );
	*id = g_new0( gchar, BUFSIZ );
	*system = g_new0( gchar, BUFSIZ );

  	/* get the dtd type and uri from the element */
	ret = sscanf( element, "%[^\"]\"%[^\"]\" %s", public, *id, *system );
	if( ret == 1 ) {
		ret = 2;
		strncpy( *id, element, BUFSIZ );
	}

	*id = g_strstrip( *id );
	*system = g_strstrip( *system );

	g_free( public );

	return ( ret == 2 || ret == 3 );
}

gboolean screem_dtd_attr_is_uri( ScreemDTD *dtd,
				 const gchar *element, const gchar *attr )
{
	ScreemDTDAttribute *att;
	gboolean ret;

	ret = FALSE;
	att= screem_dtd_valid_attr( dtd, element, attr );

	if( att ) {
		ret = att->isuri;
	}

	return ret;
}

/****************************************************************************
 * End of parsing section                                                   *
 ****************************************************************************/
ScreemDTDClose screem_dtd_element_close_state( ScreemDTD *dtd,
					       const gchar *name )
{
	ScreemDTDElement *element;
	ScreemDTDClose state;

	element = screem_dtd_valid_element( dtd, name );

	/* default is required */
	state = MUST;

	if( element ) {
		state = element->close;
	}

	return state;
}

// FIXME: only here temporarily
static gchar* next_tag_( const gchar *text, gint *pos, gchar **name )
{
        gchar *tag;
	gchar *temp;

	if( ! ( tag = strchr( text, '<' ) ) )
		return NULL;

	temp = tag + 1;
	if( ! strncmp( "!--", temp, 3 ) ) {
		if( name )
			*name = g_strdup( "!--" );
		temp = strstr( temp, "-->" );
	} else {
		if( name ) {
			while( *temp != ' ' && *temp != '>' && *temp != '\0' )
				temp ++;
			*name = g_strndup( tag + 1, temp - tag - 1 );
		}
		if( *temp != '>' )
			temp = strchr( temp, '>' );
	}

	if( ! temp ) {
		if( name )
			g_free( *name );
		return NULL;
	}

	*pos += (temp - text);

        return g_strndup( tag, temp - tag + 1 );
}

static void screem_dtd_free_data( gpointer data, GClosure *closure )
{
	g_idle_add( (GSourceFunc)g_free, data );
}


static gboolean screem_dtd_copy_entities( gpointer key, gpointer value,
					  gpointer data )
{
	ScreemDTD *dtd;

	dtd = SCREEM_DTD( data );

	if( ! g_hash_table_lookup( dtd->private->entities, key ) ) {
		g_hash_table_insert( dtd->private->entities,
				     g_strdup( key ),
				     g_strdup( value ) );
	}

	return FALSE;
}

static gboolean screem_dtd_copy_elements( gpointer key, gpointer value,
					  gpointer data )
{
	ScreemDTD *dtd;


	dtd = SCREEM_DTD( data );

	if( ! g_hash_table_lookup( dtd->private->elements, key ) ) {
		ScreemDTDElement *element;

		element = screem_dtd_element_copy( (ScreemDTDElement*)value );

		g_hash_table_insert( dtd->private->elements,
				     g_strdup( key ),
				     element );
	}
	
	return FALSE;
}

static gboolean screem_dtd_copy_attlist( gpointer key, gpointer value,
					 gpointer data )
{
	ScreemDTD *dtd;

	dtd = SCREEM_DTD( data );

	if( ! g_hash_table_lookup( dtd->private->attlists, key ) ) {
		ScreemDTDElement *element;

		element = screem_dtd_element_copy( (ScreemDTDElement*)value );


		g_hash_table_insert( dtd->private->attlists,
				     g_strdup( key ),
				     element );
	}

	return FALSE;
}

static gboolean screem_dtd_check_root( gpointer key, gpointer value,
				       gpointer data )
{
	ScreemDTD *dtd;
	const gchar *name;
	ScreemDTDElement *element;
	ScreemDTDElement *el;
	gboolean match;
	
	dtd = SCREEM_DTD( data );
	name = g_object_get_data( G_OBJECT( dtd ), "lookup_root" );
	
	element = (ScreemDTDElement*)value;
	el = screem_dtd_lookup_element( dtd, name, TRUE );

	match = TRUE;
	if( el != element ) {
		if( screem_dtd_valid_child( dtd, element->name, name ) ) {
			match = FALSE;
		}
	}

	if( ! match ) {
		g_object_set_data( G_OBJECT( dtd ), "lookedup_root", element );
	}

	return FALSE;
}


static gboolean screem_dtd_merge_attlist( gpointer key, gpointer value,
					  gpointer data )
{
	ScreemDTD *dtd;
	ScreemDTDElement *aelement;
	ScreemDTDElement *relement;

	dtd = SCREEM_DTD( data );

	aelement = (ScreemDTDElement*)value;

	relement = screem_dtd_lookup_element( dtd, aelement->name, TRUE );
	if( relement ) {
		/* copy attributes */
		GSList *list;
		ScreemDTDAttribute *attribute;

	       	for( list = aelement->attributes; list; list = list->next ) {
			attribute = screem_dtd_copy_attribute( (ScreemDTDAttribute*) list->data );
			relement->attributes = 
				g_slist_prepend( relement->attributes, attribute );
		}
       	}

	return FALSE;
}

static gboolean screem_dtd_sort_elements( gpointer key, gpointer value,
					  gpointer data )
{
	ScreemDTDPrivate *priv;
	ScreemDTDElement *cur;
       	GSList *tmp;

	priv = (ScreemDTDPrivate*)data;

	cur = (ScreemDTDElement*)value;

	cur->allowed = 
		g_slist_sort( cur->allowed,
			      (GCompareFunc)screem_dtd_compare_allowed );
	cur->attributes = 
		g_slist_sort( cur->attributes,
			      (GCompareFunc)screem_dtd_compare_attributes );

	if( ! priv->root ) {
		/* sets the root element to cur if cur is the root element
		   of the dtd - i.e.  cur is not a valid child of any
		   element */
		screem_dtd_set_root_element( priv->self, cur->name );
	} 

	if( ! priv->sorted_elements ) {
		priv->sorted_elements = g_slist_prepend( NULL, value );
	} else {
		for( tmp = priv->sorted_elements; tmp; tmp = tmp->next ) {
			cur = tmp->data;
			
			if( strcmp( key, cur->name ) < 0 ) {
				break;
			}
		}
		if( ! tmp ) {
			priv->sorted_elements =
				g_slist_append( priv->sorted_elements, value );
		} else {
			priv->sorted_elements = 
				g_slist_insert_before( priv->sorted_elements,
						       tmp,
						       value );
		}
	}
	
	return FALSE;
}

static gint screem_dtd_compare_allowed( gconstpointer a, gconstpointer b )
{
	const ScreemDTDAllowed *aa;
	const ScreemDTDAllowed *ab;

	aa = (const ScreemDTDAllowed*)a;
	ab = (const ScreemDTDAllowed*)b;

	return ( strcmp( aa->name, ab->name ) );
}

static gint screem_dtd_compare_attributes( gconstpointer a, gconstpointer b )
{
	const ScreemDTDAttribute *aa;
	const ScreemDTDAttribute *ab;

	aa = (const ScreemDTDAttribute*)a;
	ab = (const ScreemDTDAttribute*)b;

	return ( strcmp( aa->name, ab->name ) );
}

static gboolean g_casestr_func( gpointer key, gpointer value, gpointer data )
{
	ScreemDTD *dtd;
	const gchar *name;
	ScreemDTDElement *element;
	gboolean match;

	dtd = SCREEM_DTD( data );
	name = g_object_get_data( G_OBJECT( dtd ), "lookup_element" );
	
	element = (ScreemDTDElement*)value;
	match = ( ! g_strcasecmp( element->name, name ) );

	if( match ) {
		g_object_set_data( G_OBJECT( dtd ), "lookedup_element", element );
	}

	return match;
}


static ScreemDTDElement *screem_dtd_element_copy( const ScreemDTDElement *orig )
{
	ScreemDTDElement *element;
	GSList *list;
	ScreemDTDAllowed *allow;
	ScreemDTDAllowed *origal;
	ScreemDTDAttribute *attribute;

	element = g_new0( ScreemDTDElement, 1 );
	element->name = g_strdup( orig->name );
	element->close = orig->close;
	
	for( list = orig->allowed; list; list = list->next ) {
		origal = (ScreemDTDAllowed*)list->data;
		
		allow = g_new0( ScreemDTDAllowed, 1 );
		allow->name = g_strdup( origal->name );
		allow->min = origal->min;
		allow->max = origal->max;
		allow->set = origal->set;
		
		element->allowed = g_slist_prepend( element->allowed,
						    allow );
	}
	element->allowed = g_slist_reverse( element->allowed );
	
	for( list = orig->attributes; list; list = list->next ) {
		attribute = screem_dtd_copy_attribute( (ScreemDTDAttribute*) list->data );
		element->attributes = 
			g_slist_prepend( element->attributes, attribute );
					 
	}
	element->attributes = g_slist_reverse( element->attributes );

	return element;
}

static ScreemDTDAttribute *
screem_dtd_copy_attribute( const ScreemDTDAttribute *attr )
{
	ScreemDTDAttribute *attribute;
	GSList *values;

	attribute = g_new0( ScreemDTDAttribute, 1 );
	attribute->name = g_strdup( attr->name );
	
	for( values = attr->values; values;
	     values = values->next ) {
		attribute->values = 
			g_slist_prepend( attribute->values, 
					 g_strdup( values->data ) );
	}
	attribute->values =
		g_slist_reverse( attribute->values );
		
	attribute->default_value = 
		g_strdup( attr->default_value );

	return attribute;
}


const GSList *screem_dtd_get_elements( ScreemDTD *dtd )
{
	g_return_val_if_fail( SCREEM_IS_DTD( dtd ), NULL );

	return dtd->private->sorted_elements;
}

const gchar *screem_dtd_element_get_name( const ScreemDTDElement *element )
{
	return element->name;
}

const GSList *screem_dtd_element_get_attrs( const ScreemDTDElement *element )
{
	return element->attributes;
}

const gchar *
screem_dtd_attribute_get_name( const ScreemDTDAttribute *attribute )
{
	return attribute->name;
}

const GSList *
screem_dtd_attribute_get_values( const ScreemDTDAttribute *attribute )
{
	return attribute->values;
}


const gchar *screem_dtd_root_element_name( ScreemDTD *dtd )
{
	g_return_val_if_fail( SCREEM_IS_DTD( dtd ), NULL );

	return dtd->private->root;
}

static void screem_dtd_set_root_element( ScreemDTD *dtd, const gchar *name )
{
	gboolean ret;

	g_object_set_data( G_OBJECT( dtd ), "lookup_root", (gpointer)name );
	g_hash_table_foreach( dtd->private->elements,
			      (GHFunc)screem_dtd_check_root,
			      dtd );
	g_object_set_data( G_OBJECT( dtd ), "lookup_root", NULL );
	ret = ( ! g_object_get_data( G_OBJECT( dtd ), "lookedup_root" ) );
	
	if( ret ) {
		dtd->private->root = name;
	} 

	g_object_set_data( G_OBJECT( dtd ), "lookedup_root", NULL );
}
