/*  Screem:  screem-editor.c
 *
 *  The editor widget
 *
 *  Copyright (C) 1999-2001  David A Knight
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  For contact information with the author of this source code please see
 *  the AUTHORS file.  If there is no AUTHORS file present then check the
 *  about box under the help menu for a contact address
 */

#include <config.h>

#include <ctype.h>

#include <gmodule.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>

#include <gtk/gtktextview.h>
#include <gtk/gtktextbuffer.h>
#include <gtk/gtkselection.h>
#include <gtk/gtkscrolledwindow.h>
#include <gtk/gtkdnd.h>
#include <gtk/gtkmenuitem.h>
#include <gtk/gtkmain.h>
#include <gtk/gtktable.h>
#include <gtk/gtklabel.h>
#include <gtk/gtkcombo.h>
#include <gtk/gtkentry.h>
#include <gtk/gtkvbox.h>
#include <gtk/gtkimage.h>
#include <gtk/gtkstock.h>
#include <gtk/gtkeventbox.h>
#include <gtk/gtkbutton.h>
#include <gtk/gtksizegroup.h>

#include <gdk/gdkkeysyms.h>

#include <gconf/gconf-client.h>

#include <libgnomevfs/gnome-vfs.h>
#include <libgnomevfs/gnome-vfs-mime-handlers.h>
#include <libgnomevfs/gnome-vfs-mime-utils.h>

#include <libgnomeui/gnome-popup-menu.h>
#include <libgnomeui/gnome-stock-icons.h>
#include <libgnome/gnome-i18n.h>

#include "screem-application.h"
#include "screem-window.h"
#include "screem-editor.h"
#include "screem-site.h"
#include "screem-markup.h"
#include "screem-page.h"
#include "screem-plugin.h"

#include "editMenu.h"
#include "fileops.h"
#include "support.h"

#include "xml.h"

#include "php3.h"

#include "pageUI.h"

#include "libegg/menu/egg-menu.h"

#include "gdl/gdl.h"

#include "screemmarshal.h"

enum {
	REQUEST_CONTEXT,
	LAST_SIGNAL
};

static guint screem_editor_signals[ LAST_SIGNAL ] = { 0 };

typedef enum _Drop_actions {
        RELATIVE_PATH,
        FULL_PATH,
        TAG,
        ATTRIBUTE,
        INLINE     /* inserts the file contents (text files only) */
} Drop_actions;

typedef struct {
	ScreemEditor *editor;
	guint pos;
} ScreemEditorDelaySetPos;

#define GUTTER_SIZE 16

static void screem_editor_init( ScreemEditor *editor );
static void screem_editor_class_init( ScreemEditorClass *editor );
static void screem_editor_finalize( GObject *object );
static void screem_editor_window_set( ScreemView *view );
static void screem_editor_size_request( GtkWidget *widget, 
					GtkRequisition *req );
static void screem_editor_realize( GtkWidget *widget );
static void screem_editor_show( GtkWidget *widget );
static void screem_editor_hide( GtkWidget *widget );


static void screem_editor_display( ScreemView *view );
static void screem_editor_display_page( ScreemEditor *editor, 
					ScreemPage *page );

static gboolean screem_editor_mark_set( GtkTextBuffer *buffer,
					const GtkTextIter *it,
					GtkTextMark *mark,
					gpointer data );
static gboolean screem_editor_motion( GtkWidget *widget, 
				      GdkDragContext *context, 
				      gint x, gint y, guint time, 
				      ScreemEditor *editor );
static void screem_editor_drop( GtkWidget *widget, GdkDragContext *context,
				gint x, gint y, 
				GtkSelectionData *selectionData,
				guint info, guint time, ScreemEditor *editor );

static gboolean screem_editor_keypress( GtkWidget *widget, GdkEventKey *event,
					ScreemEditor *editor );
gboolean html_key_press( ScreemEditor *editor, GdkEventKey *event, 
			 ScreemEditorMacro *macro );

static gboolean screem_editor_tip( ScreemEditor *editor );
static gboolean screem_editor_tooltip( ScreemEditor *editor );

static gboolean tag_attributes_menu( ScreemEditor *editor, gchar *name );
static gboolean no_attribute( GtkWidget *widget, GdkEventKey *event, ScreemEditor *editor );


static void screem_editor_colour_notify( GConfClient *client,
					 guint cnxn_id,
					 GConfEntry *entry,
					 gpointer data );

static void screem_editor_clip_received( GtkClipboard *clipboard,
					 GtkSelectionData *selection_data,
					 gpointer data );

static void screem_editor_toggle_overwrite( GtkTextView *view,
					    ScreemEditor *editor );

static gboolean build_attribute_list( ScreemEditor *editor );
static void build_attribute_entries( ScreemEditor *editor,
				     ScreemDTD *dtd, const gchar *name,
				     GtkWidget *box,
				     GSList *attr );
static void replace_attr( GtkWidget *entry, gchar *attr_name );


static gboolean screem_editor_switch_mime( GtkWidget *widget, 
					   GdkEventButton *event,
					   ScreemEditor *editor );
static void screem_editor_mime_switch( GtkWidget *widget, 
				       ScreemEditor *editor );

static gboolean html_editor_tip( ScreemEditor *editor, ScreemWindow *window,
				 gint pos, gchar *text );

static gint screem_editor_indent( ScreemEditor *editor, gint pos );


static gboolean screem_editor_drop_uris( ScreemEditor *editor,
					 Drop_actions action,
					 const gchar *selectionData );

static void set_pos( ScreemEditorDelaySetPos *delay );

struct ScreemEditorPrivate {
	GConfClient *client;

	/* the text widget */
	GtkWidget *view;
	GtkWidget *sw;

	/* current page */
	ScreemPage *page;

	gboolean popup_open;
	gboolean inserted;

	/* screem_editor_cursor_moved() timeout handle */
	gint handle;

	/* gconf notify handles, back, text, and font */
	gint notifies[ 3 ];

	/* the signal id for "mark_set" on the current active
	   GtkTextBuffer */
	gint mark_set_signal;

	gint tooltip_timer;

	gboolean attr_up;

	GtkWidget *attr_menu;

	GtkWidget *overwriteMode;
	GtkWidget *cursorPosition;
	GtkWidget *mime;

	gboolean overwrite;


	GtkWidget *attrtab;
	GtkWidget *attributes;

	GSList *bookmarks;

	GtkWidget *types;
};

typedef enum _Drop_types {
	TARGET_URI_LIST,
	TARGET_URL,
	TARGET_MOZ,
	TARGET_COLOUR
} Drop_types;

static const GtkTargetEntry drop_types[] = {
        { "text/uri-list", 0, TARGET_URI_LIST },
	{ "text/x-moz-url", 0, TARGET_MOZ },
        { "_NETSCAPE_URL", 0, TARGET_URL },
        { "x-url/http", 0, TARGET_URL },
        { "x-url/ftp", 0, TARGET_URL },
        { "application/x-color", 0, TARGET_COLOUR }
};
static const gint num_drop_types = sizeof(drop_types) / sizeof(drop_types [0]);

/* DnD menu */
static GnomeUIInfo editor_dnd_menu[] = {
        { GNOME_APP_UI_ITEM, N_("Insert relative filename"),
          N_("Insert relative filename"),
          0, NULL, NULL,
          GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_BLANK, 0,
          GDK_CONTROL_MASK, NULL },
        { GNOME_APP_UI_ITEM, N_("Insert complete filename"),
          N_("Insert complete filename"),
          0, NULL, NULL,
          GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_BLANK, 0,
          GDK_CONTROL_MASK, NULL },
        { GNOME_APP_UI_ITEM, N_("Insert tag"), N_("Insert tag"),
          0, NULL, NULL,
          GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_BLANK, 0,
          GDK_CONTROL_MASK, NULL },
        { GNOME_APP_UI_ITEM, N_("Insert tag attibute"),
          N_("If the drop is into a tag then the data is converted into\
 an attribute for that tag"), 0, NULL, NULL,
          GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_BLANK, 0,
          GDK_CONTROL_MASK, NULL },
        { GNOME_APP_UI_ITEM, N_("Insert inline"),
          N_("Insert the text from the file into the page"), 0, NULL, NULL,
          GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_BLANK, 0,
          GDK_CONTROL_MASK, NULL },
        GNOMEUIINFO_SEPARATOR,
        { GNOME_APP_UI_ITEM, N_("Cancel drag"), N_("Cancel drag"),
          0, NULL, NULL,
          GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_BLANK, 0,
          GDK_CONTROL_MASK, NULL },
        GNOMEUIINFO_END
};

typedef gboolean (*Tipcb)( ScreemEditor *editor, gint pos, gchar *text );

static gint screem_editor_expose( GtkWidget *widget, GdkEventExpose *event,
				  ScreemEditor *editor );


static gboolean screem_editor_popup_key( GtkWidget *widget, 
						ScreemEditor *editor );
static gboolean screem_editor_popup( GtkWidget *widget, GdkEventButton *event,
					ScreemEditor *editor);

enum {
	ON = 0,
	OFF,
	BY_SET
};

ScreemEditor *screem_editor_new( ScreemWindow *window )
{
	ScreemEditor *editor;
	GType type;

	type = screem_editor_get_type();

	editor = SCREEM_EDITOR( g_object_new( type, "window", window,
					      NULL ) );

	return editor;
}

static gboolean screem_editor_popup_key( GtkWidget *widget, ScreemEditor *editor )
{
	screem_editor_popup( widget, NULL, editor );

	return TRUE;
}

gboolean screem_editor_button_release( GtkWidget *widget, 
				       GdkEventButton *event,
				       ScreemEditor *editor )
{
	ScreemEditorPrivate *private;

	private = editor->private;

	if( event && event->button == 1 &&
	    event->window ==
	    gtk_text_view_get_window( GTK_TEXT_VIEW( private->view ),
				      GTK_TEXT_WINDOW_LEFT ) ) {
		/* left click on the line numbers,
		   we want to be able to drag and select
		   entire lines, so lets cheat on pass the
		   event onto the main window, this callback is needed
		   so the drag is released properly */
		GdkEventButton child_event;

		child_event = *event;
		child_event.window =
			gtk_text_view_get_window( GTK_TEXT_VIEW(private->view),
						  GTK_TEXT_WINDOW_TEXT );
		gtk_widget_event( private->view, (GdkEvent*)&child_event );
	}

	return FALSE;
}

static gboolean screem_editor_popup( GtkWidget *widget, GdkEventButton *event,
					ScreemEditor *editor )
{
	gint cpos;
	gint pos2;
    	gint pos;
	gint tag_pos;
	gint tag_len;
	gchar *text;
	gchar *tag;
	gchar *tag2;
	GtkWidget *menu;
	GtkWidget *submenu;
	GtkWidget *menu_item;
	GtkTextBuffer *buffer;
	GtkTextTagTable *table;

	gboolean tag_known;

	ScreemPage *page;
	ScreemDTD *dtd;

	const gchar *pathname;
	const gchar *mime_type;

	gint start;
	gint len;
	gboolean sel;

	ScreemEditorPrivate *private;
	ScreemApplication *app;
	ScreemWindow *window;
	
	gboolean feature_markup;

	GSList *addeditems = NULL;
	GSList *tmp;
	
	private = editor->private;
	g_object_get( G_OBJECT( editor ), "window", &window, NULL );
	app = SCREEM_APPLICATION( window->application );

	page = private->page;

	if( ! page ) {
		return FALSE;
	}

	pathname = screem_page_get_pathname( page );

	if( event && event->button == 1 &&
	    event->window ==
	    gtk_text_view_get_window( GTK_TEXT_VIEW( private->view ),
				      GTK_TEXT_WINDOW_LEFT ) ) {
		/* left click on the line numbers,
		   we want to be able to drag and select
		   entire lines, so lets cheat on pass the
		   event onto the main window */
		GdkEventButton child_event;

		child_event = *event;
		child_event.window =
			gtk_text_view_get_window( GTK_TEXT_VIEW(private->view),
						  GTK_TEXT_WINDOW_TEXT );
		/* we do this so we select from the start of the line */
		child_event.x = 0;
		gtk_widget_event( private->view, (GdkEvent*)&child_event );
		return TRUE;
	}

	if( event && event->button == 1 && event->type == GDK_2BUTTON_PRESS ) {
		/* double click event */
		gint x;
		gint y;
		gint bx;
		gint by;
		GtkTextIter it;
		gchar *dirname;
		gchar *word;
		gchar *file;
		gchar *pattern;

		x = (gint)event->x + 0.5;
		y = (gint)event->y + 0.5;
		gtk_text_view_window_to_buffer_coords(GTK_TEXT_VIEW(private->view ),
						      GTK_TEXT_WINDOW_TEXT,
						      x, y, &bx, &by );
		gtk_text_view_get_iter_at_location(GTK_TEXT_VIEW(private->view),
						   &it, bx, by);

		pos2 = gtk_text_iter_get_offset( &it );

		text = screem_page_get_data( page );
		word = screem_editor_get_word( pos2, text );
		g_free( text );
		
		if( pathname ) {
			gchar *temp;
			temp = gnome_vfs_get_local_path_from_uri( pathname );
			if( temp ) {
				dirname = g_dirname( temp );
				g_free( temp );
			} else {
				dirname = g_strdup( "" );
			}
		} else {
			dirname = g_strdup( "" );
		}

		file = screem_support_ctags( dirname, word, &pattern );

		if( file ) {
			ScreemSite *site;

			site = screem_window_get_current( window );
			screem_page_open_with_filename( site, window,
							file );
			page = screem_site_locate_page( site, file );
			if( page && pattern ) {
				/* perform search for pattern */
				gint len;
				gchar *found;

				text = screem_page_get_data( page );
				found = find_text( text, pattern, NULL, &len );
				if( found ) {
					screem_editor_set_pos_delayed( editor, found - text );
				}

				g_free( text );
			}
			g_free( file );
			if( pattern ) {
				g_free( pattern );
			}
		} else {

		}

		g_free( dirname );
		g_free( word );

		return FALSE;
	}

	if( event && event->button != 3 )
		return FALSE;

	private->popup_open = TRUE;

	if( event ) {
		g_signal_stop_emission_by_name( G_OBJECT( widget ),
						"button_press_event" );
	} else {
		g_signal_stop_emission_by_name( G_OBJECT( widget ),
						"popup_menu" );
	}
	
	buffer = gtk_text_view_get_buffer( GTK_TEXT_VIEW( private->view ) );
	table = gtk_text_buffer_get_tag_table( buffer );

	mime_type = screem_page_get_mime_type( page );
	cpos = screem_editor_get_pos( editor );

	/* get position of click */
	if( event ) {
		gint x;
		gint y;
		gint bx;
		gint by;
		GtkTextIter it;

		x = (gint)event->x + 0.5;
		y = (gint)event->y + 0.5;
		gtk_text_view_window_to_buffer_coords(GTK_TEXT_VIEW(private->view ),
						      GTK_TEXT_WINDOW_TEXT,
						      x, y, &bx, &by );
		gtk_text_view_get_iter_at_location(GTK_TEXT_VIEW(private->view),
						   &it, bx, by);
		pos2 = pos = gtk_text_iter_get_offset( &it );
	} else {
		pos2 = pos = cpos;
	}

	sel = screem_editor_has_selection( editor, &start, &len );
	screem_editor_set_pos( editor, pos );

	if( sel ) {
		screem_editor_select_region( editor, start, len - start );
	}

	/* create the menu */
       	menu = egg_menu_merge_get_widget( EGG_MENU_MERGE( window->merge ),
					"/popups/editormenu" );
	
	/* id the tag we are inside so we can offer attribute editing */
	
	text = screem_page_get_data( page );

	feature_markup = ( g_object_get_data( G_OBJECT( table ), 
					      "feature_markup" ) != NULL );

	if( feature_markup &&
	    screem_markup_is_tag( text, pos, &pos, NULL ) ) {
		tag_pos = pos;
		tag = screem_markup_next_tag( text, pos, NULL, &pos, &tag2 );
		tag_len = pos - tag_pos + 1;

		dtd = screem_window_get_current_dtd( window );
		menu_item = screem_dtd_attribute_menu( dtd, tag2, editor );
		if( menu_item ) {
			gtk_menu_shell_prepend( GTK_MENU_SHELL( menu ), 
						menu_item );
			addeditems = g_slist_prepend( addeditems, menu_item );
		}
		tag_known = (gboolean)menu_item;

		/* add the wizard option */
		if( (menu_item = screem_plugin_tag_wizard( app, 
							   window,
							   tag2, tag,
							   tag_pos, 
							   tag_len ) ) ) {
			gtk_menu_shell_append( GTK_MENU_SHELL( menu ),
					       menu_item );
			addeditems = g_slist_prepend( addeditems, menu_item );
		}

		g_free( tag );

		/* add the help option */
		if( 0 && tag_known && ! strcmp( "text/html", mime_type ) ) {
			gchar *title = g_strdup_printf( _( "<%s> Help" ),
							tag2 );
			menu_item = gtk_menu_item_new_with_label( title );    
			g_free( title );
			gtk_widget_show( menu_item );
			gtk_menu_append( GTK_MENU( menu ), menu_item );
			g_signal_connect_data( G_OBJECT( menu_item ),
					       "activate",
					       G_CALLBACK( NULL ),
					       (gpointer)g_strdup( tag2 ),
					       (GClosureNotify)g_free, 0 );
			addeditems = g_slist_prepend( addeditems, menu_item );
		}
		g_free( tag2 );
	} else if( feature_markup ) {
		/* add the insert option */
		g_signal_emit( G_OBJECT( editor ),
			       screem_editor_signals[ REQUEST_CONTEXT ],
			       0, pos, &tag2, NULL, NULL, NULL, TRUE );

		dtd = screem_window_get_current_dtd( window );

		if( tag2 && 
		    ( menu_item = screem_dtd_allowed_menu( dtd, tag2 ) ) ) {
			gtk_menu_shell_append( GTK_MENU_SHELL( menu ), 
					       menu_item );
			addeditems = g_slist_prepend( addeditems, menu_item );
		}
	} else if( ! strcmp( "text/x-perl", mime_type ) ) {
		/* create perl menu */
		/*menu_item = perl_menu();
		  gtk_menu_shell_prepend( GTK_MENU_SHELL( menu ), menu_item );
		
		addeditems = g_slist_prepend( addeditems, menu_item );
		*/
		
	} else if( ! strcmp( "text/x-java", mime_type ) ) {
		/* create java menu */
		/*menu_item = java_menu();
		  gtk_menu_shell_prepend( GTK_MENU_SHELL( menu ), menu_item );
		
		addeditems = g_slist_prepend( addeditems, menu_item );
		*/
	}
	g_free( text );

	menu_item = gtk_menu_item_new_with_mnemonic( _( "Input _Methods" ) );
	addeditems = g_slist_prepend( addeditems, menu_item );
	gtk_widget_show( menu_item );
	submenu = gtk_menu_new ();
	gtk_menu_item_set_submenu( GTK_MENU_ITEM( menu_item ), submenu );
	gtk_menu_shell_append( GTK_MENU_SHELL( menu ), menu_item );
	gtk_im_multicontext_append_menuitems( GTK_IM_MULTICONTEXT( GTK_TEXT_VIEW(private->view)->im_context ),
						GTK_MENU_SHELL( submenu ) );

	/* display menu */
	private->inserted = FALSE;
	gnome_popup_menu_do_popup_modal( menu, 0, 0, event, 0, private->view );

	for( tmp = addeditems; tmp; tmp = tmp->next ) {
		gtk_container_remove( GTK_CONTAINER( menu ), 
					GTK_WIDGET( tmp->data ) );
	}
	g_slist_free( addeditems );
	
	if( ! private->inserted )
		screem_editor_set_pos( editor, cpos );
	
	if( sel ) {
		screem_editor_select_region( editor, start, len - start );
	}

	private->popup_open = FALSE;
		
	return TRUE;
}

void screem_editor_insert( ScreemEditor *editor, gint pos, const gchar *text )
{
	gint len;
	gint start;
	const gchar *pathname;
	const gchar *mime_type;

	GtkTextBuffer *buffer;
	GtkTextIter it;

	buffer = 
		gtk_text_view_get_buffer(GTK_TEXT_VIEW(editor->private->view));

	if( ! text || ! editor->private->page )
		return;

	pathname = screem_page_get_pathname( editor->private->page );
	mime_type = screem_page_get_mime_type( editor->private->page );

	len = screem_editor_get_length( editor );

	if( pos == -1 )
		pos = screem_editor_get_pos( editor );

	if( pos > len )
		pos = len;

	start = pos;

	len = strlen( text );

	gtk_text_buffer_get_iter_at_offset( buffer, &it, pos );
	gtk_text_buffer_insert( buffer, &it, text, len );

	screem_editor_focus( editor );
}

gboolean screem_editor_focus( ScreemEditor *editor )
{
	gtk_widget_grab_focus( editor->private->view );
	
	return FALSE;
}

guint screem_editor_get_length( ScreemEditor *editor )
{
	GtkTextBuffer *buffer;

	buffer = 
		gtk_text_view_get_buffer(GTK_TEXT_VIEW(editor->private->view));

	return gtk_text_buffer_get_char_count( buffer );
}

gchar *screem_editor_get_text( ScreemEditor *editor, guint start, guint len )
{
	GtkTextBuffer *buffer;
	GtkTextIter it;
	GtkTextIter eit;

	guint end = len;

	buffer =
		gtk_text_view_get_buffer(GTK_TEXT_VIEW(editor->private->view));

	if( end == 0 )
		end = screem_editor_get_length( editor );
	else
		end += start;

	gtk_text_buffer_get_iter_at_offset( buffer, &it, start );
	gtk_text_buffer_get_iter_at_offset( buffer, &eit, end );

	return gtk_text_buffer_get_text( buffer, &it, &eit, TRUE );
}

gboolean screem_editor_has_selection( ScreemEditor *editor, guint *start, 
				      guint *end )
{
	guint temp;
	gboolean ret;
	GtkTextBuffer *buffer;
	GtkTextIter it;
	GtkTextIter eit;

	buffer = 
		gtk_text_view_get_buffer(GTK_TEXT_VIEW(editor->private->view));

	if( ! start )
		start = &temp;
	if( ! end )
		end = &temp;

	ret = gtk_text_buffer_get_selection_bounds( buffer, &it, &eit );

	*start = gtk_text_iter_get_offset( &it );
	*end = gtk_text_iter_get_offset( &eit );

	return ret;
}

void screem_editor_select_region( ScreemEditor *editor, guint start, guint len )
{
        GtkTextBuffer* buffer = NULL;
        GtkTextIter it;
	GtkTextIter eit;

        buffer = 
		gtk_text_view_get_buffer(GTK_TEXT_VIEW(editor->private->view));

	if( len == 0 )
		len = screem_editor_get_length( editor );

	gtk_text_buffer_get_iter_at_offset( buffer, &it, start );
	gtk_text_buffer_get_iter_at_offset( buffer, &eit, start + len );

        gtk_text_buffer_place_cursor( buffer, &eit );

        gtk_text_buffer_move_mark( buffer,
				   gtk_text_buffer_get_mark(buffer, 
							    "selection_bound"),
				   &it );
	gtk_text_view_scroll_mark_onscreen(GTK_TEXT_VIEW(editor->private->view),
					   gtk_text_buffer_get_mark(buffer,
								    "insert"));
}

void screem_editor_cut( ScreemEditor *editor )
{
	GtkTextBuffer *buffer;

	buffer = 
		gtk_text_view_get_buffer(GTK_TEXT_VIEW(editor->private->view));

	gtk_text_buffer_cut_clipboard( buffer,
				       gtk_clipboard_get( GDK_NONE ),
				       TRUE );

	gtk_text_view_scroll_mark_onscreen(GTK_TEXT_VIEW(editor->private->view),
					   gtk_text_buffer_get_mark(buffer,
								    "insert"));
}

void screem_editor_copy( ScreemEditor *editor )
{
	GtkTextBuffer *buffer;

	buffer = 
		gtk_text_view_get_buffer(GTK_TEXT_VIEW(editor->private->view));

	gtk_text_buffer_copy_clipboard( buffer,
					gtk_clipboard_get( GDK_NONE ) );
	
	gtk_text_view_scroll_mark_onscreen(GTK_TEXT_VIEW(editor->private->view),
					   gtk_text_buffer_get_mark(buffer,
								    "insert"));
}

void screem_editor_paste( ScreemEditor *editor )
{
	GtkTextBuffer *buffer;

	buffer = 
		gtk_text_view_get_buffer(GTK_TEXT_VIEW(editor->private->view));

	gtk_text_buffer_paste_clipboard( buffer,
					 gtk_clipboard_get( GDK_NONE ),
					 NULL,
					 TRUE );

	gtk_text_view_scroll_mark_onscreen(GTK_TEXT_VIEW(editor->private->view),
					   gtk_text_buffer_get_mark(buffer,
								    "insert"));

}

void screem_editor_paste_encoded( ScreemEditor *editor )
{
	gtk_clipboard_request_contents( gtk_clipboard_get( GDK_NONE ),
					gdk_atom_intern( "UTF8_STRING", 
							 FALSE ),
					screem_editor_clip_received, editor );
}

void screem_editor_clear_selection( ScreemEditor *editor )
{
	GtkTextBuffer *buffer;

	buffer = 
		gtk_text_view_get_buffer(GTK_TEXT_VIEW(editor->private->view));

	gtk_text_buffer_delete_selection( buffer, FALSE, TRUE );

	gtk_text_view_scroll_mark_onscreen(GTK_TEXT_VIEW(editor->private->view),
					   gtk_text_buffer_get_mark(buffer,
								    "insert"));
}

void screem_editor_set_pos_delayed( ScreemEditor *editor, guint pos )
{
	ScreemEditorDelaySetPos *delay;

	delay = g_new0(ScreemEditorDelaySetPos, 1 );
	delay->editor = editor;
	delay->pos = pos;
	g_idle_add( (GSourceFunc)set_pos, delay );
	g_idle_add( (GSourceFunc)screem_editor_focus, editor );
}

void screem_editor_set_pos( ScreemEditor *editor, guint pos )
{
	GtkTextView *view;
	GtkTextBuffer *buffer;
	GtkTextIter it;

	view = GTK_TEXT_VIEW( editor->private->view );

	buffer = gtk_text_view_get_buffer( view );

	gtk_text_buffer_get_iter_at_offset( buffer, &it, pos );

	gtk_text_buffer_place_cursor( buffer, &it );

	gtk_text_view_scroll_to_iter( view, &it, 0.0, FALSE, 0.0, 0.0 );
}

guint screem_editor_get_pos( ScreemEditor *editor )
{
	GtkTextBuffer *buffer;
	GtkTextMark *mark;
	GtkTextIter it;

	buffer = 
		gtk_text_view_get_buffer(GTK_TEXT_VIEW(editor->private->view));
	mark = gtk_text_buffer_get_mark( buffer, "insert" );

	gtk_text_buffer_get_iter_at_mark( buffer, &it, mark );

	return gtk_text_iter_get_offset( &it );
}

void screem_editor_delete_forward( ScreemEditor *editor, guint pos, guint len )
{
	GtkTextBuffer *buffer;
	GtkTextIter it;
	GtkTextIter eit;

	if( len == 0 )
		len = ( screem_editor_get_length( editor ) - pos );

	buffer = 
		gtk_text_view_get_buffer(GTK_TEXT_VIEW(editor->private->view));
 
	gtk_text_buffer_get_iter_at_offset( buffer, &it, pos );
	gtk_text_buffer_get_iter_at_offset( buffer, &eit, pos + len );

	gtk_text_buffer_delete( buffer, &it, &eit );
}

void screem_editor_insert_markup( ScreemEditor *editor, 
				  const gchar *open_element, 
				  const gchar *close_element )
{
	gboolean has_selection;
	guint start;
	guint end;
	gint pos;
	gint ipos;
	ScreemPage *page;

	page = editor->private->page;
	
	if( ! page )
		return;

	pos = screem_editor_get_pos( editor );
	ipos = pos;

	has_selection = screem_editor_has_selection( editor, &start, &end );

	if( has_selection ) {
		if( open_element ) {
			screem_editor_insert( editor, start, open_element );
                        end += strlen( open_element );
		}
		pos = end;
		if( close_element ) {
                        screem_editor_insert( editor, end, close_element );
                        pos += strlen( close_element );
                }
	} else {
		if( open_element ) {
			screem_editor_insert( editor, pos, open_element );
			pos += strlen( open_element );
		}
		if( close_element ) {
			screem_editor_insert( editor, pos, close_element );
			pos += strlen( close_element );
		}

		if( open_element )
			pos = ipos + strlen( open_element );
		else if( close_element )
			pos = ipos + strlen( close_element );
	}
	screem_editor_set_pos( editor, pos );

	if( editor->private->popup_open )
		editor->private->inserted = TRUE;
}

/* This is horrible but its the cleanest way I could think of */
void screem_editor_insert_attr_cb( GtkWidget *item, const gchar *attribute )
{
	ScreemEditor *editor;

	editor = SCREEM_EDITOR( g_object_get_data( G_OBJECT( item ),
						   "editor" ) );
	
	screem_editor_insert_attribute( editor, attribute );
}

void screem_editor_insert_attribute( ScreemEditor *editor, 
				     const gchar *attribute )
{
	gint pos;
	gint cpos;
        gint start;
        gchar *insert;
        gchar *text;
        gchar *tag;
        gchar *attr;
	ScreemPage *page;

	gchar *temp;

	page = editor->private->page;
	
	if( ! page || ! attribute )
		return;

        cpos = pos = screem_editor_get_pos( editor );
        text = screem_page_get_data( page );

        if( ! screem_markup_is_tag( text, pos, &pos, NULL ) ) {
		g_free( text );
		return;
        }
        start = pos;
        /* pos == the start of the tag */
        tag = screem_markup_next_tag( text, pos, NULL, &pos, NULL );

  	/* does the tag already have this attribute?
           = match:  [A-Za-z]*=\"[^"]*\"
           other: [A-Za-z]*
        */
      	temp = g_strconcat( " ", attribute, NULL );
        if( ( attr = find_text( temp, "=", NULL, NULL ) ) ) {
                attr = g_strndup( temp, attr - temp );
                insert = g_strdup_printf( "%s=\\\"[^\\\"]*\\\"", attr );
                g_free( attr );
        } else {
                insert = g_strdup( temp );
        }

        /* now we try and replace it */
        if( ! ( attr = find_text( tag, insert, temp, NULL ) ) ) {
                /* we didn't find it so we just insert it */
                g_free( insert );
		/* pos is at the end of the tag, however, we may
		   be dealing with xml, hence we need to insert
		   the attribute before a possible closing / */
		if( text[ pos - 1 ] == '/' ) {
			pos --;
		}
		insert = g_strdup_printf( " %s", attribute );
                screem_editor_insert( editor, pos, insert );
		pos += strlen( attribute ) + 1;
	} else {
                /* erase all the tag's attributes */
		g_free( tag );
                tag = attr;
		screem_editor_delete_forward( editor, start, pos - start + 1 );
                insert = g_strdup( tag );
                screem_editor_insert( editor, start, insert );
		pos = start + strlen( insert );
        }

        g_free( text );
	g_free( tag );
	g_free( insert );
	g_free( temp );

	editor->private->inserted = TRUE;

	screem_editor_set_pos( editor, pos );
}

void screem_editor_insert_file( ScreemEditor *editor, const gchar *filename )
{
	ScreemPage *page;
	gchar *data;
	gint pos;

	page = screem_page_new();

	screem_page_set_pathname( page, filename );
	screem_page_load( page );

	pos = screem_editor_get_pos( editor );
	data = screem_page_get_data( page );
	screem_editor_insert( editor, pos, data );
	
	g_free( data );
	
	g_object_unref( page );
}

void screem_editor_undo( ScreemEditor *editor )
{
	ScreemPage *page;

	page = editor->private->page;

	if( ! page )
		return;

	screem_page_undo( page );
}

void screem_editor_redo( ScreemEditor *editor )
{
	ScreemPage *page;

	page = editor->private->page;

	if( ! page )
		return;

	screem_page_redo( page );
}

static gint screem_editor_indent( ScreemEditor *editor, gint pos )
{
	guint depth;
	guint start;
	guint end;
	GString *indent;
	GtkTextIter it;
	GtkTextIter *copy;
	GtkTextBuffer *buffer;
	gunichar c;
	guint offset;
	gboolean corrected;

	guint sel_start;
	guint sel_end;
	gint sel_offset;
	gboolean built;

	if( ! screem_editor_has_selection( editor, &sel_start, &sel_end ) ) {
		sel_start = pos;
		sel_end = sel_start;
	}

	buffer =gtk_text_view_get_buffer(GTK_TEXT_VIEW(editor->private->view));

	/* convert to line numbers */
	gtk_text_buffer_get_iter_at_offset( buffer, &it, sel_start );
	sel_start = gtk_text_iter_get_line( &it );
	gtk_text_buffer_get_iter_at_offset( buffer, &it, sel_end );
	sel_end = gtk_text_iter_get_line( &it );

	for( sel_offset = 0, built = FALSE; sel_start <= sel_end;
		++ sel_start ) {
		pos = sel_start;
		gtk_text_buffer_get_iter_at_line( buffer, &it, pos );
		pos = gtk_text_iter_get_offset( &it );
	
		g_signal_emit( G_OBJECT( editor ),
			       screem_editor_signals[ REQUEST_CONTEXT ],
			       0, pos - sel_offset, NULL, &depth, 
			 	&start, &end, ! built );
		built = TRUE;
	
		/* if the first char on a line is < then we will have
		   depth 1 > than we should */
		corrected = FALSE;
		if( depth != 0 && gtk_text_iter_get_char( &it ) == '<' ) {
			depth --;
			corrected = TRUE;
		}
		
		copy = gtk_text_iter_copy( &it );
		offset = 0;
		
		do {
			offset ++;
			c = gtk_text_iter_get_char( copy );
			if( ! g_unichar_isspace( gtk_text_iter_get_char( copy ) ) ||
			    c == '\n' ) {
				break;
			}
			gtk_text_iter_forward_char( copy );
		} while( ! gtk_text_iter_is_end( copy ) );
		
		gtk_text_buffer_delete( buffer, &it, copy );
		gtk_text_iter_free( copy );
		gtk_text_buffer_get_iter_at_offset( buffer, &it, pos );
		
		/* handle close of context correctly */
		copy = gtk_text_iter_copy( &it );
		c = gtk_text_iter_get_char( copy );
		if( c == '<' ) {
			gtk_text_iter_forward_char( copy );
			if( ! gtk_text_iter_is_end( copy ) ) {
				c = gtk_text_iter_get_char( copy );
			}
			if( c == '/' ) {
				/* closing tag, does it close the context? */
				do {
					gtk_text_iter_forward_char( copy );
					if( gtk_text_iter_is_end( copy ) ) {
						break;
					}
					c = gtk_text_iter_get_char( copy );
				} while( c != '>' );
			}
			pos = gtk_text_iter_get_offset( copy ) + 1;
			pos += offset;
			/* is pos outside the current context? */
			if( pos >= end && depth != 0 && ! corrected ) {
				depth --;
			}
		}
		gtk_text_iter_free( copy );
		
		indent = g_string_new_len( "", depth );
		memset( indent->str, '\t', depth );

		gtk_text_buffer_insert( buffer, &it, indent->str, indent->len );
		
		sel_offset -= offset;
		sel_offset += 1 + indent->len;
		
		g_string_free( indent, TRUE );
	}
	
	pos = gtk_text_iter_get_offset( &it );
	pos += sel_offset;

	return pos;
}

void screem_editor_goto_line( ScreemEditor *editor, gint line )
{
	ScreemPage *page;
	GtkTextIter it;
	GtkTextView *view;
	GtkTextBuffer *buffer;

	page = editor->private->page;

	if( ! page )
		return;

	view = GTK_TEXT_VIEW(editor->private->view);
	buffer = screem_page_get_text_buffer( page );

	gtk_text_buffer_get_iter_at_line( buffer, &it, line - 1 );
	gtk_text_view_scroll_to_iter( view, &it, 0.0, FALSE, 0.0, 0.0 );
}

void screem_editor_encode_text( ScreemEditor *editor )
{
	gchar *text;
	gchar *ret;
	guint start = 0;
	guint len = 0;
	ScreemPage *page;

	page = editor->private->page;

	if( ! page )
		return;

	if( screem_editor_has_selection( editor, &start, &len ) )
		len = len - start;

	text = screem_editor_get_text( editor, start, len );

	ret = screem_markup_encode_text( text );

	screem_editor_delete_forward( editor, start, len );
	screem_editor_insert( editor, start, ret );

	g_free( ret );
	g_free( text );
}

gchar* screem_editor_get_word( gint pos, gchar *text )
{
	gint start;
	gint end;
	gchar *word;

	for( start = pos; start; start -- ) {
		if( ( ! isalnum( text[ start ] ) ) && text[ start ] != '_' )
			break;
	}
	for( end = pos; text[ end ]; end ++ ) {
		if( ( ! isalnum( text[ end ] ) ) && text[ end ] != '_' )
			break;
	}

	if( start != pos )
		start ++;

	if( end != pos )
		end --;

    	if( start == end )
		word = NULL;
	else
		word = g_strndup( &text[ start ], end - start + 1 );

	return word;
}

void screem_editor_set_colours( ScreemEditor *editor )
{
	GdkColor back;
	GdkColor text;
	GConfClient *client;
	ScreemEditorPrivate *private;
 	PangoFontDescription *font = NULL;
	gchar *value;

	private = editor->private;
	client = private->client;

	value = gconf_client_get_string( client, "/apps/screem/editor/back",
					 NULL );
	if( ! value ) {
		value = g_strdup( "#ffffffffffff" );
	}
	gdk_color_parse( value, &back );
	g_free( value );
	
	gtk_widget_modify_base( GTK_WIDGET( private->view ),
				GTK_STATE_NORMAL, &back );

	value = gconf_client_get_string( client, "/apps/screem/editor/text",
					 NULL );
	if( ! value ) {
		value = g_strdup( "#000000000000" );
	}
	gdk_color_parse( value, &text );

	g_free( value );

	gtk_widget_modify_text( GTK_WIDGET( private->view ),
				GTK_STATE_NORMAL, &text );

	value = gconf_client_get_string( client, "/apps/screem/editor/font",
					 NULL );
	if( value ) {
		font = pango_font_description_from_string( value );
		gtk_widget_modify_font( GTK_WIDGET( private->view ), font );
		pango_font_description_free( font );
		g_free( value );
	}

}

static void screem_editor_print( ScreemEditor *editor, gboolean preview )
{
	screem_page_print( editor->private->page, preview );
}

void screem_editor_expose_lines( ScreemEditor *editor )
{
	GdkWindow *window;
	GdkRectangle rect;

	/* we only want to force an expose on the currently visible
	   lines numbers incase a bookmark has been added/removed
	   in view */

	window = gtk_text_view_get_window(GTK_TEXT_VIEW(editor->private->view),
					  GTK_TEXT_WINDOW_LEFT);

   	gtk_text_view_get_visible_rect(GTK_TEXT_VIEW(editor->private->view),
				       &rect);

	gdk_window_invalidate_rect( window, &rect, TRUE );
	gdk_window_process_updates( window, TRUE );
}

/* static stuff */
static void screem_editor_display( ScreemView *view )
{
	ScreemPage *page;

	g_object_get( G_OBJECT( view ), "page", &page, NULL );

	screem_editor_display_page( SCREEM_EDITOR( view ), page );
}

static void screem_editor_display_page( ScreemEditor *editor, 
					ScreemPage *page )
{
	GtkTextBuffer *buffer;

	if( page ) {
		GSList *types;
		GSList *tmp;
		GtkWidget *item;
		const gchar *pathname;
		const gchar *type;
		const gchar *showtype;

		pathname = screem_page_get_pathname( page );
		type = NULL;
		if( pathname ) {
			type = gnome_vfs_get_mime_type( pathname );
		} 
		if( ! type ) {
			type = "text/plain";
		}
		showtype = gnome_vfs_mime_get_description( type );

		if( ! showtype ) {
			showtype = type;
		}

		if( editor->private->types ) {
			gtk_widget_destroy( editor->private->types );
		}

		types = screem_page_get_mime_types( page );

		editor->private->types = gtk_menu_new();
		
		item = gtk_menu_item_new_with_label( showtype );
		gtk_widget_show( item );
		gtk_menu_shell_append( GTK_MENU_SHELL(editor->private->types ),
				       item );
		g_object_set_data( G_OBJECT( item ), "label", (gchar*)type );
		g_signal_connect( G_OBJECT( item ), "activate",
				  G_CALLBACK( screem_editor_mime_switch ),
				  editor );
		for( tmp = types; tmp; tmp = tmp->next ) {
			showtype = gnome_vfs_mime_get_description( tmp->data );
			if( ! showtype ) {
				showtype = tmp->data;
			}
			item = gtk_menu_item_new_with_label( showtype );
			gtk_widget_show( item );
			g_object_set_data( G_OBJECT( item ), "label",
					   tmp->data );
			g_signal_connect( G_OBJECT( item ), "activate",
					  G_CALLBACK( screem_editor_mime_switch ),
					  editor );
			gtk_menu_shell_append( GTK_MENU_SHELL( editor->private->types ), item );
		}
		
		g_slist_free( types );
	} else {
		editor->private->mark_set_signal = 0;
	}

	/* disconnect mark_set signal from the buffer */
	buffer = gtk_text_view_get_buffer( GTK_TEXT_VIEW( editor->private->view ) );

	if( editor->private->mark_set_signal ) {
		g_signal_handler_disconnect( G_OBJECT( buffer ),
					     editor->private->mark_set_signal);
	}

	gtk_text_view_set_buffer( GTK_TEXT_VIEW( editor->private->view ),
				  NULL );
	gtk_text_view_set_editable( GTK_TEXT_VIEW( editor->private->view ),
				    FALSE );
	
	
	if( editor->private->page ) {
		g_slist_foreach( editor->private->bookmarks,
				 (GFunc)gtk_widget_destroy, NULL );

		g_slist_free( editor->private->bookmarks );
		editor->private->bookmarks = NULL;
	}

	editor->private->page = NULL;

	gtk_button_set_label( GTK_BUTTON( editor->private->mime ), 
					  _( "No Page" ) );

	if( page ) {
		GtkTextTagTable *table;
		const gchar *type;
		const gchar *showtype;

		buffer = screem_page_get_text_buffer( page );

		table = gtk_text_buffer_get_tag_table( buffer );

		editor->private->page = page;

		gtk_text_view_set_buffer( GTK_TEXT_VIEW( editor->private->view ),
					  buffer );
		
		gtk_text_view_set_editable( GTK_TEXT_VIEW( editor->private->view ),
					    TRUE );
		type = screem_page_get_mime_type( page );
		showtype = gnome_vfs_mime_get_description( type );
		if( ! showtype ) {
			showtype = type;
		}
		gtk_button_set_label( GTK_BUTTON( editor->private->mime ),
				      showtype );
       	} else {
		gtk_button_set_label( GTK_BUTTON( editor->private->mime ),
				      "text/plain" );


		/* yes we did set this to NULL, but that means we
		   get a newed one back so we can actually clear the display
		   it will be unrefed when we set it to a proper one later */
		buffer = gtk_text_view_get_buffer( GTK_TEXT_VIEW( editor->private->view ) );

		gtk_text_view_set_buffer( GTK_TEXT_VIEW( editor->private->view ), buffer );
	}

	/* connect mark_set signal */
	editor->private->mark_set_signal = 
		g_signal_connect( G_OBJECT( buffer ),
				  "mark_set",
				  G_CALLBACK( screem_editor_mark_set ),
				  editor );
}

static gboolean screem_editor_mark_set( GtkTextBuffer *buffer,
					const GtkTextIter *it,
					GtkTextMark *mark,
					gpointer data )
{
	ScreemEditor *editor;
	ScreemPage *page;

	g_return_val_if_fail( SCREEM_IS_EDITOR( data ), FALSE );

	editor = SCREEM_EDITOR( data );


	page = editor->private->page;

	if( page ) {
		gint row;
		gint col;
		gchar *mesg;

		if( editor->private->handle != -1 ) {
			g_source_remove( editor->private->handle );
		}
		
		editor->private->handle = 
			g_timeout_add( 1000, (GSourceFunc)build_attribute_list,
				       editor );
		
		gtk_statusbar_pop( GTK_STATUSBAR( editor->private->cursorPosition ),
				   0 );
		
		row = gtk_text_iter_get_line( it );
		col = gtk_text_iter_get_line_offset( it );
		mesg = g_strdup_printf( _("  Ln %d, Col. %d"), row + 1, col + 1 );

		gtk_statusbar_push( GTK_STATUSBAR( editor->private->cursorPosition ),
				    0, mesg );
		g_free( mesg );
	}

	return TRUE;
}

static gboolean screem_editor_motion( GtkWidget *widget, 
				      GdkDragContext *context, 
				      gint x, gint y, guint time, 
				      ScreemEditor *editor )
{
	GdkDragAction action;
       
	GdkModifierType modifiers;
        gdk_window_get_pointer (NULL, NULL, NULL, &modifiers);

	
        if( context->suggested_action != GDK_ACTION_ASK )
                action = GDK_ACTION_COPY;
        else
                action = GDK_ACTION_ASK;

	if ((modifiers & GDK_MOD1_MASK) != 0)
                action = GDK_ACTION_ASK;
        
        gdk_drag_status( context, action, time );

        return TRUE;
}

static void screem_editor_drop( GtkWidget *widget, GdkDragContext *context,
				gint x, gint y, 
				GtkSelectionData *selectionData,
				guint info, guint time, ScreemEditor *editor )
{
	ScreemPage *page;
	GtkWidget *popup;
	gint item;
	gint pos;
	guint16 *colours;
	gchar *text = NULL;
	gchar *temp;
	Drop_actions action;

	const gchar *pathname;

	ScreemWindow *window;
	ScreemSite *site;

	const gchar *spath;

	gint bx;
	gint by;
	GtkTextIter it;

	gboolean handled;

	handled = FALSE;

	page = editor->private->page;
	g_object_get( G_OBJECT( editor ), "window", &window, NULL );
	site = screem_window_get_current( window );
	spath = screem_site_get_pathname( site );
	
	if( ! page ) {
		gtk_drag_finish( context, FALSE, FALSE, time );
		return;
	}

	pathname = screem_page_get_pathname( page );

	/* if we are doing a right drag then we need to ask the user
	   what to do */
	if( context->action == GDK_ACTION_ASK ) {
                popup = gnome_popup_menu_new( editor_dnd_menu );
                item = gnome_popup_menu_do_popup_modal( popup, 0, 0, 0, 0, 0 );
                switch( item ) {
                case 0: /* insert relative filename */
                        action = RELATIVE_PATH;
                        break;
                case 1: /* insert complete filename */
                        action = FULL_PATH;
                        break;
                case 2: /* insert tag */
                        action = TAG;
                        break;
                case 3: /* insert tag attribute */
                        action = ATTRIBUTE;
                        break;
                case 4: /* insert text inline */
                        action = INLINE;
                        break;
                default:
			gtk_drag_finish( context, FALSE, FALSE, time ); 
			return;
                        break;
                }
		gtk_widget_destroy( popup );
        } else {
                action = RELATIVE_PATH;
	}

	/* position the cursor at the drop location */
	gtk_text_view_window_to_buffer_coords(GTK_TEXT_VIEW(editor->private->view ),
					      GTK_TEXT_WINDOW_WIDGET,
					      x, y, &bx, &by );
	gtk_text_view_get_iter_at_location(GTK_TEXT_VIEW(editor->private->view),
					   &it, bx, by);
	pos = gtk_text_iter_get_offset( &it );

	screem_editor_set_pos( editor, pos );


	if( info == TARGET_MOZ ) {
		glong read;
		glong written;

		text = g_utf16_to_utf8( selectionData->data,
					selectionData->length,
					&read, &written,
					NULL );
	} else {
		text = g_strdup( selectionData->data );
	}
	
	/* handle the different drop types */
	switch( info ) {
	case TARGET_MOZ:
	case TARGET_URL:
	case TARGET_URI_LIST:
		handled = screem_editor_drop_uris( editor, 
						   action,
						   text );

		break;
	case TARGET_COLOUR:
		colours = (guint16*)selectionData->data;
                temp = g_strdup_printf( "\"#%.2x%.2x%.2x\"",colours[ 0 ] >> 8, 
                                        colours[ 1 ] >> 8, colours[ 2 ] >> 8 );
		switch( action ) {
		case TAG:
                        text = g_strdup_printf( "<font color=%s> </font>",
                                                temp );
                        g_free( temp );
                        break;
		case ATTRIBUTE:
			text = screem_page_get_data( page );
                        if( screem_markup_is_tag( text, pos, NULL, NULL ) ) {
                                /* we are inside a tag */
                                g_free( text );
                                text = g_strdup_printf( " color=%s", temp );
                                screem_editor_insert_attribute( editor, text );
                                g_free( text );
				gtk_drag_finish( context, TRUE, FALSE, time );
                                return;
                        } else {
                                /* not in a tag, but we will insert it anyway*/
                                g_free( text );
                                text = g_strdup_printf( " color=%s", temp );
                                g_free( temp );
                        }
                        break;
		default:
			text = temp;
			break;
		}
		if( text ) {
			screem_editor_insert( editor, pos, text );
			g_free( text );
			handled = TRUE;
		}
		break;
	default:
		/* we don't know about this type, and as such
		   this case should never execute */
		screem_window_show_error( SCREEM_WINDOW( window ),
					  _("Error: invalid drop type for editor occured\n") );
		g_free( text );
		break;
	}

	gtk_drag_finish( context, handled, FALSE, time );
}

static gboolean screem_editor_keypress( GtkWidget *widget, GdkEventKey *event, 
					ScreemEditor *editor )
{
	ScreemPage *page;

	const gchar *pathname;

	GList *list;
	GModule *self;

	gchar *string;
	ScreemEditorMacro *macro;
	gint pos;

	gboolean press_handled;

	ScreemEditorPrivate *private;
	ScreemApplication *application;
	GConfClient *client;

	GtkTextBuffer *buffer;
	GtkTextTagTable *table;
	gboolean feature_markup;

	ScreemWindow *window;

	private = editor->private;
	client = private->client;

	page = private->page;

	g_object_get( G_OBJECT( editor ), "window", &window, NULL );

	application = SCREEM_APPLICATION( SCREEM_WINDOW( window )->application );

	if( ! page )
		return FALSE;

	press_handled = FALSE;

	pathname = screem_page_get_pathname( page );

	self = g_module_open( NULL, G_MODULE_BIND_LAZY );

	/* does string match any of the editor keys set by the user? */
	string = NULL;
	if( application->macros ) {
		string = convert_keysym_state_to_string( event->keyval, 
							 event->state );
	}
	macro = NULL;

	for( list = application->macros; list; list = list->next ) {
		macro = (ScreemEditorMacro*)list->data;
		if( ! strcmp( string, macro->key_name ) ) {
			if( macro->action == TEXT ) {
				pos = screem_editor_get_pos( editor );
				screem_editor_insert( editor, pos, 
						      macro->text );
				pos += strlen( macro->text );
				screem_editor_set_pos( editor, pos );
				g_free( string );
				return TRUE;
			}
			break;
		} else {
			macro = NULL;
		}
	}
	if( string ) {
		g_free( string );
	}

	buffer = gtk_text_view_get_buffer( GTK_TEXT_VIEW( private->view ) );
	table = gtk_text_buffer_get_tag_table( buffer );
	feature_markup = ( g_object_get_data( G_OBJECT( table ),
					      "feature_markup" ) != NULL );

	if( feature_markup ) {
		press_handled = html_key_press( editor, event, NULL );
	}
	
	
	return press_handled;
}

gboolean html_key_press( ScreemEditor *editor, GdkEventKey *event, 
			 ScreemEditorMacro *macro )
{
	gchar *text;
	gchar *tag = NULL;
	gchar *tag2;
	gchar *format;
	gint pos;
	gint start;
	gint len;
	gint len2;
	gboolean empty = FALSE;
	ScreemPage *page;
	const gchar *charset;

	ScreemEditorPrivate *private;
	GConfClient *client;
	ScreemDTD *dtd;
	ScreemWindow *window;

	gint entinsert;

	g_object_get( G_OBJECT( editor ), "window", &window, NULL );

	private = editor->private;
	client = private->client;

	pos = screem_editor_get_pos( editor );
	page = private->page;
	charset = screem_page_get_charset( page );

	text = screem_page_get_data( page );

	/* Are we in an inline for scripting languages, 
	   ie  <? ?>  <% %> <?php ?> <!-- --> if so we don't want
	   any features, apart from possibly auto indent */
	if( ( screem_markup_is_tag( text, pos, &start, NULL ) ) &&
	    event->keyval != GDK_Tab &&
	    strncmp( "<?xml ", &text[ start ], strlen( "<?xml " ) ) &&
	    ( text[ start + 1 ] == '?' || text[ start + 1 ] == '%' ||
	      text[ start + 1 ] == '!' ) ) {
		g_free( text );
		return FALSE;
	}

	/* act upon key combos, the user must want them otherwise
	   they wouldn't have pressed the combo so the check
	   is made before the script check */
	if( macro ) {
		switch( macro->action ) {
		case ADD_CLOSE:
			/* FIXME: place code in sensible place,
			 *        check tag is allowed to be closed
			 */

			if( start ) {
				start --;
				tag = screem_markup_next_tag( text,
							      start, 
							      NULL,
							      &start, &tag2);
				g_free( tag );
			} else
				tag2 = NULL;
			if( tag2 && tag2[ 0 ] != '/' &&
			    !screem_markup_next_tag_close( text, tag2,
							   start ) ){
				tag = g_strconcat( "</", tag2, ">", NULL );
				start ++;
				screem_editor_insert( editor, start, tag );
				screem_editor_set_pos( editor, start );
				g_free( tag );
				g_free( tag2 );
				g_free( text );
				g_signal_stop_emission_by_name(G_OBJECT(private->view), 
							       "key_press_event");
				return TRUE;
			} else{
				g_free( tag2 );
			}
			break;
		default:
			/* we don't recognise it, this shouldn't happen */
			g_assert( FALSE );
			break;
		}
	}

#if 0
	g_signal_emit( G_OBJECT( editor ),
		       screem_editor_signals[ REQUEST_CONTEXT ],
		       0, pos, &tag, NULL, NULL, NULL, TRUE );

	/* if CDATA is a valid child then we don't want to
	   offer auto tag completion etc. */
	g_print( "IN TAG: %s\n", tag );
	if( tag ) {
		if( ! g_strcasecmp( "script", tag ) ) {
			/* FIXME: should id what language the script is
			   and support key presses for that language,
			   eg bracket expansion etc */
			g_free( tag );
			g_free( text );
			return FALSE;
		}
		g_free( tag );
	}
	tag = NULL;
#endif
	switch( event->keyval ) {
	case GDK_Return:
		g_free( text );
		if( gconf_client_get_bool( client, "/apps/screem/editor/auto_indent", NULL ) ) {
			screem_editor_insert( editor, pos, "\n" );
			pos ++;
			screem_editor_indent( editor, pos );
			g_signal_stop_emission_by_name(G_OBJECT(private->view),
						       "key_press_event" );
			return TRUE;
		}
		break;
	case GDK_space:
		if( ( private->attr_up ) || 
		    ( ! gconf_client_get_bool( client, "/apps/screem/editor/inline_tagging", NULL ) ) ||
		      ( screem_markup_is_attribute( text, pos, NULL, NULL ) )
		    || ( ! start ) )  {
			g_free( text );
			return FALSE;
		}

		start --;
		tag = screem_markup_next_tag(text, start, &start, NULL, &tag2);
		if( ! tag ) {
			return FALSE;
		}
		len = strlen( tag );

		len2 = strlen( tag2 );
		g_free( text );

		if( pos <= start + len2 ) {
			/* not in the tag, or space pressed
			   in the tag name */
			g_free( tag2 );
			g_free( tag );
			return FALSE;
		}

		/* if intelliclose is enabled, and the tag can't be
		   closed, and we are an XML document we should put
		   in the closing / part of the tag if it isn't already
		   present */
		if( gconf_client_get_bool( client,
					   "/apps/screem/editor/intelliclose",
					   NULL  ) ) {
			ScreemDTDClose close;
			const gchar *id;

			dtd = screem_window_get_current_dtd( SCREEM_WINDOW( window ) );
			id = NULL;
			close = MUST;
			if( dtd ) {
				id = g_object_get_data( G_OBJECT( dtd ),
							"type" );
				close = screem_dtd_element_close_state( dtd,
									tag2 );
			}
			if( close == MUSTNT && id && 
			    ( ( ! strncmp( "<?xml ",
					   text, strlen( "<?xml " ) ) ) ||
			      ( strstr( id, " XHTML " ) || 
				strstr( id, " XML " ) ) ) &&
			    tag[ len - 2 ] != '/' ) {
				screem_editor_insert( editor, start + len - 1,
						      " /" );
				screem_editor_set_pos( editor, pos );
			}
		}
		/* we must know about the tag and also not be positioned
		   inside its name in the editor */
		if( tag_attributes_menu( editor, tag2 ) ) {
			g_signal_stop_emission_by_name(G_OBJECT(private->view),
						       "key_press_event" );
			g_free( tag2 );
			g_free( tag );
			return TRUE;
		}
		g_free( tag2 );
		g_free( tag );
		break;
	case GDK_slash:
		if( ! gconf_client_get_bool(client,
					    "/apps/screem/editor/intelliclose",
					    NULL ) ||  ! start  ) {
			g_free( text );
			return FALSE;
		}
		if( ( ( start < ( pos - 1 ) ) &&  text[ pos - 1 ] != '<' )
		    || text[ pos ] != '>' ) {
			g_free( text );
			return FALSE;
		}
	
		screem_editor_insert( editor, pos, "/" );

		g_signal_emit( G_OBJECT( editor ),
			       screem_editor_signals[ REQUEST_CONTEXT ],
			       0, pos, &tag2, NULL, NULL, NULL, TRUE );
		if( tag2 && 
		    ! screem_markup_next_tag_close( text, tag2, start + 1 ) ) {
			/* we found a tag that needed closing */
			screem_editor_insert( editor, pos + 1, tag2 );
			g_signal_stop_emission_by_name(G_OBJECT(private->view),
						       "key_press_event");
			screem_editor_set_pos( editor, pos + strlen( tag2 ) + 2 );
			g_free( tag2 );
		} else {
			if( tag2 )
				tag =g_strdup_printf(_("%s is already closed"),
						     tag2 );
			else
				tag=g_strdup_printf(_("no tag needs closing"));
			screem_window_show_message( SCREEM_WINDOW( window ),
						    tag );
			g_free( tag );
			g_free( tag2 );
		}

		g_free( text );
	   		
		return TRUE;
		break;
	default:
		if( ! ( tag = screem_markup_char_to_ent( event->keyval ) ) ) {
			g_free( text );
			return FALSE;
		}
		
		/* we're about to insert a character with a special HTML
		   meaning (tag or entity). To make sure that this can
		   be done we have to check if we're already inside another
		   entity. Look for a semicolon after the current position
		   AND an ampersand before the current position.
		   This search is aborted as soon as we find a character that
		   is part of another HTML construct */
		if( pos != 0 ) {
			empty = ( text[ pos - 1 ] == '&' && 
				  text[ pos ] == ';' );
		} else {
			empty = FALSE;
		}
		
		if( ! empty && screem_markup_is_entity(text,pos,NULL,NULL) ) {
			/* let the widget handle insertion as normal as
			   the entity
			   already exists and is not empty */
			g_free( text );
			return FALSE;
		}

		if( empty ) {
			format = "%s";
		} else if( text[ pos ] == ';' ) {
			format = "&%s";
		} else if( pos > 0 && text[ pos - 1 ] == '&' ) {
			format = "%s;";
		} else {
			format = "&%s;";
		}

		tag = g_strdup_printf( format, tag );

		/* special cases */
		if( event->keyval == GDK_ampersand && ! empty ) {
			g_free( tag );
			tag = g_strdup_printf( format , "" ) ;
		} else if( event->keyval == GDK_quotedbl &&
			   ( !(empty |= screem_markup_is_attribute(text,
								   pos,
								   NULL,
								   NULL)) ) ) {
			/* if pos - 1 == "=" then insert two, place between */
			g_free( tag );
			if( text[ pos - 1 ] == '=' &&
			   ( text[ pos ] == ' ' || text[ pos ] == '>' ) )
				tag = g_strdup_printf( "\"\"" );
			else {
				g_free( text );
				return FALSE;
			}
		} else if( event->keyval == GDK_Tab && ! empty ) {
			g_free( tag );
			if( ! gconf_client_get_bool( client, "/apps/screem/editor/auto_indent", NULL ) ) {
				/* insert a normal tab as not in an entity,
				   and auto indent is off */
				g_free( text );
				return FALSE;
			}
			screem_editor_indent( editor, pos );
			g_free( text );
			g_signal_stop_emission_by_name(G_OBJECT(private->view),
						       "key_press_event" );
			return TRUE;
		} else if( event->keyval == GDK_less && ! empty ) {
			/* We add this manually along with a
			   GDK_greater, unless we are found to be
			   inside a tag */
			g_free( tag );
			if( screem_markup_is_tag( text, pos - 1, NULL, NULL ) &&
			    text[ pos - 1 ] != '>' ) {
				/* in tag so let widget do it */
				g_free( text );
				return FALSE;
			}
			tag = g_strdup_printf( "<>" );
		} else if( event->keyval == GDK_greater && ! empty ) {
			/* to keep consistant with GDK_less we
			   do a normal insert if we aren't inside
			   and empty entity */
			g_free( tag );
			g_free( text );
			return FALSE;
		}
		
		g_free( text );

		/* are we performing entity insertion? */
		entinsert = gconf_client_get_int( client,
						  "/apps/screem/editor/entityinsertion", NULL );
		if( entinsert != ON ) {
			switch( event->keyval ) {
			case GDK_quotedbl:
			case GDK_Tab:
			case GDK_less:
			case GDK_greater:
				if( empty && entinsert == OFF ) {
					/* we only don't do these for off */
					g_free( tag );
					return FALSE;
				}
				break;
			case GDK_ampersand:
				if( entinsert == BY_SET )
					break;
			default:
				if( entinsert == BY_SET &&
				    ! g_strcasecmp( "ISO-8859-1", charset ) )
					break;
				g_free( tag );
				return FALSE;
			}
		}

		screem_editor_insert( editor, pos, tag );
		if( empty && event->keyval != GDK_quotedbl ) {
			pos += ( strlen(tag) + 1 );
		} else if( event->keyval == GDK_less ||
			   event->keyval == GDK_ampersand ||
			   ! strcmp( "\"\"", tag ) ) {
			pos ++;
		} else {
			pos += strlen( tag );
		}
		
		g_free( tag );
		screem_editor_set_pos( editor, pos );
		g_signal_stop_emission_by_name( G_OBJECT( private->view ), 
						"key_press_event" );
		return TRUE;
		break;
	}
	return FALSE;
}

static gboolean screem_editor_tip( ScreemEditor *editor )
{
	ScreemEditorPrivate *private;

	private = editor->private;

	if( private->tooltip_timer != -1 ) {
		g_source_remove( private->tooltip_timer );
	}
	private->tooltip_timer = g_timeout_add( 2000, 
						(GSourceFunc)screem_editor_tooltip,
						editor );

	return FALSE;
}

static gboolean screem_editor_tooltip( ScreemEditor *editor )
{
	gint pos;
	gchar *text;
	ScreemPage *page;
	const gchar *mime_type;
	GModule *self;
	gboolean handled;

	GdkWindow *win;
	gint x;
	gint y;
	GdkModifierType mask;
	gint bx;
	gint by;
	GtkTextIter it;

	GtkTextBuffer *buffer;
	GtkTextTagTable *table;
	gboolean feature_markup;

	page = editor->private->page;

	if( ! page ) {
		return FALSE;
	}

	mime_type = screem_page_get_mime_type( page );

	self = g_module_open( NULL, G_MODULE_BIND_LAZY );

	win = gtk_text_view_get_window(GTK_TEXT_VIEW(editor->private->view),
				       GTK_TEXT_WINDOW_TEXT);
	gdk_window_get_pointer( win, &x, &y, &mask );

	gtk_text_view_window_to_buffer_coords(GTK_TEXT_VIEW(editor->private->view ),
					      GTK_TEXT_WINDOW_TEXT,
					      x, y, &bx, &by );
	gtk_text_view_get_iter_at_location(GTK_TEXT_VIEW(editor->private->view),
					   &it, bx, by);
	pos = gtk_text_iter_get_offset( &it );

	text = screem_page_get_data( page );

	handled = FALSE;

	buffer = gtk_text_view_get_buffer( GTK_TEXT_VIEW( editor->private->view ) );
	table = gtk_text_buffer_get_tag_table( buffer );
	feature_markup = ( g_object_get_data( G_OBJECT( table ),
					      "feature_markup" ) != NULL );

	if( feature_markup ) {
		ScreemWindow *window;

		g_object_get( G_OBJECT( editor ), "window", &window, NULL );
		handled = html_editor_tip( editor, window, pos, text );
		
		if( ! handled && ! strcmp( "application/x-php", mime_type ) ) {
			handled = php_tooltip( editor, window,
					       pos, text );
		}
	}

	g_free( text );

	editor->private->tooltip_timer = -1;

	return FALSE;
}

static gboolean html_editor_tip( ScreemEditor *editor, ScreemWindow *window,
				 gint pos, gchar *text )
{
	ScreemDTD *dtd;
	gchar *tag;
	gchar *name;
	ScreemDTDElement *element;
	gchar *message;

	if( screem_markup_is_tag( text, pos, &pos, NULL ) ) {
		tag = screem_markup_next_tag( text, pos, NULL, &pos, &name );
		/* so we can tooltip PHP/ASP etc functions */
		if( tag[ 1 ] == '?' || tag[ 1 ] == '!' ) {
			g_free( name );
			return FALSE;
		}
		/* we have the tag, look it up in the DTD */
		dtd = screem_window_get_current_dtd( window );
		element = screem_dtd_valid_element( dtd, name );
		if( ! element && *name != '/' ) {
			message = g_strdup_printf( _("%s is not a valid element of this documents DTD" ), name );
			screem_window_show_message( window, message );

			g_free( message );
		} else if( *name != '/' ) {
			/* its valid, but is it allowed here? */
			message = g_strdup( "" );

			g_free( message );
		}

		g_free( name );
		g_free( tag );
		return TRUE;
	}

	return FALSE;
}

static gboolean tag_attributes_menu( ScreemEditor *editor, gchar *name )
{
	/* passing NULL here as the first param causes the default DTD to
	   be used */
	ScreemDTD *dtd;
	GtkWidget *menu;
	ScreemWindow *window;

	g_object_get( G_OBJECT( editor ), "window", &window, NULL );
	dtd = screem_window_get_current_dtd( window );

	editor->private->attr_menu = NULL;
	if( dtd ) {
		editor->private->attr_menu = 
			screem_dtd_attribute_menu( dtd, name,
						   editor );
	}
	
	if( ! editor->private->attr_menu ) {
		return FALSE;
	}

	menu = gtk_menu_new();
	gtk_menu_shell_append( GTK_MENU_SHELL( menu ),
			       editor->private->attr_menu );

	/* attach a callback for a key press, that way the user
	   can ignore the popup and keep typing */
	g_signal_connect( G_OBJECT( menu ), "key_press_event",
			  G_CALLBACK( no_attribute ), editor );
	gnome_popup_menu_do_popup( menu, 0, 0, 0, 0, 0 );

	return TRUE;
}

static gboolean no_attribute( GtkWidget *widget, GdkEventKey *event, ScreemEditor *editor )
{
        gint pos;
	gboolean ret;

        if( ! isascii( event->keyval ) && event->keyval != GDK_Escape )
                return FALSE;

        g_signal_stop_emission_by_name( G_OBJECT( widget ),
					"key_press_event" );

        /* if attrMenu is NULL then we must have come from the
           editor popup menu, which doesn't want this functionality */
        if( ! editor->private->attr_menu )
                return FALSE;

        /* insert initial press of space */
	pos = screem_editor_get_pos( editor );
        screem_editor_insert( editor, pos, " " );
	screem_editor_set_pos( editor, ++ pos );
        /* a key was pressed while the attribute menu was open,
           so emit a key_press_event signal on the editor */

        editor->private->attr_up = TRUE;
	if( event->keyval != ' ' )
		g_signal_emit_by_name( G_OBJECT( editor->private->view ),
				       "key_press_event", event, &ret );
        /* kill the menu */
        gtk_widget_destroy( widget );
        editor->private->attr_menu = NULL;
        editor->private->attr_up = FALSE;

	return ret;
}

/* Line number code ( get_lines(), screem_editor_expose() ) is from / based on 
   testtext.c in gtk+  ( and fixed up with src from gedit-view.c in gedit */
static void get_lines( GtkTextView  *text_view,
		       gint          first_y,
		       gint          last_y,
		       GArray       *buffer_coords,
		       GArray       *numbers,
		       gint         *countp,
		       gint         *totalCount )
{
	GtkTextIter iter;
	gint count;
	gint size;  
	gint last;
	
	gint line_num;

	if( totalCount ) {
		GtkTextBuffer *buffer;	

		buffer = gtk_text_view_get_buffer( text_view );

		gtk_text_buffer_get_end_iter( buffer, &iter );
		*totalCount = gtk_text_iter_get_line( &iter );
	}

	/* Get iter at first y */
	gtk_text_view_get_line_at_y( text_view, &iter, first_y, NULL );
	
	/* For each iter, get its location and add it to the arrays.
	 * Stop when we pass last_y
	 */
	count = 0;
	size = 0;
	
	last = -1;
	line_num = gtk_text_iter_get_line( &iter );
	while( ! gtk_text_iter_is_end( &iter ) ) { 
		gint y;
		gint height;
		
		gtk_text_view_get_line_yrange( text_view, &iter, &y, &height );
		last = line_num;
		g_array_append_val( buffer_coords, y );
		g_array_append_val( numbers, line_num );
		++ count;
		++ line_num;
		if( ( y + height ) >= last_y ) {
			break;
		}
		gtk_text_iter_forward_line( &iter );
	}
	       
	if( gtk_text_iter_is_end( &iter ) &&( last != line_num ) ) {
		gint y;
		gint height;
		
                gtk_text_view_get_line_yrange( text_view, &iter, &y, &height );
		g_array_append_val( buffer_coords, y );
		g_array_append_val( numbers, line_num );
		++count;
	}
	
	*countp = count;
}

static gint screem_editor_expose( GtkWidget *widget, GdkEventExpose *event,
				  ScreemEditor *editor )
{
	gint layout_width;
	gint count;
	gint total;
	GArray *numbers;
	GArray *pixels;
	gint first_y;
	gint last_y;
	gint i;
	gchar *str;
	GdkWindow *left_win;
	PangoLayout *layout;
	GtkTextView *text_view;
	GtkTextWindowType type;
	GdkDrawable *target;

	gint line = 0;
	gint first = 0;
	gint last = 0;

	ScreemPage *page;

	page = editor->private->page;

	if( ! page ) {
		return TRUE;
	}

	text_view = GTK_TEXT_VIEW( widget );
	
	/* See if this expose is on the line numbers window */
	left_win = gtk_text_view_get_window( text_view,
					     GTK_TEXT_WINDOW_LEFT );
		
	if( event->window == left_win ) {
		type = GTK_TEXT_WINDOW_LEFT;
		target = left_win;
	} else {
		return FALSE;
	}
	
	first_y = event->area.y;
	last_y = first_y + event->area.height;
	
	gtk_text_view_window_to_buffer_coords( text_view,
					       type,
					       0,
					       first_y,
					       NULL,
					       &first_y );
	
	gtk_text_view_window_to_buffer_coords( text_view,
					       type,
					       0,
					       last_y,
					       NULL,
					       &last_y );
	
	numbers = g_array_new( FALSE, FALSE, sizeof (gint) );
	pixels = g_array_new( FALSE, FALSE, sizeof (gint) );
	
	get_lines( text_view,
		   first_y,
		   last_y,
		   pixels,
		   numbers,
		   &count,
		   &total );

	layout = gtk_widget_create_pango_layout( widget, "" );

	/* Set size, we use total * 10 as we want 1 digit of padding */
        str = g_strdup_printf( "%d", MAX( 99, total * 10 ) );
	pango_layout_set_text (layout, str, -1);
        g_free (str);
	pango_layout_get_pixel_size( layout, &layout_width, NULL );
	gtk_text_view_set_border_window_size( text_view,
					      GTK_TEXT_WINDOW_LEFT,
					      layout_width + 4 + GUTTER_SIZE );
	
	/* Draw fully internationalized numbers! */

     	i = 0;
	if( count > 0 ) {
		line = g_array_index( numbers, gint, 0 );
		first = line;
		last = g_array_index( numbers, gint, numbers->len - 1 );
	}
	while( i < count ) {
		gint pos;
		const gchar *bookmark;

		gtk_text_view_buffer_to_window_coords( text_view,
						       type,
						       0,
						       g_array_index( pixels, 
								      gint, i),
						       NULL,
						       &pos );
		/* check for bookmarks etc */
		bookmark = screem_page_bookmarked(editor->private->page, 
						  ++ line );

		if( bookmark ) {
			GSList *list;
			GtkWidget *child;

			for( child = NULL, list = editor->private->bookmarks;
			     list; list = list->next ) {
				gint l;

				l = GPOINTER_TO_INT( g_object_get_data( G_OBJECT( list->data ), "line" ) );
				if( l == line ) {
					child = GTK_WIDGET( list->data );
					break;
				} else if( l < first || l > last ) {
					gtk_text_view_move_child( text_view,
								  GTK_WIDGET( list->data ),
								  -32, 0 );
				}
			}
			if( child ) {
				gtk_text_view_move_child( text_view,
							  child,
							  layout_width + 2,
							  pos );
			} else {
				GtkWidget *event;
				GtkWidget *child;
				
				event = gtk_event_box_new();
				child = gtk_image_new_from_stock("Screem_Bookmark",
								 GTK_ICON_SIZE_MENU );
				gtk_container_add( GTK_CONTAINER( event ),
						   child );
				g_object_set_data( G_OBJECT( event ), "line",
						   GINT_TO_POINTER( line ) );

				editor->private->bookmarks = g_slist_prepend( editor->private->bookmarks, event );
				gtk_widget_show_all( event );
				gtk_text_view_add_child_in_window(text_view,
								  event,
								  GTK_TEXT_WINDOW_LEFT,
								  layout_width + 2,
								  pos );
			}
		} else {
			GSList *list;

			list = editor->private->bookmarks;
			while( list ) {
				gint l;

				l = GPOINTER_TO_INT( g_object_get_data( G_OBJECT( list->data ),
									"line" ) );
				if( l == line ) {
					GSList *tmp;
					gtk_widget_destroy( GTK_WIDGET( list->data ) );
					tmp = list;
					list = list->next;
					editor->private->bookmarks = g_slist_remove( editor->private->bookmarks, tmp->data );
				} else {
					list = list->next;
				}
			}
		}
		
		str = g_strdup_printf("%d", line );
				
		pango_layout_set_markup( layout, str, -1 );
	
		gtk_paint_layout( widget->style,
				  target,
				  GTK_WIDGET_STATE( widget ),
				  FALSE,
				  NULL,
				  widget,
				  NULL,
				  2, pos,
				  layout );
			
		g_free( str );
			
		++i;
	}
	g_array_free( pixels, TRUE );
	g_array_free( numbers, TRUE );
	
	g_object_unref( G_OBJECT( layout ) );
	
	return TRUE;
}


static void screem_editor_colour_notify( GConfClient *client,
					 guint cnxn_id,
					 GConfEntry *entry,
					 gpointer data )
{
	screem_editor_set_colours( SCREEM_EDITOR( data ) );
}


static void screem_editor_clip_received( GtkClipboard *clipboard,
					 GtkSelectionData *selection_data,
					 gpointer data )
{
	ScreemEditor *editor;

	editor = SCREEM_EDITOR( data );

	if( selection_data->length > 0 ) {
		gchar *text = 
			screem_markup_encode_text( selection_data->data );
		screem_editor_insert( editor, -1, text );
		g_free( text );
	}
}

static void screem_editor_toggle_overwrite( GtkTextView *view,
					    ScreemEditor *editor )
{
	gchar *mesg;

	gtk_statusbar_pop( GTK_STATUSBAR( editor->private->overwriteMode ),
			   0 );

	editor->private->overwrite = ! editor->private->overwrite;

	if( editor->private->overwrite ) {
		mesg = g_strdup( _( "  OVR" ) );
	} else {
		mesg = g_strdup( _( "  INS" ) );
	}

	gtk_statusbar_push( GTK_STATUSBAR( editor->private->overwriteMode ),
			    0, mesg );

	g_free( mesg );
}

static gboolean build_attribute_list( ScreemEditor *editor )
{
    	GtkWidget *box;
	GtkBoxChild *bchild;
    
	GSList *set_attr;

	gchar *name;

	gint pos;
	gchar *text;
	gchar *tag;

	GList *list;

	ScreemDTD *dtd;
	ScreemPage *page;
	ScreemWindow *window;

	static gint ppos = -1;

	g_object_get( G_OBJECT( editor ), "window", &window, NULL );

	editor->private->handle = -1;

	page = editor->private->page;

	if( page ) {
		box = editor->private->attributes;
		
		/* id which tag we are in */
		pos = screem_editor_get_pos( editor );
		text = screem_page_get_data( page );
		
		if( ! screem_markup_is_tag( text, pos, &pos, NULL ) ) {
			ppos = pos;
			gtk_widget_set_sensitive( box, FALSE );
			g_free( text );
			return FALSE;
		} else if( ppos == pos ) {
			g_free( text );
			return FALSE;
		}
		ppos = pos;
		
		tag = screem_markup_next_tag( text, pos, NULL, &pos, &name );
		if( tag ) {
			set_attr = screem_markup_build_attributes_list( tag, NULL );
			set_attr = g_slist_reverse( set_attr );
		} else {
			name = NULL;
			set_attr = NULL;
		}
		
		dtd = screem_window_get_current_dtd( window );
		
		g_free( text );
		g_free( tag );
		
		gtk_widget_hide( box );
		
		while( ( list = GTK_BOX( box )->children ) ) {
			bchild = list->data;
			gtk_container_remove( GTK_CONTAINER( box ),
					      bchild->widget );
		}
		if( name ) {
			build_attribute_entries( editor, dtd, name, box,
						set_attr );
			/* destroy all of set_attr */
			g_slist_foreach( set_attr, (GFunc)g_free, NULL );
			g_slist_free( set_attr );
		}
		
		if( name ) {
			g_free( name );
		}
		
		gtk_widget_show_all( box );
		gtk_widget_set_sensitive( box, TRUE );
	}

	return FALSE;
}

static void build_attribute_entries( ScreemEditor *editor,
				     ScreemDTD *dtd, const gchar *name,
				     GtkWidget *box,
				     GSList *attr )
{
    	GtkWidget *line;
	GtkWidget *label;
	GtkWidget *combo;
	GtkWidget *entry;

	GList *combo_list;
	GSList *list;
	GtkSizeGroup *group;

	ScreemDTDElement *element;
	const GSList *attrs;

	if( ! dtd ) {
		return;
	}

	element = screem_dtd_valid_element( dtd, name );
	if( ! element ) {
		return;
	}

	attrs = screem_dtd_element_get_attrs( element );

	group = gtk_size_group_new( GTK_SIZE_GROUP_BOTH );
	
	for( ;attrs; attrs = attrs->next ) {
		ScreemDTDAttribute *att;
		const gchar *attr_name;
		const GSList *values;

		att = (ScreemDTDAttribute*)attrs->data;
		attr_name = screem_dtd_attribute_get_name( att );
		
		values = screem_dtd_attribute_get_values( att );

		/* add entrys */
		line = gtk_table_new( 2, 1, FALSE );
		label = gtk_label_new( attr_name );
		gtk_label_set_justify( GTK_LABEL( label ), GTK_JUSTIFY_LEFT );
		gtk_misc_set_alignment( GTK_MISC( label ), 0.0, 0.5 );
		gtk_table_attach( GTK_TABLE( line ), label, 0, 1, 0, 1,
				  GTK_FILL, 
				  GTK_FILL, 4, 4 );
		gtk_size_group_add_widget( group, label );

		combo = gtk_combo_new();
		gtk_widget_set_size_request( combo, 128, -1 );
		entry = GTK_COMBO( combo )->entry;
		gtk_table_attach( GTK_TABLE( line ), combo, 1, 2, 0, 1,
				  GTK_FILL | GTK_EXPAND, GTK_FILL, 4, 4 );
		gtk_box_pack_start( GTK_BOX( box ), line, FALSE, FALSE, 0 );
		
		/* add combo entries */
		combo_list = NULL;
		for( ; values; values = values->next ) {
			combo_list = g_list_prepend( combo_list, values->data );
		}
		if( combo_list ) {
			combo_list = g_list_reverse( combo_list );
			gtk_combo_set_popdown_strings( GTK_COMBO( combo ), 
						       combo_list );
			g_list_free( combo_list );
		}
		
		/* 
		 * set the entry value to whatever value the
		 * attribute currently has in the tag, or make it
		 * empty if the attribute isn't used
		 */

		/* find name in the attr list */
		for( list = attr; list; list = list->next ) {
			if( ! strcmp( attr_name, (gchar*)list->data ) )
				break;
			list = list->next;
		}
		
		if( ! list ) {
			gtk_entry_set_text( GTK_ENTRY( entry ), "" );
		} else if( ! list->next ) {
			gtk_entry_set_text( GTK_ENTRY( entry ), "1" );
		} else {
			gtk_entry_set_text( GTK_ENTRY( entry ),
					    list->next->data );
		}

		g_object_set_data( G_OBJECT( entry ), "editor", editor );

		gtk_combo_disable_activate( GTK_COMBO( combo ) );
		g_signal_connect( G_OBJECT( entry ), "activate",
				  G_CALLBACK( replace_attr ), 
				  (gpointer)attr_name );

		gtk_widget_show_all( line );
	}
}

static void replace_attr( GtkWidget *entry, gchar *attr_name )
{
	ScreemEditor *editor;
	const gchar *text;
	gchar *attr;
	gint pos;

	editor = SCREEM_EDITOR( g_object_get_data( G_OBJECT( entry ),
						   "editor" ) );

	text = gtk_entry_get_text( GTK_ENTRY( entry ) );

	attr = g_strconcat( attr_name, "=\"", text, "\"", NULL );

	pos = screem_editor_get_pos( editor );
	screem_editor_insert_attribute( editor, attr );
	screem_editor_set_pos( editor, pos );

	g_free( attr );
}


static gboolean screem_editor_switch_mime( GtkWidget *widget, 
					   GdkEventButton *event,
					   ScreemEditor *editor )
{
	g_signal_stop_emission_by_name( G_OBJECT( widget ),
					"button_press_event" );
	
	gnome_popup_menu_do_popup_modal( editor->private->types,
					 0, 0, event, 0, widget );
	gtk_widget_hide( editor->private->types );

	return TRUE;
}

static void screem_editor_mime_switch( GtkWidget *widget, 
				       ScreemEditor *editor )
{
	const gchar *mime_type;
	ScreemPage *page;

	page = editor->private->page;

	mime_type = g_object_get_data( G_OBJECT( widget ), "label" );

	screem_editor_display_page( editor, NULL );

	screem_page_set_mime_type( page, mime_type );

	screem_editor_display_page( editor, page );
}


static gboolean screem_editor_drop_uris( ScreemEditor *editor,
					 Drop_actions action,
					 const gchar *selectionData )
{
	gboolean online;
	GList *uris;
	GList *tmp;
	GnomeVFSURI *uri;
	
	ScreemPage *page;
	const gchar *pathname;
	const gchar *mime_type;
	gboolean ret;

	guint pos;

	g_object_get( G_OBJECT( editor ), "online", &online, NULL );
	page = editor->private->page;

	ret = FALSE;

	pathname = screem_page_get_pathname( page );

	uris = gnome_vfs_uri_list_parse( selectionData );
	for( tmp = uris; tmp; tmp = tmp->next ) {
		gchar *url;
		uri = (GnomeVFSURI*)tmp->data;
		
		url = gnome_vfs_uri_to_string( uri, 0 );
		if( online || gnome_vfs_uri_is_local( uri ) ) {
			mime_type = gnome_vfs_get_mime_type( url );
		} else {
			mime_type = NULL;
		}
		/* INLINE action selected, can only do this,
		   if we have a mime type that is for a text file */
		if( action == INLINE && 
		    screem_page_is_mime_type_page( mime_type ) ) {
			screem_editor_insert_file( editor,
						   url );
			g_free( url );
			url = NULL;
			ret = TRUE;
		} else if( action != FULL_PATH && pathname ) {
			gchar *dir;
			gchar *temp;
			
			dir = g_dirname( pathname );
			temp = url;
			url = relative_path( url, dir );
			g_free( dir );
				g_free( temp );
		}
		if( action == TAG ) {
			gchar *temp;

			temp = url;
			if( mime_type && ! strncmp(mime_type, "image/", 
						   strlen("image/") ) ) {
				url = g_strdup_printf( "<img src=\"%s\" alt=\"\">", temp );
			} else {
				url = g_strdup_printf( "<a href=\"%s\"> </a>",
						       temp );
			}
			g_free( temp );
		}
		if( url ) {
			pos = screem_editor_get_pos( editor );
			screem_editor_insert( editor, pos, url );
			g_free( url );
			ret = TRUE;
		}
		
		gnome_vfs_uri_unref( uri );
	}
	g_list_free( uris );

	return ret;
}

static void set_pos( ScreemEditorDelaySetPos *delay )
{
	screem_editor_set_pos( delay->editor, delay->pos );
	g_free( delay );
}

/* G Object stuff */

#define PARENT_TYPE SCREEM_TYPE_VIEW

static gpointer parent_class;

static void screem_editor_class_init( ScreemEditorClass *klass )
{
	GObjectClass *object_class;
	GtkWidgetClass *widget_class;

	object_class = G_OBJECT_CLASS( klass );
	widget_class = (GtkWidgetClass *)klass;
	parent_class = g_type_class_peek_parent( klass );

	object_class->finalize = screem_editor_finalize;

	widget_class->show = screem_editor_show;
	widget_class->hide = screem_editor_hide;
	widget_class->realize = screem_editor_realize;
	widget_class->size_request = screem_editor_size_request;

	screem_editor_signals[ REQUEST_CONTEXT ] = 
		g_signal_new( "request_context",
			      G_OBJECT_CLASS_TYPE( object_class ),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET( ScreemEditorClass, 
					       request_context),
			      NULL, NULL,
			      screem_marshal_VOID__UINT_POINTER_POINTER_POINTER_POINTER_BOOLEAN,
			      G_TYPE_NONE, 6,
			      G_TYPE_UINT,
			      G_TYPE_POINTER,
			      G_TYPE_POINTER,
			      G_TYPE_POINTER,
			      G_TYPE_POINTER,
			      G_TYPE_BOOLEAN );
}

static void screem_editor_init( ScreemEditor *editor )
{
	GtkWidget *sw;
	GtkWidget *text;
	ScreemEditorPrivate *private;

	SCREEM_VIEW( editor )->window_set = screem_editor_window_set;
	SCREEM_VIEW( editor )->display = screem_editor_display;
	SCREEM_VIEW( editor )->print = screem_editor_print;
	SCREEM_VIEW( editor )->undo = screem_editor_undo;
	SCREEM_VIEW( editor )->redo = screem_editor_redo;

	private = editor->private = g_new0( ScreemEditorPrivate, 1 );

	private->client = gconf_client_get_default();
	gconf_client_add_dir( private->client, "/apps/screem/editor",
			      GCONF_CLIENT_PRELOAD_ONELEVEL, NULL );

	private->handle = -1;
	private->popup_open = FALSE;
	private->inserted = FALSE;
	private->attr_up = FALSE;
	private->tooltip_timer = -1;

	private->view = text = gtk_text_view_new();
	gtk_widget_show( text );

	gtk_text_view_set_border_window_size( GTK_TEXT_VIEW( text ),
					      GTK_TEXT_WINDOW_LEFT,
					      30 );

	sw = private->sw = gtk_scrolled_window_new( NULL, NULL );
	gtk_widget_show( sw );
	gtk_scrolled_window_set_policy( GTK_SCROLLED_WINDOW( sw ), 
                                        GTK_POLICY_AUTOMATIC,
                                        GTK_POLICY_AUTOMATIC );
	gtk_scrolled_window_set_shadow_type( GTK_SCROLLED_WINDOW( sw ),
						GTK_SHADOW_IN );
	gtk_container_add( GTK_CONTAINER( sw ), text );

	g_signal_connect( G_OBJECT( text ), "button_press_event",
			  G_CALLBACK( screem_editor_popup ), editor );
	
	g_signal_connect( G_OBJECT( text ), "button_release_event",
			  G_CALLBACK( screem_editor_button_release ),
			  editor );

	g_signal_connect( G_OBJECT( text ), "popup_menu",
			  G_CALLBACK( screem_editor_popup_key ), editor );
	
	g_signal_connect( G_OBJECT( text ), "key_press_event",
			  G_CALLBACK( screem_editor_keypress ), editor );
	
	
	g_signal_connect( G_OBJECT( text ), "expose_event",
			  G_CALLBACK( screem_editor_expose ),
			  editor );
	
	gtk_drag_dest_set( text,
			   GTK_DEST_DEFAULT_MOTION | 
			   GTK_DEST_DEFAULT_HIGHLIGHT,
			   drop_types, num_drop_types,
			   GDK_ACTION_COPY | GDK_ACTION_ASK );
	
	g_signal_connect( G_OBJECT( text ), "drag_data_received",
			  G_CALLBACK( screem_editor_drop ), editor );
	g_signal_connect( G_OBJECT( text ), "drag_motion",
			  G_CALLBACK( screem_editor_motion ), editor );

	g_signal_connect( G_OBJECT( editor->private->view ),
			  "toggle_overwrite",
			  G_CALLBACK( screem_editor_toggle_overwrite ),
			  editor );

	private->notifies[ 0 ] = 
		gconf_client_notify_add( private->client,
					 "/apps/screem/editor/back",
					 screem_editor_colour_notify,
					 editor, NULL, NULL );
	private->notifies[ 1 ] = 
		gconf_client_notify_add( private->client,
					 "/apps/screem/editor/text",
					 screem_editor_colour_notify,
					 editor, NULL, NULL );
	private->notifies[ 2 ] = 
		gconf_client_notify_add( private->client,
					 "/apps/screem/editor/font",
					 screem_editor_colour_notify,
					 editor, NULL, NULL );
	
	gtk_text_view_set_wrap_mode( GTK_TEXT_VIEW( text ), 
				     GTK_WRAP_WORD );
	
	
	g_signal_connect_data( G_OBJECT( text ),
			       "motion_notify_event",
			       G_CALLBACK( screem_editor_tip ), 
			       editor,
			       NULL, G_CONNECT_SWAPPED );

	gtk_container_add( GTK_CONTAINER( editor ), sw );
}

static void screem_editor_finalize( GObject *object )
{
	ScreemEditor *editor;
	ScreemEditorPrivate *private;
	
	editor = SCREEM_EDITOR( object );
	private = editor->private;
	
	if( private->handle != -1 ) {
		g_source_remove( private->handle );
	}
	if( private->tooltip_timer != -1 ) {
		g_source_remove( private->tooltip_timer );
	}

	g_free( private );
	
	G_OBJECT_CLASS( parent_class )->finalize( object );
}

static void screem_editor_window_set( ScreemView *view )
{
	ScreemEditor *editor;
	ScreemEditorPrivate *private;
	GtkWidget *status;
	ScreemWindow *window;
	GtkWidget *sw;
	GtkWidget *widget;
	
	editor = SCREEM_EDITOR( view );
	private = editor->private;

	g_object_get( G_OBJECT( view ), "window", &window, NULL );

	private->overwriteMode = gtk_statusbar_new();
	gtk_statusbar_set_has_resize_grip( GTK_STATUSBAR( private->overwriteMode ),
						TRUE );
	gtk_widget_show( private->overwriteMode );
	gtk_box_pack_end( GTK_BOX( window->status ), private->overwriteMode,
				FALSE, FALSE, 0 );
	gtk_widget_set_size_request( private->overwriteMode, 80, 10 );
	
	private->cursorPosition = gtk_statusbar_new();
	gtk_statusbar_set_has_resize_grip( GTK_STATUSBAR( private->cursorPosition ),
						FALSE );
	gtk_widget_show( private->cursorPosition );
	gtk_widget_set_size_request( private->cursorPosition, 150, 10 );	
	gtk_box_pack_end( GTK_BOX( window->status ), private->cursorPosition, 
				FALSE, FALSE, 0 );

	private->mime = gtk_button_new();
	gtk_button_set_relief( GTK_BUTTON( private->mime  ), GTK_RELIEF_NONE );
	gtk_widget_show( private->mime  );
	gtk_box_pack_end( GTK_BOX( window->status ), private->mime,
				FALSE, FALSE, 0 );
	g_signal_connect( G_OBJECT( private->mime ), "button_press_event",
			  G_CALLBACK( screem_editor_switch_mime ),
			  editor );
	gtk_widget_set_size_request( private->mime , 120, 10 );

	private->overwrite = TRUE;
	screem_editor_toggle_overwrite( GTK_TEXT_VIEW( private->view ),
					editor );
	
	/* add attributes tab */
	private->attrtab = sw = 
		gtk_scrolled_window_new( NULL, NULL );
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(sw),
				       GTK_POLICY_AUTOMATIC, 
				       GTK_POLICY_AUTOMATIC );
	gtk_scrolled_window_set_shadow_type( GTK_SCROLLED_WINDOW( sw ),
						GTK_SHADOW_IN );
	private->attributes = gtk_vbox_new( 0, FALSE );
	gtk_scrolled_window_add_with_viewport( GTK_SCROLLED_WINDOW(sw),
					       private->attributes );
	gtk_widget_show_all( sw );
	
	widget = gdl_dock_item_new( "Attributes", _( "Attributes" ), GTK_STOCK_INDEX,
					GDL_DOCK_ITEM_BEH_NORMAL );
	gtk_container_add( GTK_CONTAINER( widget ), sw );
	gdl_dock_add_item( GDL_DOCK( window->dock ), GDL_DOCK_ITEM( widget ),
					GDL_DOCK_RIGHT );
	gtk_widget_show( widget );
					  
}

static void screem_editor_size_request( GtkWidget *widget, GtkRequisition *req )
{
	GTK_WIDGET_CLASS( parent_class )->size_request( widget, req );
}

static void screem_editor_realize( GtkWidget *widget )
{
	GTK_WIDGET_CLASS( parent_class )->realize( widget );
}

static void screem_editor_show( GtkWidget *widget )
{
	ScreemEditor *editor;
	ScreemEditorPrivate *private;
	ScreemWindow *window;
	
	editor = SCREEM_EDITOR( widget );
	private = editor->private;
	
	g_object_get( G_OBJECT( editor ), "window", &window, NULL );
	
	screem_editor_set_colours( editor );

	gtk_widget_set_sensitive( private->attrtab, TRUE );

	GTK_WIDGET_CLASS( parent_class )->show( widget );

	gtk_widget_show( private->view );
	gtk_widget_show( private->sw );
	
	if( window->status_text ) {
		gtk_statusbar_set_has_resize_grip( GTK_STATUSBAR( window->status_text ),
							FALSE );
	}
	
	gtk_widget_show( private->overwriteMode );
	gtk_widget_show( private->cursorPosition );
	gtk_widget_show( private->mime );
}

static void screem_editor_hide( GtkWidget *widget )
{
	ScreemEditor *editor;
	ScreemEditorPrivate *private;
	ScreemWindow *window;
	
	editor = SCREEM_EDITOR( widget );
	private = editor->private;

	g_object_get( G_OBJECT( editor ), "window", &window, NULL );
	
	if( window->status_text ) {
		gtk_statusbar_set_has_resize_grip( GTK_STATUSBAR( window->status_text ),
							TRUE );
	}
	
	gtk_widget_hide( private->overwriteMode );
	gtk_widget_hide( private->cursorPosition );
	gtk_widget_hide( private->mime );
	
	gtk_widget_set_sensitive( private->attrtab, FALSE );
	
	GTK_WIDGET_CLASS( parent_class )->hide( widget );
}

GType screem_editor_get_type()
{
	static GType type = 0;
	
	if( ! type ) {
		static const GTypeInfo info = {
			sizeof( ScreemEditorClass ),
			NULL, /* base init */
			NULL, /* base finalise */
			(GClassInitFunc)screem_editor_class_init,
			NULL, /* class finalise */
			NULL, /* class data */
			sizeof( ScreemEditor ),
			0, /* n_preallocs */
			(GInstanceInitFunc)screem_editor_init
		};
		
		type = g_type_register_static( PARENT_TYPE,
					       "ScreemEditor",
					       &info, 0 );
	}

	return type;
}
