/*  Screem:  screem-page.c
 *
 *  The ScreemPage object
 *
 *  Copyright (C) 2001 David A Knight
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  For contact information with the author of this source code please see
 *  the AUTHORS file.  If there is no AUTHORS file present then check the
 *  about box under the help menu for a contact address
 */

#include <config.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <pcreposix.h>

#include <libgnomevfs/gnome-vfs-mime-handlers.h>
#include <libgnomevfs/gnome-vfs-mime-utils.h>
#include <libgnomevfs/gnome-vfs-uri.h>
#include <libgnomevfs/gnome-vfs-ops.h>
#include <libgnomevfs/gnome-vfs-utils.h>

#include <gconf/gconf-client.h>

#include <glib/gconvert.h>
#include <glib/gunicode.h>

#include <gtk/gtktreemodel.h>
#include <gtk/gtkliststore.h>

#include "screem-page.h"

#include "fileops.h"

/* yuk */
#include "pageUI.h"

/* yuk */
#include "screem-markup.h"

/* yuk */
#include "editMenu.h"

enum {
	CHANGED,
	DOCTYPE_CHANGED,
	CHARSET_CHANGED,
	SAVED,
	LAST_SIGNAL
};

static guint screem_page_signals[ LAST_SIGNAL ] = { 0 };


static void screem_page_class_init( ScreemPageClass *klass );
static void screem_page_init( ScreemPage *page );
static void screem_page_finalize( GObject *object );

static gboolean bookmark_store_find( GtkTreeModel *model, GtkTreePath *path,
				     GtkTreeIter *iter, gpointer data );

static void changed( GtkTextBuffer *buffer, ScreemPage *page );
static void select_syntax( gpointer key, gpointer value, gpointer user_data );

static void insert_text( GtkTextBuffer *buffer, GtkTextIter *it,
			 const gchar *text, gint length, ScreemPage *page );

static void delete_range_before( GtkTextBuffer *buffer, GtkTextIter *it,
				 GtkTextIter *eit, ScreemPage *page );
static void delete_range( GtkTextBuffer *buffer, GtkTextIter *it,
			  GtkTextIter *eit, ScreemPage *page );

static gint hi_trigger( ScreemPage *page, GtkTextIter *it, GtkTextIter *eit );
static inline GtkTextTag *hi_trigger_fine_match( GList *tags, const gchar *str,
						 gint *pat_len,
						 GtkTextTag **ftag,
						 GtkTextTag **btag );

static void screem_page_tag_added( GtkTextTagTable *table, GtkTextTag *tag,
				   ScreemPage *page );
static void screem_page_tag_removed( GtkTextTagTable *table, GtkTextTag *tag,
				     ScreemPage *page );
static void screem_page_tag_changed( GtkTextTagTable *table, GtkTextTag *tag,
				     gboolean size_changed,
				     ScreemPage *page );

static void screem_page_monitor_cb( GnomeVFSMonitorHandle *handle,
				    const gchar *monitor_uri,
				    const gchar *info_uri,
				    GnomeVFSMonitorEventType type,
				    gpointer data );

static void screem_page_doctype_check( ScreemPage *page, gint pos, 
				       gint length );
static void screem_page_charset_check( ScreemPage *page, gint pos, 
				       gint length );

static void get_syntax_types( gpointer key, gpointer value, 
			      gpointer user_data );

struct ScreemPagePrivate {
	gchar *pathname;     /* pathname for the page */
	gint tab;            /* the tab number of the page if open */
	const gchar *mime_type;
	gboolean changed;

	GtkTextBuffer *buffer; /* the actual page data stored in
				  a GtkTextBuffer */
	GtkTextTagTable *table;

	GHashTable *tag_tables;

	gboolean loaded;
	gboolean opened;

	GConfClient *client;
	gboolean highlight;  /* do we think highlighting is on? */
	gint highlight_notify;

	GHashTable *bookmarks;
	GtkListStore *bookmarkstore;

	GSList *undo;
	GSList *redo;
	gboolean undoing;

	GnomeVFSMonitorHandle *monitor;
	gboolean modified;

	gint doctype_pos;
	gchar *doctype;

	gint charset_pos;
	gint charset_len;
	gchar *charset;
};

typedef struct ScreemPageUndo {
	gboolean insert;
	gchar *text;
	gint pos;
} ScreemPageUndo;

static ScreemPageUndo* screem_page_undo_new( gboolean insert, 
					     const gchar *text, gint pos )
{
	ScreemPageUndo *undo;

	/* no need to waste time doing g_new0 as we set all members */
	undo = g_new( ScreemPageUndo, 1 );
	undo->insert = insert;
	undo->text = g_strdup( text );
	undo->pos = pos;

	return undo;
}

static void screem_page_undo_destroy( ScreemPageUndo *undo )
{
	g_free( undo->text );
	g_free( undo );
}

static void screem_page_clear_undo_redo_list( ScreemPage *page,
					      gboolean redo )
{
	GSList *list;

	g_return_if_fail( SCREEM_IS_PAGE( page ) );

	if( redo ) {
		list = page->private->redo;
		page->private->redo = NULL;
	} else {
		list = page->private->undo;
		page->private->undo = NULL;
	}

	g_slist_foreach( list, (GFunc)screem_page_undo_destroy, NULL );
	g_slist_free( list );
}

void screem_page_undo( ScreemPage *page )
{
	ScreemPageUndo *undo;
	GSList *list;

	g_return_if_fail( SCREEM_IS_PAGE( page ) );

	list = page->private->undo;
	if( list ) {
		GtkTextBuffer *buffer;
		GtkTextIter it;
		GtkTextIter eit;
		undo = (ScreemPageUndo*)list->data;
		page->private->undoing = TRUE;

		buffer = screem_page_get_text_buffer( page );
		gtk_text_buffer_get_iter_at_offset( buffer, &it, undo->pos );
		gtk_text_buffer_get_iter_at_offset( buffer, &eit, 
						    undo->pos -
						    strlen( undo->text ) );
		if( ! undo->insert ) {
			gtk_text_buffer_insert( buffer, &it, undo->text,
						strlen( undo->text ) );
		} else {
			gtk_text_buffer_delete( buffer, &eit, &it );
		}
		page->private->undoing = FALSE;
		page->private->undo = g_slist_remove( page->private->undo,
						      undo );
		
		page->private->redo = g_slist_prepend( page->private->redo,
						       undo );
	}
}

void screem_page_redo( ScreemPage *page )
{
	ScreemPageUndo *undo;
	GSList *list;

	g_return_if_fail( SCREEM_IS_PAGE( page ) );

	list = page->private->redo;
	if( list ) {
		GtkTextBuffer *buffer;
		GtkTextIter it;
		GtkTextIter eit;
		gint pos;

		undo = (ScreemPageUndo*)list->data;

		pos = undo->pos;

		buffer = screem_page_get_text_buffer( page );

		page->private->undoing = TRUE;

		if( undo->insert ) {
			if( pos > 0 ) {
				pos --;
			}
			gtk_text_buffer_get_iter_at_offset( buffer, &it, pos );
			gtk_text_buffer_get_iter_at_offset( buffer, &eit, 
							    pos -
							    strlen( undo->text ) );
			gtk_text_buffer_insert( buffer, &it, undo->text,
						strlen( undo->text ) );
		} else {
			gtk_text_buffer_get_iter_at_offset( buffer, &it, pos );
			gtk_text_buffer_get_iter_at_offset(buffer, &eit, 
							   pos +
							   strlen(undo->text));
			gtk_text_buffer_delete( buffer, &it, &eit );
		}
		page->private->undoing = FALSE;

		page->private->redo = g_slist_remove( page->private->redo,
						      undo );
		page->private->undo = g_slist_prepend( page->private->undo,
						       undo );
	}
}

GSList *screem_page_get_mime_types( ScreemPage *page )
{
	GSList *ret;

	g_return_val_if_fail( SCREEM_IS_PAGE( page ), NULL );

	ret = NULL;

	if( page->private->tag_tables ) {
		g_hash_table_foreach( page->private->tag_tables,
				      get_syntax_types, &ret );
	}

	return ret;
}

static void get_syntax_types( gpointer key, gpointer value, 
			      gpointer user_data )
{
	GSList **list;
	GSList *tmp;
	GtkTextTagTable *table;

	list = (GSList**)user_data;

	table = GTK_TEXT_TAG_TABLE( value );
	tmp = (GSList*)g_object_get_data( G_OBJECT( table ), "types" );

	for( ; tmp; tmp = tmp->next ) {
		*list = g_slist_prepend( *list, tmp->data );
	}
}

ScreemPage *screem_page_new( void )
{
	ScreemPage *page;
	GType type;

	type = screem_page_get_type();

	page = SCREEM_PAGE( g_object_new( type, NULL ) );

	return page;
}

gboolean screem_page_is_loaded( ScreemPage *page )
{
	g_return_val_if_fail( SCREEM_IS_PAGE( page ), FALSE );

	return page->private->loaded;
}

gboolean screem_page_is_open( ScreemPage *page )
{
	g_return_val_if_fail( SCREEM_IS_PAGE( page ), FALSE );

	return page->private->opened;
}

void screem_page_set_open( ScreemPage *page, gboolean opened )
{
	g_return_if_fail( SCREEM_IS_PAGE( page ) );

	page->private->opened = opened;
}


void screem_page_set_pathname( ScreemPage *page, const gchar *path )
{
	gchar *data;
	gboolean loaded;

	g_return_if_fail( SCREEM_IS_PAGE( page ) );

	if( page->private->pathname && path != page->private->pathname ) {
		g_free( page->private->pathname );
	}

	page->private->mime_type = 
		gnome_vfs_get_mime_type( path );

	if( path != page->private->pathname ) {
		if( path ) {
			page->private->pathname = g_strdup( path );
		} else {
			page->private->pathname = NULL;
		}
	}

	page->private->table = NULL;
	if( page->private->mime_type &&
	    page->private->tag_tables ) {
		g_hash_table_foreach( page->private->tag_tables,
				      select_syntax, page );
	}

	data = screem_page_get_data( page );
	g_object_unref( page->private->buffer );

	page->private->buffer = gtk_text_buffer_new( page->private->table );

	loaded = screem_page_is_loaded( page );
	screem_page_set_data( page, data );
	page->private->loaded = loaded;

	g_free( data );

	g_signal_connect( G_OBJECT( page->private->buffer ),
			  "changed", G_CALLBACK( changed ), page );
	g_signal_connect_after( G_OBJECT( page->private->buffer ),
				"insert_text", 
				G_CALLBACK( insert_text ),
				page );
	g_signal_connect( G_OBJECT( page->private->buffer ),
			  "delete_range",
			  G_CALLBACK( delete_range_before ),
			  page );
	g_signal_connect_after( G_OBJECT( page->private->buffer ),
				"delete_range",
				G_CALLBACK( delete_range ),
				page );

	if( page->private->monitor ) {
		gnome_vfs_monitor_cancel( page->private->monitor );
		page->private->monitor = NULL;
	}
	if( page->private->pathname ) {
		gnome_vfs_monitor_add( &page->private->monitor,
				       page->private->pathname,
				       GNOME_VFS_MONITOR_FILE,
				       screem_page_monitor_cb, page );
	}
}

const gchar *screem_page_get_pathname( ScreemPage *page )
{
	g_return_val_if_fail( SCREEM_IS_PAGE( page ), NULL );

	return page->private->pathname;
}

void screem_page_set_mime_type( ScreemPage *page, const gchar *mime_type )
{
	gboolean loaded;
	gchar *data;

	g_return_if_fail( SCREEM_IS_PAGE( page ) );
	
	if( ! mime_type ) {
		mime_type = "text/plain";
	}

	page->private->mime_type = mime_type;
	page->private->table = NULL;
	if( page->private->tag_tables ) {
		g_hash_table_foreach( page->private->tag_tables,
				      select_syntax, page );
	}

	data = screem_page_get_data( page );
	g_object_unref( page->private->buffer );

	page->private->buffer = gtk_text_buffer_new( page->private->table );

	loaded = screem_page_is_loaded( page );
	screem_page_set_data( page, data );
	page->private->loaded = loaded;

	g_free( data );
	
	g_signal_connect( G_OBJECT( page->private->buffer ),
			  "changed", G_CALLBACK( changed ), page );
	g_signal_connect_after( G_OBJECT( page->private->buffer ),
				"insert_text", 
				G_CALLBACK( insert_text ),
				page );
	g_signal_connect( G_OBJECT( page->private->buffer ),
			  "delete_range",
			  G_CALLBACK( delete_range_before ),
			  page );
	g_signal_connect_after( G_OBJECT( page->private->buffer ),
				"delete_range",
				G_CALLBACK( delete_range ),
				page );
	hi_trigger( page, NULL, NULL );
}

const gchar *screem_page_get_mime_type( ScreemPage *page )
{
	g_return_val_if_fail( SCREEM_IS_PAGE( page ), NULL );

	return page->private->mime_type;
}

const gchar *screem_page_get_doctype( ScreemPage *page )
{
	g_return_val_if_fail( SCREEM_IS_PAGE( page ), NULL );

	return page->private->doctype;
}


void screem_page_set_charset( ScreemPage *page, const gchar *charset )
{
	g_return_if_fail( SCREEM_IS_PAGE( page ) );

	if( page->private->charset ) {
		g_free( page->private->charset );
	}
	page->private->charset = g_strdup( charset );
}

const gchar *screem_page_get_charset( ScreemPage *page )
{
	g_return_val_if_fail( SCREEM_IS_PAGE( page ), NULL );

	return page->private->charset;
}

void screem_page_set_data( ScreemPage *page, const gchar *data )
{
	const gchar *end;
	gsize len;

	g_return_if_fail( SCREEM_IS_PAGE( page ) );

	page->private->undoing = TRUE;

	/* ensure the text is UTF8, otherwise GtkTextBuffer can't
	   hold it */

	if( data && ! g_utf8_validate( data, strlen( data ), &end ) ) {
		gchar *temp = NULL;
		const gchar *charset;
		gboolean utf8;
		const char *encodings_to_try[ 2 ] = { 0, 0 };

		utf8 = g_get_charset( &charset );

		if( ! utf8 ) {
			encodings_to_try[ 0 ] = charset;
		}
		if( g_ascii_strcasecmp( charset, "ISO-8859-1" ) ) {
			encodings_to_try[ 1 ] = "ISO-8859-1";
		}

		if( encodings_to_try[ 0 ] ) {
			temp = g_convert( data, strlen( data ), "UTF-8", 
					  encodings_to_try[ 0 ], 
					  NULL, &len, NULL );
		}
		if( ! temp && encodings_to_try[ 1 ] ) {
			temp = g_convert( data, strlen( data ), "UTF-8", 
					  encodings_to_try[ 1 ], 
					  NULL, &len, NULL );
		}

		if( temp ) {
			data = temp;
		}
	}

	if( data ) {
		gtk_text_buffer_set_text( page->private->buffer, data,
					  strlen( data ) );
	}

	page->private->undoing = FALSE;

	page->private->loaded = TRUE;
}

gchar *screem_page_get_data( ScreemPage *page )
{
	GtkTextIter it;
	GtkTextIter eit;

	g_return_val_if_fail( SCREEM_IS_PAGE( page ), NULL );

	gtk_text_buffer_get_start_iter( page->private->buffer, &it );
	gtk_text_buffer_get_end_iter( page->private->buffer, &eit );

	return gtk_text_buffer_get_text( page->private->buffer, &it, &eit,
					 TRUE );
}

void screem_page_set_changed( ScreemPage *page, gboolean flag )
{
	g_return_if_fail( SCREEM_IS_PAGE( page ) );

	g_signal_emit( G_OBJECT( page ), 
		       screem_page_signals[ CHANGED ],
		       0, flag );

	page->private->changed = flag;
}

gboolean screem_page_get_changed( ScreemPage *page )
{
	g_return_val_if_fail( SCREEM_IS_PAGE( page ), FALSE );

	return page->private->changed;
}

gboolean screem_page_load( ScreemPage *page )
{
	const gchar *path;

	GString *data;

        g_return_val_if_fail( SCREEM_IS_PAGE( page ), FALSE );

	if( screem_page_is_loaded( page ) ) {
		return TRUE;
	}

	path = screem_page_get_pathname( page );

	data = load_file( path );

	if( data ) {
		screem_page_set_data( page, data->str );
		g_string_free( data, TRUE );

		screem_page_set_changed( page, FALSE );

		page->private->loaded = TRUE;
	}

	page->private->modified = FALSE;

	return ( data != NULL );
}

gboolean screem_page_revert( ScreemPage *page )
{
	g_return_val_if_fail( SCREEM_IS_PAGE( page ), FALSE );

	page->private->loaded = FALSE;

	return screem_page_load( page );
}

gboolean screem_page_save( ScreemPage *page )
{
	size_t size;

	gchar *data;
	gboolean retval;

	gchar *charset;

	g_return_val_if_fail( SCREEM_IS_PAGE( page ), FALSE );

	if( ( ! screem_page_get_changed( page ) ) ||
	    ( ! screem_page_save_check( page ) ) ) {
		return TRUE;
	}

	data = screem_page_get_data( page );

	size = strlen( data );

	if( page->private->charset ) {
		charset = g_strdup( page->private->charset );
	} else {
		GConfClient *client;

		client = gconf_client_get_default();

		charset = gconf_client_get_string( client,
						   "/apps/screem/editor/default_charset",
						   NULL );
		g_object_unref( G_OBJECT( client ) );
	}

	if( g_strcasecmp( "UTF-8", charset ) ) {
		/* need to convert from UTF-8 to charset */
		gchar *temp;
		gint len;
		
		temp = g_convert( data, size, charset, 
				  "UTF-8", 
				  NULL, &len, NULL );

		if( temp ) {
			g_free( data );
			size = strlen( temp );
			data = temp;
		} else {
			/* failed to convert to the requested charset */
			g_free( data );
			return FALSE;
		}
	}

	g_free( charset );

	/* stop monitoring while we write to the file ourselves */
	if( page->private->monitor ) {
		gnome_vfs_monitor_cancel( page->private->monitor );
	}
	page->private->monitor = NULL;

	retval = save_file( screem_page_get_pathname( page ),
			    data, 
			    GNOME_VFS_PERM_USER_READ | 
			    GNOME_VFS_PERM_USER_WRITE |
			    GNOME_VFS_PERM_GROUP_READ |
			    GNOME_VFS_PERM_GROUP_WRITE |
			    GNOME_VFS_PERM_OTHER_READ );

	g_free( data );

	if( retval ) {
		g_signal_emit( G_OBJECT( page ),
				       screem_page_signals[ SAVED ], 0 );
	}
	
	screem_page_set_changed( page, FALSE );
	page->private->modified = FALSE;

      	if( ! screem_page_get_mime_type( page ) ) {
		screem_page_set_pathname( page, 
					  screem_page_get_pathname( page ) );
	} else {
		gnome_vfs_monitor_add( &page->private->monitor,
				       page->private->pathname,
				       GNOME_VFS_MONITOR_FILE,
				       screem_page_monitor_cb, page );
	}
	return retval;
}

gboolean screem_page_is_file_page( const gchar *path )
{
	const gchar *mime_type;

	g_return_val_if_fail( path != NULL, FALSE );
	

	mime_type = gnome_vfs_get_mime_type( path );

	return screem_page_is_mime_type_page( mime_type );
}

gboolean screem_page_is_mime_type_page( const gchar *mime_type )
{
	gboolean is_page;

	if( ! mime_type ) {
		is_page = FALSE;
	} else {
		is_page = ! strncmp( mime_type, "text/", strlen( "text/" ) );
		is_page |= ! strcmp( mime_type, "application/x-php" );
		is_page |= ! strcmp( mime_type, "application/x-asp" );
		is_page |= ! strcmp( mime_type, "application/x-cgi" );
	}
	
	return is_page;
}


gboolean screem_page_get_modified( ScreemPage *page )
{
	g_return_val_if_fail( SCREEM_IS_PAGE( page ), FALSE );

	return page->private->modified;
}

GtkTextBuffer *screem_page_get_text_buffer( ScreemPage *page )
{
	g_return_val_if_fail( page != NULL, NULL );

	return page->private->buffer;
}

GtkListStore *screem_page_get_bookmark_model( ScreemPage *page )
{
	g_return_val_if_fail( SCREEM_IS_PAGE( page ), NULL );

	return page->private->bookmarkstore;
}

void screem_page_add_bookmark( ScreemPage *page, gint line, const gchar *name )
{
	ScreemPagePrivate *private;
	ScreemPageBookMark *bookmark;
	GHashTable *table;
	GtkTreeIter it;

	g_return_if_fail( SCREEM_IS_PAGE( page ) );

	private = page->private;

	table = private->bookmarks;

	bookmark = g_hash_table_lookup( table, name );
	if( bookmark ) {
		/* already got a bookmark with this name, just update
		   the line */
		g_free( bookmark->line_number );
		bookmark->line_number = g_strdup_printf( "%d", line );

		gtk_tree_model_foreach( GTK_TREE_MODEL(private->bookmarkstore),
					bookmark_store_find, bookmark );	
		it = *(GtkTreeIter*)bookmark->iter;
	} else {
		bookmark = g_new0( ScreemPageBookMark, 1 );
		bookmark->line_number = g_strdup_printf( "%d", line );
		bookmark->name = g_string_new( name );

		gtk_list_store_append( private->bookmarkstore, &it );

		g_hash_table_insert( table, bookmark->name->str, bookmark );
	}

	gtk_list_store_set( private->bookmarkstore, &it,
			    0, bookmark->name->str,
			    1, bookmark->line_number,
			    2, bookmark, -1 );
	
}

void screem_page_remove_bookmark( ScreemPage *page, const gchar *name )
{
	ScreemPagePrivate *private;
	ScreemPageBookMark *bookmark;
	GHashTable *table;

	g_return_if_fail( SCREEM_IS_PAGE( page ) );

	private = page->private;

	table = private->bookmarks;

	bookmark = g_hash_table_lookup( table, name );
	if( bookmark ) {
		/* found the bookmark we are after */
		GtkTreeIter it;

		g_hash_table_remove( table, name );

		gtk_tree_model_foreach( GTK_TREE_MODEL(private->bookmarkstore),
					bookmark_store_find, bookmark );
		it = *(GtkTreeIter*)bookmark->iter;
		gtk_list_store_remove( private->bookmarkstore,
				       &it );

		g_string_free( bookmark->name, TRUE );
		g_free( bookmark );
	}
}

gint screem_page_get_bookmark( ScreemPage *page, const gchar *name )
{
	ScreemPagePrivate *private;
	ScreemPageBookMark *bookmark;
	GHashTable *table;

	g_return_val_if_fail( SCREEM_IS_PAGE( page ), -1 );

	private = page->private;

	table = private->bookmarks;

	bookmark = g_hash_table_lookup( table, name );
	if( bookmark ) {
		return atoi( bookmark->line_number );
	} else {
		return -1;
	}
}

static gboolean bookmark_scan( gpointer key, gpointer value, gpointer data )
{
	ScreemPageBookMark **bookmark;
	ScreemPageBookMark *current;

	bookmark = (ScreemPageBookMark**)data;

	current = (ScreemPageBookMark*)value;

	if( atoi(current->line_number) == atoi( (*bookmark)->line_number ) ) {
		*bookmark = current;

		return TRUE;
	}

	return FALSE;
}

static gboolean screem_page_clear_bookmarks( gpointer key, gpointer value, 
					     gpointer data )
{
	screem_page_remove_bookmark( SCREEM_PAGE( data ), key );

	return FALSE;
}

const gchar* screem_page_bookmarked( ScreemPage *page, gint line )
{
 	ScreemPagePrivate *private;
	ScreemPageBookMark bm;
	ScreemPageBookMark *bookmark;
	GHashTable *table;

	g_return_val_if_fail( SCREEM_IS_PAGE( page ), NULL );

	private = page->private;

	table = private->bookmarks;

	bookmark = &bm;
	bm.name = NULL;
	bm.line_number = g_strdup_printf( "%d", line );
	g_hash_table_foreach( table, (GHFunc)bookmark_scan,
			      (gpointer)&bookmark );
	g_free( bm.line_number );

	if( bookmark->name ) {
		return bookmark->name->str;
	}

	return NULL;
}

static gboolean bookmark_store_find( GtkTreeModel *model, GtkTreePath *path,
				     GtkTreeIter *iter, gpointer data )
{
	ScreemPageBookMark *bookmark;
	ScreemPageBookMark *bookmark2;
     	GValue value = {0,};

	bookmark = (ScreemPageBookMark*)data;
	
	gtk_tree_model_get_value( GTK_TREE_MODEL( model ),
				  iter, 2, &value );
	bookmark2 = g_value_get_pointer( &value );
	
	bookmark->iter = NULL;
	if( !strcmp( bookmark->name->str, bookmark2->name->str ) )
		bookmark->iter = iter;

	return ( bookmark->iter != NULL );
}

static void changed( GtkTextBuffer *buffer, ScreemPage *page )
{
	g_return_if_fail( SCREEM_IS_PAGE( page ) );

	screem_page_set_changed( page, TRUE );
}

static void insert_text( GtkTextBuffer *buffer, GtkTextIter *it,
			 const gchar *text, gint length, ScreemPage *page )
{
	GtkTextIter nit = *it;
	GtkTextIter neit;
	GtkTextTagTable *table;

	gint pos;

	pos = gtk_text_iter_get_offset( it );

	table = gtk_text_buffer_get_tag_table( buffer );
		
	gtk_text_buffer_get_iter_at_offset( buffer,
					    &neit, pos + length );

	screem_page_doctype_check( page, pos, length );
	screem_page_charset_check( page, pos, length );
	
	/* undo stuff */
	if( ! page->private->undoing )
	{
		ScreemPageUndo *undo;
		undo = screem_page_undo_new( TRUE, text, pos );
		page->private->undo = g_slist_prepend( page->private->undo,
						       undo );

		/* clear redo list */
		screem_page_clear_undo_redo_list( page, TRUE );
	}

	/* search back to previous tag start */
	if( gtk_text_tag_table_get_size( table ) ) {
		gtk_text_iter_backward_char( &nit );
		while( ! gtk_text_iter_is_start( &nit ) ) {
			if( gtk_text_iter_begins_tag( &nit, NULL ) ) {
				/* we are only interested if we aren't in a sub tag */
				GSList *tags;
				GSList *tmp;
				tags = gtk_text_iter_get_tags( &nit );
				for( tmp = tags; tmp; tmp = tmp->next ) {
					if( g_object_get_data( G_OBJECT( tmp->data ),
							       "sub" ) ) {
						break;
					}
				}
				g_slist_free( tags );
				if( ! tmp ) {
					break;
				}
			}
			gtk_text_iter_backward_char( &nit );
		}
		
		/* search to start of next tag */
		gtk_text_iter_forward_char( &neit );
		while( ! gtk_text_iter_is_end( &neit ) ) {
			if( gtk_text_iter_begins_tag( &neit, NULL ) ) {
				/* we are only interested if we aren't in a sub tag */
				GSList *tags;
				GSList *tmp;
				tags = gtk_text_iter_get_tags( &neit );
				for( tmp = tags; tmp; tmp = tmp->next ) {
					if( g_object_get_data( G_OBJECT( tmp->data ),
							       "sub" ) ) {
						break;
					}
				}
				g_slist_free( tags );
				if( ! tmp ) {
					break;
				}
			}
		gtk_text_iter_forward_char( &neit );
		}
	}

	
	hi_trigger( page, &nit, &neit );
}

static void delete_range_before( GtkTextBuffer *buffer, GtkTextIter *it,
				 GtkTextIter *eit, ScreemPage *page )
{
	if( ! page->private->undoing )
	{
		ScreemPageUndo *undo;
		gint pos;
		gchar *text;
		
		text = gtk_text_buffer_get_text( buffer, it, eit, TRUE );
		pos = gtk_text_iter_get_offset( it );
		undo = screem_page_undo_new( FALSE, text, pos );
		page->private->undo = g_slist_prepend( page->private->undo,
						       undo );
		g_free( text );

		/* clear redo list */
		screem_page_clear_undo_redo_list( page, TRUE );
	}
}

static void delete_range( GtkTextBuffer *buffer, GtkTextIter *it,
			  GtkTextIter *eit, ScreemPage *page )
{
	GtkTextIter nit = *it;
	GtkTextIter neit = *eit;

	gint pos;
	gint length;

	pos = gtk_text_iter_get_offset( it );
	length = gtk_text_iter_get_offset( eit );
	length = length - pos;

	screem_page_doctype_check( page, pos, length );
	screem_page_charset_check( page, pos, length );
	
	/* search back to previous tag start */
	if( gtk_text_iter_get_tags( it ) || gtk_text_iter_get_tags( eit ) ) {
		gtk_text_iter_backward_char( &nit );
		while( ! gtk_text_iter_is_start( &nit ) ) {
			if( gtk_text_iter_begins_tag( &nit, NULL ) ) {
				/* we are only interested if we aren't in a sub tag */
				GSList *tags;
				GSList *tmp;
				tags = gtk_text_iter_get_tags( &nit );
				for( tmp = tags; tmp; tmp = tmp->next ) {
					if( g_object_get_data( G_OBJECT( tmp->data ),
							       "sub" ) ) {
						break;
					}
				}
				g_slist_free( tags );
				if( ! tmp ) {
					break;
				}
			}
			gtk_text_iter_backward_char( &nit );
		}
		
		/* search to start of next tag */
		gtk_text_iter_forward_char( &neit );
		while( ! gtk_text_iter_is_end( &neit ) ) {
			if( gtk_text_iter_begins_tag( &neit, NULL ) ) {
				/* we are only interested if we aren't in a sub tag */
				GSList *tags;
				GSList *tmp;
				tags = gtk_text_iter_get_tags( &neit );
				for( tmp = tags; tmp; tmp = tmp->next ) {
					if( g_object_get_data( G_OBJECT( tmp->data ),
							       "sub" ) ) {
						break;
					}
				}
				g_slist_free( tags );
				if( ! tmp ) {
					break;
				}
		}
			gtk_text_iter_forward_char( &neit );
		}
		hi_trigger( page, &nit, &neit );
	}
}

static void select_syntax( gpointer key, gpointer value, gpointer user_data )
{
	GtkTextTagTable *table;
	GList *mime_types;
	ScreemPage *page;
	const gchar *mime_type;

	page = SCREEM_PAGE( user_data );
	table = GTK_TEXT_TAG_TABLE( value );

	g_assert( GTK_IS_TEXT_TAG_TABLE( table ) );

	g_return_if_fail( page != NULL );

	mime_type = screem_page_get_mime_type( page );

	if( ( ! page->private->table ) && mime_type ) {
		mime_types = (GList*)g_object_get_data( G_OBJECT( table ), "types" );
		
		while( mime_types ) {
			const gchar *type = (const gchar*)mime_types->data;
			
			g_assert( type );
			
			if( ! strcmp( mime_type, type ) ) {
				/* found the syntax table */
				page->private->table = table;
				
				/* we need to know when tags are added / removed / 
				   changed,
				   which is done in the preferences dialog */
				g_signal_connect_after( G_OBJECT( table ), "tag_added",
							G_CALLBACK(screem_page_tag_added),
							page );
				g_signal_connect( G_OBJECT( table ), "tag_removed",
						  G_CALLBACK( screem_page_tag_removed),
						  page );
				g_signal_connect_after( G_OBJECT( table ),
							"tag_changed",
							G_CALLBACK( screem_page_tag_changed ),
							page );
				break;
			}
			mime_types = mime_types->next;
		}
	}
}

static gint hi_trigger( ScreemPage *page, GtkTextIter *it, GtkTextIter *eit )
{
	GtkTextBuffer *buffer;
	GtkTextTagTable *table;
	GtkTextTag *tag;
	GtkTextIter hit;
	GtkTextIter heit;

	gchar *text;
	const gchar *end;
	const gchar *end_chunk;
	gint pat_len;
	gchar *str;
	gint pos;
	gint spos;
	gint epos;

	size_t nmatch = 1;
	regmatch_t pmatch[ 1 ];

	regex_t *sall;
	GList *table_tags;

	GtkTextTag *ftag;
	GtkTextTag *btag;
	GtkTextTag *sub;

	regex_t *regex;
	GtkTextIter nit;
	GtkTextIter neit;
	GList *subs;

	g_return_val_if_fail( SCREEM_IS_PAGE( page ), TRUE );

	if( ! page->private->highlight ) {
		return TRUE;
	}

	buffer = screem_page_get_text_buffer( page );
	table = gtk_text_buffer_get_tag_table( buffer );
	sall = (regex_t*)g_object_get_data( G_OBJECT( table ), 
					    "start_patterns" );

	if( ! sall ) {
		return TRUE;
	}

	if( ! it ) {
		gtk_text_buffer_get_start_iter( buffer, &hit );
		it = &hit;
	}
	if( ! eit ) {
		gtk_text_buffer_get_end_iter( buffer, &heit );
		eit = &heit;
	}

	text = gtk_text_buffer_get_text( buffer, it, eit, TRUE );
	pos = 0;
	spos = gtk_text_iter_get_offset( it );
	epos = gtk_text_iter_get_offset( eit );

	end = text + epos - spos;

	end_chunk = text;

	gtk_text_buffer_remove_all_tags( buffer, it, eit );

	table_tags = (GList*)g_object_get_data( G_OBJECT( table ),
						"tags" );

	do {
		/* we are done if there are no matches left */
		if( regexec( sall, end_chunk, nmatch, pmatch, 0 ) ) {
			break;
		}
		
		/* calculate range */
		end_chunk += pmatch[ 0 ].rm_so;
		pat_len = pmatch[ 0 ].rm_eo - pmatch[ 0 ].rm_so;

		/* we have a match for one of the expressions,
		   from end_chunk to end_chunk + pat_len */
		str = g_strndup( end_chunk, pat_len );

		/* find finer match */
		tag = hi_trigger_fine_match( table_tags, str,
					     &pat_len,
					     &ftag, &btag );
		g_free( str );
		pos = end_chunk - text;

		end_chunk += pat_len;

		gtk_text_buffer_get_iter_at_offset( buffer, &nit, 
						    pos + spos );
		
		/* apply sub / end matches */
		if( tag ) {
			regex = g_object_get_data( G_OBJECT( tag ),
						   "sub_patterns" );
		} else {
			regex = NULL;
		}
		if( regex ) {
			GtkTextIter snit;
			GtkTextIter *sit;
			regex_t *subreg;

			subs = g_object_get_data( G_OBJECT( tag ),
						  "subs" );
			do {
			
				if( regexec( regex, end_chunk, nmatch, 
					     pmatch, 0 ) ) {
					break;
				}
				end_chunk += pmatch[ 0 ].rm_so;
				pat_len = pmatch[ 0 ].rm_eo - pmatch[ 0 ].rm_so;
				str = g_strndup( end_chunk, pat_len );
				
				sub = hi_trigger_fine_match( subs, str,
							     &pat_len,
							     &ftag, &btag );
				g_free( str );
				if( ! sub ) {
					break;
				}

				/* move match_start to character after
				   the matched expression */
				pos = end_chunk - text;
				
				end_chunk += pat_len;
				
				gtk_text_buffer_get_iter_at_offset( buffer, 
								    &snit, 
								    pos + 
								    spos );
				
				/* locate end pattern if one exists for
				   this tag */
				subreg = g_object_get_data( G_OBJECT( sub ),
							    "end" );
				
				sit = &snit;
				if( sub == tag ) {
					sit = &nit;
				} else if( subreg ) {
					/* match regex against text
					   to find end of
					   region */
					if( ! regexec( subreg, end_chunk,
						       nmatch, 
						       pmatch, 0 ) ) {
						/* calculate range */
						pat_len = pmatch[ 0 ].rm_eo;

						end_chunk += pat_len;
						pat_len += pmatch[ 0 ].rm_eo -
							pmatch[ 0 ].rm_so;
					} else {
						pat_len += end - 
							end_chunk;
						end_chunk = end;
					}
				}
				
				/* get iter for end of start match,
				   or end of end match if one
				   exists */
				gtk_text_buffer_get_iter_at_offset( buffer, 
								    &neit,
								    pos +
								    pat_len +
								    spos);
				if( ftag ) {
					gtk_text_buffer_apply_tag( buffer, ftag, 
								   sit, &neit );
				}
				if( btag ) {
					gtk_text_buffer_apply_tag( buffer, btag,
								   sit, &neit );
				}
			} while( end_chunk != end && sub != tag );
		} else {
			/* not a block highlight, won't have subs */
			gtk_text_buffer_get_iter_at_offset( buffer, &neit,
							    pos + pat_len +
							    spos);
			if( ftag ) {
				gtk_text_buffer_apply_tag( buffer, ftag, 
							   &nit, &neit );
			}
			if( btag ) {
				gtk_text_buffer_apply_tag( buffer, btag,
							   &nit, &neit );
			}
		}
	} while( end_chunk < end );

	g_free( text );

	return TRUE;
}

static inline GtkTextTag *hi_trigger_fine_match( GList *tags, const gchar *str,
						 gint *pat_len,
						 GtkTextTag **ftag,
						 GtkTextTag **btag )
{
	regex_t *regex;
	GtkTextTag *tag;
	GList *tmp;
	size_t nmatch = 1;
	regmatch_t pmatch[ 1 ];

	*ftag = NULL;
	*btag = NULL;

	/* match str against the correct regexp,
	   we start at tags->next to avoid matching against the start
	   expression we have already matched, non sub matches the
	   first item in the list is NULL to account for this*/
	tmp = tags;
       	for( tags = tags->next, tag = NULL; tags; tags = tags->next ) {
		tag = GTK_TEXT_TAG( tags->data );

		if( tag ) {
			regex = (regex_t*)
				g_object_get_data( G_OBJECT( tag ),"start" );
		} else {
			regex = NULL;
		}
		if( ( regex && ! regexec( regex, str, nmatch, pmatch, 0 ) ) &&
		    pmatch[ 0 ].rm_so == 0 ) {
			/* is this one in use?, if in use
			   break out of the loop */
			*ftag = GTK_TEXT_TAG( g_object_get_data(G_OBJECT(tag),
								"fore_tag") );
			*btag = GTK_TEXT_TAG( g_object_get_data(G_OBJECT(tag),
								"back_tag") );
			
			if( *ftag && ! g_object_get_data( G_OBJECT( *ftag ),
							 "usefore" ) ) {
				*ftag = NULL;
			}
			if( *btag && ! g_object_get_data( G_OBJECT( *btag ),
							 "useback" ) ) {
				*btag = NULL;
			}
			if( *ftag || *btag ) {
				*pat_len = pmatch[ 0 ].rm_eo - 
					pmatch[ 0 ].rm_so;
				break;
			}
		}
		tag = NULL;
	}
	if( ! tag ) {
		/* possibly match the end pattern, which is always first
		   in the list of sub tags, and this is the only case
		   where tag should be NULL due to a regex on the complete set
		   of start patterns */
		tag = GTK_TEXT_TAG( tmp->data );
		if( tag ) {
			regex = (regex_t*)g_object_get_data( G_OBJECT( tag ),
								"end" );
		} else {
			regex = NULL;
		}
		if( ( regex && ! regexec( regex, str, nmatch, pmatch, 0 ) ) &&
		    pmatch[ 0 ].rm_so == 0 ) {
			/* is this one in use?, if in use
			   break out of the loop */
			*ftag = GTK_TEXT_TAG( g_object_get_data(G_OBJECT(tag),
								"fore_tag") );
			*btag = GTK_TEXT_TAG( g_object_get_data(G_OBJECT(tag),
								"back_tag") );
			
			if( *ftag && ! g_object_get_data( G_OBJECT( *ftag ),
							 "usefore" ) ) {
				*ftag = NULL;
			}
			if( *btag && ! g_object_get_data( G_OBJECT( *btag ),
							 "useback" ) ) {
				*btag = NULL;
			}
			if( *ftag || *btag ) {
				*pat_len = pmatch[ 0 ].rm_eo - 
					pmatch[ 0 ].rm_so;
			} else {
				tag = NULL;
			}
		} else {
			tag = NULL;
		}
	}

	return tag;
}

static void screem_page_tag_added( GtkTextTagTable *table, GtkTextTag *tag,
				   ScreemPage *page )
{
	GtkTextIter it;
	GtkTextIter eit;
	GtkTextBuffer *buffer;

	buffer = screem_page_get_text_buffer( page );

	gtk_text_buffer_get_start_iter( buffer, &it );
	gtk_text_buffer_get_end_iter( buffer, &eit );

	hi_trigger( page, &it, &eit );
}

static void screem_page_tag_removed( GtkTextTagTable *table, GtkTextTag *tag,
				     ScreemPage *page )
{
	GtkTextIter it;
	GtkTextIter eit;
	GtkTextBuffer *buffer;

	buffer = screem_page_get_text_buffer( page );

	gtk_text_buffer_get_start_iter( buffer, &it );
	gtk_text_buffer_get_end_iter( buffer, &eit );

	hi_trigger( page, &it, &eit );
}


static void screem_page_tag_changed( GtkTextTagTable *table, GtkTextTag *tag,
				     gboolean size_changed,
				     ScreemPage *page )
{
	GtkTextIter it;
	GtkTextIter eit;
	GtkTextBuffer *buffer;

	buffer = screem_page_get_text_buffer( page );

	gtk_text_buffer_get_start_iter( buffer, &it );
	gtk_text_buffer_get_end_iter( buffer, &eit );

	hi_trigger( page, &it, &eit );
}

gint screem_page_get_line_count( ScreemPage *page )
{
	GtkTextBuffer *buffer;

	g_return_val_if_fail( SCREEM_IS_PAGE( page ), 0 );

	buffer = screem_page_get_text_buffer( page );

	return gtk_text_buffer_get_line_count( buffer );
}

static void screem_page_highlight_notify( GConfClient *client,
					  guint cnxn_id,
					  GConfEntry *entry,
					  gpointer data )
{
	ScreemPage *page;

	page = SCREEM_PAGE( data );

	if( entry->value->type == GCONF_VALUE_BOOL ) {
		gboolean state = gconf_value_get_bool( entry->value );
		GtkTextIter it;
		GtkTextIter eit;
		GtkTextBuffer *buffer;

		buffer = screem_page_get_text_buffer( page );
		gtk_text_buffer_get_start_iter( buffer, &it );
		gtk_text_buffer_get_end_iter( buffer, &eit );

		if( state && ! page->private->highlight ) {
			/* do highlighting */
			page->private->highlight = state;
			hi_trigger( page, &it, &eit );
		} else if( ! state && page->private->highlight ) {
			/* remove highlighting */
			page->private->highlight = state;
			gtk_text_buffer_remove_all_tags( buffer, &it, &eit );
		}
	}
}

static void screem_page_monitor_cb( GnomeVFSMonitorHandle *handle,
				    const gchar *monitor_uri,
				    const gchar *info_uri,
				    GnomeVFSMonitorEventType type,
				    gpointer data )
{
	ScreemPage *page;

	page = SCREEM_PAGE( data );

	switch( type ) {
	case GNOME_VFS_MONITOR_EVENT_CHANGED:
		page->private->modified = TRUE;
		break;
	case GNOME_VFS_MONITOR_EVENT_DELETED:
		/* FIXME: hmm, should alert the user really */
		break;
	case GNOME_VFS_MONITOR_EVENT_STARTEXECUTING:
		break;
	case GNOME_VFS_MONITOR_EVENT_STOPEXECUTING:
		break;
	case GNOME_VFS_MONITOR_EVENT_CREATED:
		page->private->modified = FALSE;
		break;
	case GNOME_VFS_MONITOR_EVENT_METADATA_CHANGED:
		break;
	default:
		break;
	}
}

static void screem_page_doctype_check( ScreemPage *page, gint pos, 
				       gint length )
{
	ScreemPagePrivate *private;
	gboolean set;
	gboolean getdoctype;

	private = page->private;

	set = FALSE;
	if( private->table ) {
		getdoctype = ( g_object_get_data( G_OBJECT( private->table ),
						  "feature_doctype" ) != NULL);
	} else {
		getdoctype = FALSE;
	}

	/* only do for types that need a doctype */
	if( getdoctype ) {
	   	if( ! private->doctype ) {
			set = TRUE;
		} else { 
			gint len;
			
			len = strlen( private->doctype );
			
			if( ( pos <= ( private->doctype_pos + len ) ) &&
			    ( pos + length ) >= private->doctype_pos ) {
				set = TRUE;
			}
		}
		if( set ) {
			/* set the doctype */
			gchar *text;
			gchar *doctype;
			
			text = screem_page_get_data( page );
			
			doctype = find_text( text,
					     "<![Dd][Oo][Cc][Ty][Yy][Pp][Ee] ",
					     NULL, NULL );
			if( doctype ) {
				GString *tag;
				
				tag = g_string_new( "" );
				
				private->doctype_pos = doctype - text;
				
				doctype --;
				do {
					doctype ++;
					g_string_append_c( tag, *doctype );
				} while( *doctype != '\0' && *doctype != '>' );
				
				if( private->doctype ) {
					g_free( private->doctype );
				}
				private->doctype = tag->str;
				g_string_free( tag, FALSE );
			}
			
			g_free( text );
			
			if( private->doctype ) {
				g_signal_emit( G_OBJECT( page ),
					       screem_page_signals[ DOCTYPE_CHANGED ],
					       0, private->doctype );
			}
		}
	}
}

static void screem_page_charset_check( ScreemPage *page, gint pos, 
				       gint length )
{
	ScreemPagePrivate *private;
	gboolean set = FALSE;
	gboolean getcharset;

	private = page->private;

	if( private->table ) {
		getcharset = ( g_object_get_data( G_OBJECT( private->table ),
						  "feature_charset" ) != NULL);
	} else {
		getcharset = FALSE;
	}

	/* only do for types that need a charset */
	if( ! getcharset ) {
		return;
	}

	if( ! private->charset ) {
		set = TRUE;
	} else { 
		gint len;

		len = private->charset_len;

		if( ( pos <= ( private->charset_pos + len ) ) &&
		    ( pos + length ) >= private->charset_pos ) {
			set = TRUE;
		}
	}
	if( set ) {
		/* set the charset */
		gchar *text;
		gchar *tag;
		gchar *name;
		gint pos;
		gint start;
		gint end;
		gint len;

		pos = 0;
		text = screem_page_get_data( page );
		len = strlen( text );

		g_free( private->charset );
		private->charset = NULL;
		private->charset_pos = 0;

		/* first we look for <?xml encoding="" ?>
		   then we check <meta http-equiv="content-type" />
		   then we check <meta name="charset" /> 
		   if we have an <?xml ?> without an encoding="" we
		   enforce UTF-8 no matter what <meta /> exist */

		while( ( tag = screem_markup_next_tag( text, pos,
						       &start, &end, 
						       &name ) ) ) {
			pos = end;
			if( ! strcmp( "?xml", name ) ) {
				/* xml PI element */
				GSList *list;
				GSList *attr;

				list=screem_markup_build_attributes_list(tag,
									 NULL);
				for( attr = list; attr; attr = attr->next ) {
					if(! strcmp( "encoding", 
						     attr->next->data ) &&
					   attr->data ) {
						private->charset = g_strdup( attr->data );
						break;
					} else {
						attr = NULL;
						break;
					}
					attr = attr->next;
				}
				
				if( ! attr ) {
					private->charset = g_strdup( "UTF-8" );
					private->charset_pos = start;
					private->charset_len = end - start;
				}

				g_slist_foreach( list, (GFunc)g_free, NULL );
				g_slist_free( list );
				pos = len;
			} else if( ! g_strcasecmp( "meta", name ) ) {
				/* meta element */
				GSList *list;
				GSList *attr;
				gboolean content;
				
				list=screem_markup_build_attributes_list(tag,
									 NULL);
				content = FALSE;

				for( attr = list; attr; attr = attr->next ) {
					if( ! strcmp( "http-equiv", 
						      attr->next->data ) && 
					    ! strcmp( "content-type",
						      attr->data ) ) {
						content = TRUE;
						break;
					} else if(! strcmp( "name", 
							    attr->next->data) &&
						  attr->data && 
						  ! strcmp( "charset",
							    attr->data)){
						break;
					}
					attr = attr->next;
				}
				if( attr ) {
					for( attr = list; attr; 
					     attr = attr->next ) {
						if( ! strcmp( "content",
							      attr->next->data ) ) {
							break;
						}
						attr = attr->next;
					}
				}
				if( attr && content ) {
					gchar *val;

					val = strstr( attr->data, "charset=" );
					if( val ) {
						GString *tmp;

						val += strlen( "charset=" );
						tmp = g_string_new( "" );
						while( *val != ' ' &&
						       *val != '\0' &&
						       *val != '"' ) {
							g_string_append_c( tmp,
									   *val);
							val ++;
						}
						private->charset = tmp->str;
						g_string_free( tmp, FALSE );
						private->charset_pos = start;
						private->charset_len = end - 
							start;
						pos =  len;
					}
				} else if( attr ) {
					private->charset = 
						g_strdup( attr->data );
					private->charset_pos = start;
					private->charset_len = end - start;
					pos = len;
				}

				g_slist_foreach( list, (GFunc)g_free, NULL );
				g_slist_free( list );
			} else if( g_strcasecmp( "html", name ) &&
				   g_strcasecmp( "head", name ) &&
				   g_strcasecmp( "title", name ) &&
				   g_strcasecmp( "link", name ) &&
				   g_strcasecmp( "script", name ) &&
				   g_strcasecmp( "style", name ) &&
				   g_strcasecmp( "!--", name ) &&
				   g_strcasecmp( "base", name ) &&
				   g_strcasecmp( "isindex", name ) &&
				   g_strcasecmp( "!DOCTYPE", name ) &&
				   *name != '/' && *name != '?' ) {
				/* limiting case, bit of a hack for html,
				   valid xml docs are supposed to have the
				   <?xml ?> element and have indeterminate 
				   element names so we can't stop if the doc is
				   invalid */
				pos = len;
			}
			g_free( tag );
			g_free( name );
		}
	
		
		g_free( text );

		if( private->charset ) {
			g_signal_emit( G_OBJECT( page ),
				       screem_page_signals[ CHARSET_CHANGED ],
				       0, private->charset );
		}
	}
}

gboolean screem_page_is_dynamic( ScreemPage *page )
{
	gboolean ret;

	g_return_val_if_fail( SCREEM_IS_PAGE( page ), FALSE );

	if( page->private->table ) {
		ret = ( g_object_get_data( G_OBJECT( page->private->table ),
					   "feature_dynamic" ) != NULL);
	} else {
		ret = FALSE;
	}

	return ret;
}

gboolean screem_page_is_template( ScreemPage *page, gchar **tag,
				  gchar **path, GSList **blocks )
{
	gchar *text;
	gint len;
	gchar *name;
	gboolean ret;

	g_return_val_if_fail( SCREEM_IS_PAGE( page ), FALSE );
	g_return_val_if_fail( tag != NULL, FALSE );
	g_return_val_if_fail( path != NULL, FALSE );
	
	text = screem_page_get_data( page );
	name = find_text( text, "<!-- #BeginTemplate[ \t]+\"[^\"]*\"[ \t]*-->",
			  NULL, &len );
	ret = ( name != NULL );
	if( ret ) {
		const gchar *pathname;
		gchar *end;
		*tag= g_strndup( name, len );
		name += strlen( "<!-- #BeginTemplate \"" );
		end = strchr( name, '"' );
		*path = g_strndup( name, end - name );
				
		pathname = screem_page_get_pathname( page );
		if( ! gnome_vfs_uris_match( pathname, *path ) ) {
			/* page is not the master template, it just uses it */
			g_free( *path );
			g_free( *tag );
			ret = FALSE;
		} else if( blocks ){
			/* build list of blocks */
			gchar *template_end;
			name = strstr( name, "-->" );
			g_assert( name );
			name += strlen( "-->" );
		
			template_end = strstr( name, "<!-- #EndTemplate -->" );
			
			do {
				gchar *ed_name = NULL;

				end = strstr( name, "<!-- #BeginEditable" );
				if( end ) {
					/* get name of editable section */
					
					ed_name = find_text( end, "\"[^\"]*\"",
								NULL, &len );
					end = strstr( end, "-->" );
					if( ed_name > end ) {
						ed_name = g_strdup( "" );
					} else {
						ed_name = g_strndup( ed_name,
									len );
					}
				}
				if( ! end ) {
					if( template_end ) {
						*blocks = g_slist_prepend( *blocks, g_strndup( name, template_end - name ) );		
					} else {
						*blocks = g_slist_prepend( *blocks, g_strdup( name ) );
					}
				} else {
					end += strlen( "-->" );
					*blocks = g_slist_prepend( *blocks, g_strndup( name, end - name ) );
					*blocks = g_slist_prepend( *blocks, ed_name );
					end = strstr( end, "<!-- #EndEditable -->" );
				}
				name = end;
			} while( name );
			*blocks = g_slist_reverse( *blocks );
		}
	}
	
	g_free( text );
	
	return ret;
}

/* G Object stuff */
#define PARENT_TYPE G_TYPE_OBJECT

static gpointer parent_class;

static void screem_page_class_init( ScreemPageClass *klass )
{
	GObjectClass *object_class;

	object_class = G_OBJECT_CLASS( klass );
	parent_class = g_type_class_peek_parent( klass );

	object_class->finalize = screem_page_finalize;

	screem_page_signals[ CHANGED ] = 
		g_signal_new( "changed",
			      G_OBJECT_CLASS_TYPE( object_class ),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET( ScreemPageClass, changed),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__BOOLEAN,
			      G_TYPE_NONE, 1,
			      G_TYPE_BOOLEAN );


	screem_page_signals[ DOCTYPE_CHANGED ] =
		g_signal_new( "doctype_changed",
			      G_OBJECT_CLASS_TYPE( object_class ),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET( ScreemPageClass, 
					       doctype_changed),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__STRING,
			      G_TYPE_NONE, 1,
			      G_TYPE_BOOLEAN );


	screem_page_signals[ CHARSET_CHANGED ] =
		g_signal_new( "charset_changed",
			      G_OBJECT_CLASS_TYPE( object_class ),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET( ScreemPageClass, 
					       charset_changed),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__STRING,
			      G_TYPE_NONE, 1,
			      G_TYPE_BOOLEAN );

	screem_page_signals[ SAVED ] =
		g_signal_new( "saved",
			      G_OBJECT_CLASS_TYPE( object_class ),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET( ScreemPageClass, 
					       saved),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__VOID,
			      G_TYPE_NONE, 0 );

}

static void screem_page_init( ScreemPage *page )
{
	page->private = g_new0( ScreemPagePrivate, 1 );

	page->private->client = gconf_client_get_default();
	page->private->highlight = gconf_client_get_bool(page->private->client,
							 "/apps/screem/editor/highlight",
							 NULL );
	page->private->highlight_notify = 
		gconf_client_notify_add( page->private->client,
					 "/apps/screem/editor/highlight",
					 screem_page_highlight_notify,
					 page, NULL, NULL );
	
	page->private->bookmarks = g_hash_table_new( g_str_hash, g_str_equal );

	/* NAME, LINE, BOOKMARK STRUCT */
	page->private->bookmarkstore = gtk_list_store_new( 3, G_TYPE_STRING,
							   G_TYPE_STRING,
							   G_TYPE_POINTER,
							   NULL );
	page->private->tag_tables = screem_application_load_syntax_tables();

	page->private->buffer = gtk_text_buffer_new( NULL );
	page->private->mime_type = "text/plain";
	
	g_signal_connect( G_OBJECT( page->private->buffer ),
			  "changed", G_CALLBACK( changed ), page );
	g_signal_connect_after( G_OBJECT( page->private->buffer ),
				"insert_text", 
				G_CALLBACK( insert_text ),
				page );
	g_signal_connect( G_OBJECT( page->private->buffer ),
			  "delete_range",
			  G_CALLBACK( delete_range_before ),
			  page );
	g_signal_connect_after( G_OBJECT( page->private->buffer ),
				"delete_range",
				G_CALLBACK( delete_range ),
				page );
}


static void screem_page_finalize( GObject *object )
{
	ScreemPage *page;

	page = SCREEM_PAGE( object );

	gconf_client_notify_remove( page->private->client, 
				    page->private->highlight_notify );

	g_hash_table_foreach( page->private->bookmarks,
			      (GHFunc)screem_page_clear_bookmarks, page );
	g_hash_table_destroy( page->private->bookmarks );
	g_object_unref( page->private->bookmarkstore );

	g_object_unref( page->private->client );

	if( page->private->pathname )
		g_free( page->private->pathname );
	if( page->private->charset )
		g_free( page->private->charset );

	if( page->private->buffer )
		g_object_unref( page->private->buffer );

	screem_page_clear_undo_redo_list( page, FALSE );
	screem_page_clear_undo_redo_list( page, TRUE );

	if( page->private->monitor ) {
		gnome_vfs_monitor_cancel( page->private->monitor );
	}

	g_free( page->private );

	G_OBJECT_CLASS( parent_class )->finalize( object );
}

GType screem_page_get_type()
{
	static GType type = 0;
	
	if( ! type ) {
		static const GTypeInfo info = {
			sizeof( ScreemPageClass ),
			NULL, /* base init */
			NULL, /* base finalise */
			(GClassInitFunc)screem_page_class_init,
			NULL, /* class finalise */
			NULL, /* class data */
			sizeof( ScreemPage ),
			0, /* n_preallocs */
			(GInstanceInitFunc)screem_page_init
		};

		type = g_type_register_static( PARENT_TYPE,
					       "ScreemPage",
					       &info, 0 );
	}

	return type;
}
