/*  Screem:  screem-site-ui.c,
 *  this file contains GUI code that deals with the site
 * 
 *  Copyright (C) 1999 - 2001  David A Knight
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  For contact information with the author of this source code please see
 *  the AUTHORS file.  If there is no AUTHORS file present then check the
 *  about box under the help menu for a contact address
 */

#include <config.h>

#include <glib/gfileutils.h>

#include <gtk/gtktogglebutton.h>

#include <libgnomeui/gnome-messagebox.h>
#include <libgnomeui/gnome-stock-icons.h>
#include <libgnomeui/gnome-dialog.h>
#include <libgnomeui/gnome-entry.h>
#include <libgnomeui/gnome-file-entry.h>
#include <libgnome/gnome-i18n.h>

#include <libgnomevfs/gnome-vfs-uri.h>
#include <libgnomevfs/gnome-vfs-ops.h>
#include <libgnomevfs/gnome-vfs-file-info.h>
#include <libgnomevfs/gnome-vfs-mime-utils.h>

#include <bonobo/bonobo-file-selector-util.h>

#include <string.h>

#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>

#include <gtk/gtkinvisible.h>

#include <glade/glade.h>

#include <libxml/tree.h>

#include "fileops.h"
#include "pageUI.h"
#include "support.h"
#include "xml.h"

#include "screem-application.h"
#include "screem-site.h"
#include "screem-site-ui.h"
#include "screem-page.h"
#include "screem-editor.h"
#include "screem-preview.h"

#include "screem-site-view.h"

#include "screem-file-browser.h"

#include "libegg/menu/egg-menu.h"

void screem_site_settings_dialog_cb( GtkWidget *widget, gint button );
void screem_site_settings_change_url( GtkWidget *widget, gpointer data );

static void screem_site_gather_stats( ScreemSite *site, GladeXML *xml );

static void screem_site_create_template_finish( GtkWidget *widget,
						gpointer data );

static void screem_site_create_template_select( GtkTreeSelection *selection,
						gpointer data );

static void screem_site_create_template_added( ScreemFileBrowser *browser,
					       const gchar *uri,
					       gpointer data );
static void screem_site_create_template_removed( ScreemFileBrowser *browser,
						 const gchar *uri,
						 gpointer data );
static xmlNodePtr screem_site_template_locate_uri( xmlNodePtr node, const gchar *uri );

void screem_site_close_site( ScreemSite *site )
{
       	g_return_if_fail( site != NULL );

	/* if this is the fake site then it can't be closed */
	if( ! screem_site_get_fake_flag( site ) ) {
		screem_site_save( site );
	}
}

gboolean screem_site_open_with_filename( ScreemWindow *window,
					 const gchar *filename )
{
	ScreemApplication *application;
	ScreemSite *site;
	gchar *path = NULL;
	GnomeVFSURI *uri;
	GnomeVFSFileInfo *info;

	g_return_val_if_fail( filename != NULL, FALSE );

	gdk_threads_enter();

	application = SCREEM_APPLICATION( window->application );

	uri = gnome_vfs_uri_new( filename );
	
	/* is the filename local? */
	if( gnome_vfs_uri_is_local( uri ) ) {
		/* yes */
		/* HACK */
		if( *filename == G_DIR_SEPARATOR ) {
			path = g_strconcat( "file://", filename, NULL );
		} else {
			path = relative_to_full( filename, NULL );
		}
	} else {
		path = g_strdup( filename );
	}
	gnome_vfs_uri_unref( uri );

	/* ensure we have a dir */
	info = gnome_vfs_file_info_new();

	gnome_vfs_get_file_info( path, info, 
				 GNOME_VFS_FILE_INFO_FOLLOW_LINKS );

	if( info->type != GNOME_VFS_FILE_TYPE_DIRECTORY )  {
		gchar *temp = g_dirname( path );
		g_free( path );
		path = temp;
	}

	gnome_vfs_file_info_unref( info );

	/* if we already have the site loaded, then don't do so again */
	site = screem_application_get_site( application, path );
	if( site ) {
		g_free( path );
		screem_window_set_current( window, site );
		gdk_threads_leave();
		return TRUE;
	}

	site = screem_site_new();
	screem_site_set_pathname( site, path );
	screem_site_set_name( site, _( "Imported Site" ) );

	g_free( path );

	if( ! screem_site_load( site ) ) {
		/* we failed to load the site */
		g_object_unref( G_OBJECT( site ) );
		gdk_threads_leave();
		return FALSE;
	}

	/* add this site to the list of loaded sites */
	screem_application_add_site( application, site );

	/* if we are importing a site open the site settings dialog */
	if( screem_site_get_is_import( site ) ) {
		screem_site_settings_dialog( site, window );
		screem_site_set_is_import( site, FALSE );
	}

	screem_window_set_current( window, site );

	/* site loaded, check for cvs update,
	   only do this if the site auto update flag is false,
	   and the user hasn't disabled displaying of the prompting dialog */

	if( screem_site_get_cvs_root( site ) && 
	    screem_site_get_auto_update( site ) ) {
		GladeXML *xml;
		GtkWidget *widget;
		gint button;
		GtkWidget *toggle;
		gboolean ask;
		gchar *gladepath;

		gladepath = gconf_client_get_string( application->client,
						     "/apps/screem/general/glade_path",
						     NULL );
		xml = glade_xml_new( gladepath,
				     "auto_cvs_update", NULL );
		g_free( gladepath );
		widget = glade_xml_get_widget( xml, "auto_cvs_update" );
		toggle = glade_xml_get_widget( xml, "dont_ask" );

		gtk_window_set_wmclass( GTK_WINDOW( widget ), 
					"Screem", "auto_cvs_update" );
		gtk_window_set_transient_for( GTK_WINDOW( widget ),
					      GTK_WINDOW( window ) );
		if( screem_site_get_auto_update_ask( site ) ) {
			button = gtk_dialog_run( GTK_DIALOG( widget ) );
			ask = !gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(toggle));
			screem_site_set_auto_update_ask( site, ask );
		} else {
			ask = TRUE;
			button = 0;
		}

		gtk_widget_destroy( widget );

		if( button == 0 ) {
			/* do a cvs update */
			EggAction *action;
			
			action = egg_action_group_get_action( EGG_ACTION_GROUP( window->action_group ),
								"UpdateSite" );
			egg_action_activate( action );
		} else if( ! ask ) {
			/* don't ask was selected and the dialog was
			   just closed */
			screem_site_set_auto_update( site, FALSE );
		}
	}

	/* do we auto load pages? */
	if( gconf_client_get_bool( application->client,
				   "/apps/screem/autoopen", NULL ) ) {
		/* yes we do */
		GList *list = screem_site_get_auto_open( site );
		const gchar *page_path;
		while( list ) {
			page_path = (gchar*)list->data;
			screem_page_open_with_filename( site, window, 
							page_path );
			list = list->next;
		}
	}

	gdk_threads_leave();
	
	return TRUE;
}

/**
 * screem_site_settings_dialog:
 * opens the site settings dialog
 *
 * return values: none
 */
void screem_site_settings_dialog( ScreemSite *site, ScreemWindow *window )
{
	GtkWidget *dialog1;
     	GtkWidget *entry;
	GtkWidget *toggle;
	UploadMethods method;
	ScreemSitePerms permissions;
	ScreemSiteSymlinks symlinks;
	const gchar *text;
	GladeXML *xml;
	GConfClient *client;
	gchar *gladepath;

	gint button;

	g_return_if_fail( site != NULL );
	g_return_if_fail( ! screem_site_get_fake_flag( site ) );

	client = gconf_client_get_default();
	gladepath = gconf_client_get_string( client, 
					     "/apps/screem/general/glade_path",
					     NULL );
	xml = glade_xml_new( gladepath, "site_settings", NULL );
	g_free( gladepath );
	g_object_unref( client );

	glade_xml_signal_autoconnect( xml );
		
	screem_site_gather_stats( site, xml );
		
	dialog1 = glade_xml_get_widget( xml, "site_settings" );
	
	/* fill in the details */
	/* name */
	entry = glade_xml_get_widget( xml, "site_name_entry" );
	entry = gnome_entry_gtk_entry( GNOME_ENTRY( entry ) );
	text = screem_site_get_name( site );
	if( ! text ) {
		text = _( "Imported Site ");
	}
	gtk_entry_set_text( GTK_ENTRY( entry ), text );
	
	/* path */
	entry = glade_xml_get_widget( xml, "site_path_entry" );
	text = screem_site_get_pathname( site );
	if( ! text ) {
		text = "";
	}
	gtk_label_set_text( GTK_LABEL( entry ), text );
	
	/* cvs root */
	entry = glade_xml_get_widget( xml, "cvs_file_entry" );
	if( ( text = screem_site_get_cvs_root( site ) ) ) {
		toggle = glade_xml_get_widget( xml, "cvs_check" );
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON( toggle ),
					     TRUE );
		entry = gnome_file_entry_gtk_entry( GNOME_FILE_ENTRY(entry ) );
		gtk_entry_set_text( GTK_ENTRY( entry ), text );
	}
	
	/* template path */
	entry = glade_xml_get_widget( xml, "template_file_entry" );
	if( ( text = screem_site_get_template_path( site ) ) ) {
		toggle = glade_xml_get_widget( xml, "template_check" );
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON( toggle ),
					     TRUE );
		entry = gnome_file_entry_gtk_entry( GNOME_FILE_ENTRY(entry ) );
		gtk_entry_set_text(GTK_ENTRY( entry ), text );
	}
	
	/* auto cvs update */
	toggle = glade_xml_get_widget( xml, "auto_cvs_update" );
	gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON( toggle ),
				      screem_site_get_auto_update( site ) );

	/* internal preview settings */
	toggle = glade_xml_get_widget( xml, "preview_static" );
	gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON( toggle ),
				      screem_site_get_preview_static( site ) );

	toggle = glade_xml_get_widget( xml, "preview_dynamic" );
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON( toggle ),
				     screem_site_get_preview_dynamic( site ) );

	/* remote url */
	entry = glade_xml_get_widget( xml, "remote_url_entry" );
	entry = gnome_entry_gtk_entry( GNOME_ENTRY( entry ) );
	text = screem_site_get_remote_url( site );
	if( ! text ) {
		text = "";
	}
	gtk_entry_set_text( GTK_ENTRY( entry ), text );
	
	/* http url */
	entry = glade_xml_get_widget( xml, "http_url_entry" );
	entry = gnome_entry_gtk_entry( GNOME_ENTRY( entry ) );
	text = screem_site_get_http_url( site );
	if( ! text ) {
		text = "";
	}
	gtk_entry_set_text( GTK_ENTRY( entry ), text );
	
	/* remote method */
	method = screem_site_get_remote_method( site );
	toggle = glade_xml_get_widget( xml, upload_strings[ method ] );
	gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON( toggle ), TRUE );
	
	/* passive ftp */
	if( screem_site_get_passive_ftp( site ) ) {
		toggle = glade_xml_get_widget( xml, "ftpmode" );
		gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON( toggle ), TRUE );
	}
	
	/* no file delete */
	if( screem_site_get_no_delete( site ) ) {
		toggle = glade_xml_get_widget( xml, "nodelete" );
		gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON( toggle ), TRUE );
	}
	
	/* check for moved files */
	if( screem_site_get_check_moved( site ) ) {
		toggle = glade_xml_get_widget( xml, "checkmoved" );
		gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON( toggle ), TRUE );
	}
	
	/* no file overwrite */
	if( screem_site_get_no_overwrite( site ) ) {
		toggle = glade_xml_get_widget( xml, "nooverwrite" );
		gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON( toggle ), TRUE );
	}
	
	/* permissions handling */
	permissions = screem_site_get_permissions( site );
	toggle = glade_xml_get_widget( xml, permission_strings[permissions] );
	gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON( toggle ), TRUE );
	
	/* symbolic link handling */
	symlinks = screem_site_get_symlinks( site );
	toggle = glade_xml_get_widget( xml, symlink_strings[ symlinks ] );
	gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON( toggle ), TRUE );
	
	/* remote path */
	entry = glade_xml_get_widget( xml, "remote_path_entry" );
	entry = gnome_entry_gtk_entry( GNOME_ENTRY( entry ) );
	text = screem_site_get_remote_path( site );
	if( ! text ) {
		text = "";
	}
	gtk_entry_set_text( GTK_ENTRY( entry ), text );
	
	/* remote user */
	entry = glade_xml_get_widget( xml, "remote_user_entry" );
	entry = gnome_entry_gtk_entry( GNOME_ENTRY( entry ) );
	text = screem_site_get_remote_user( site );
	if( ! text ) {
		text = "";
	}
	gtk_entry_set_text( GTK_ENTRY( entry ), text );
	
	/* remote password */
	entry = glade_xml_get_widget( xml, "remote_password_entry" );
	text = screem_site_get_remote_pass( site );
	if( ! text ) {
		text = "";
	}
	gtk_entry_set_text( GTK_ENTRY( entry ), text );
	
	/* attach the site to the dialog */
	g_object_set_data( G_OBJECT( dialog1 ), "site", site );
	
	gtk_window_set_transient_for( GTK_WINDOW( dialog1 ),
				      GTK_WINDOW( window ) );

	button = gtk_dialog_run( GTK_DIALOG( dialog1 ) );
	
	if( button == GTK_RESPONSE_CLOSE ) {
		screem_site_settings_dialog_cb( dialog1, button );
	}
	gtk_widget_destroy( dialog1 );
	g_object_unref( xml );
}

void screem_site_settings_change_url( GtkWidget *widget, gpointer data )
{
	const gchar *host;
	GtkWidget *url;
	GtkWidget *rhost;
	GtkWidget *path;
	GtkWidget *webdav;
	GladeXML *xml;
	const gchar *pathname;
	gchar *baseuri;

	xml = glade_get_widget_tree( widget );

	webdav = glade_xml_get_widget( xml, upload_strings[ WEBDAV ] );
	if( gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON( webdav ) ) ) {
	       	url = glade_xml_get_widget( xml, "http_url_entry" );
		path = glade_xml_get_widget( xml, "remote_path_entry" );
		rhost = glade_xml_get_widget( xml, "remote_url_entry" );
		
		host = gtk_entry_get_text( GTK_ENTRY( rhost ) );
		pathname = gtk_entry_get_text( GTK_ENTRY( path ) );
		
		baseuri = g_strconcat( "http://", host, pathname, NULL );
		gtk_entry_set_text( GTK_ENTRY( url ), baseuri );
		g_free( baseuri );
	}
}

void screem_site_settings_dialog_cb( GtkWidget *widget, gint button ) 
{
	GtkWidget *entry;
	GtkWidget *toggle;
	ScreemSite *site;
	GladeXML *xml;

	xml = glade_get_widget_tree( widget );

	site = SCREEM_SITE( g_object_get_data( G_OBJECT( widget ), "site") );
	
	/* name */
	entry = glade_xml_get_widget( xml, "site_name_entry" );
	entry = gnome_entry_gtk_entry( GNOME_ENTRY( entry ) );
	screem_site_set_name( site, 
			      gtk_entry_get_text( GTK_ENTRY(entry) ) );
	/* cvs root */
	toggle = glade_xml_get_widget( xml, "cvs_check" );
	if( gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON( toggle ) ) ) {
		entry = glade_xml_get_widget(xml,"cvs_file_entry");
		entry = gnome_file_entry_gtk_entry( GNOME_FILE_ENTRY( entry ) );
		screem_site_set_cvs_root( site, 
					  gtk_entry_get_text( GTK_ENTRY( entry ) ) );
	} else {
		screem_site_set_cvs_root( site, NULL );
	}
	
	/* template */
	toggle = glade_xml_get_widget( xml, "template_check" );
	if( gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON( toggle ) ) ) {
		entry = glade_xml_get_widget( xml, 
					      "template_file_entry" );
		entry = gnome_file_entry_gtk_entry( GNOME_FILE_ENTRY( entry ) );
		screem_site_set_template_path( site, 
					       gtk_entry_get_text( GTK_ENTRY( entry ) ) );
	} else {
		screem_site_set_template_path( site, NULL );
	}

	/* cvs update */
	toggle = glade_xml_get_widget( xml, "auto_cvs_update" );
	screem_site_set_auto_update( site,
				     gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON( toggle ) ) );

	/* preview */
	toggle = glade_xml_get_widget( xml, "preview_static" );
	screem_site_set_preview_static( site,
					gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON( toggle ) ) );

	toggle = glade_xml_get_widget( xml, "preview_dynamic" );
	screem_site_set_preview_dynamic( site,
					 gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON( toggle ) ) );
	
	/* remote url */
	entry = glade_xml_get_widget( xml, "remote_url_entry" );
	entry = gnome_entry_gtk_entry( GNOME_ENTRY( entry ) );
	screem_site_set_remote_url( site, gtk_entry_get_text( GTK_ENTRY( entry ) ) );
	/* http url */
	entry = glade_xml_get_widget( xml, "http_url_entry" );
	entry = gnome_entry_gtk_entry( GNOME_ENTRY( entry ) );
	screem_site_set_http_url( site, gtk_entry_get_text( GTK_ENTRY( entry ) ) );
	
	/* remote method */
	toggle = glade_xml_get_widget( xml, "local" );
	if(gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON(toggle) )) {
		screem_site_set_remote_method( site, LOCAL );
	}
	toggle = glade_xml_get_widget( xml, "ftp" );
	if(gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON(toggle) )) {
		screem_site_set_remote_method( site, FTP );
	}
	toggle = glade_xml_get_widget( xml, "webdav" );
	if(gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON(toggle) )) {
		screem_site_set_remote_method( site, WEBDAV );
	}
	toggle = glade_xml_get_widget( xml, "rsh" );
	if(gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON(toggle) )) {
		screem_site_set_remote_method( site, RSH );
	}
	toggle = glade_xml_get_widget( xml, "ssh" );
	if(gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON(toggle) )) {
		screem_site_set_remote_method( site, SSH );
	}
	
	
	/* passive ftp */
	toggle = glade_xml_get_widget( xml, "ftpmode" );
	screem_site_set_passive_ftp( site, gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON(toggle) ));
	
	/* no file delete */
	toggle = glade_xml_get_widget( xml, "nodelete" );
	screem_site_set_no_delete( site, gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON(toggle) ));
	
	/* check for moved files */
	toggle = glade_xml_get_widget( xml, "checkmoved" );
	screem_site_set_check_moved( site, gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON(toggle) ));
	
	/* no file overwrite */
	toggle = glade_xml_get_widget( xml, "nooverwrite" );
	screem_site_set_no_overwrite( site, gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON(toggle) ));
	
	/* permissions handling */
	toggle = glade_xml_get_widget( xml, "p_ignore" );
	if(gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON(toggle) )) {
		screem_site_set_permissions( site, PERMS_IGNORE );
	}
	toggle = glade_xml_get_widget( xml, "p_exec" );
	if(gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON(toggle) )) {
		screem_site_set_permissions( site, PERMS_EXEC );
	}
	toggle = glade_xml_get_widget( xml, "p_all" );
	if(gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON(toggle) )) {
		screem_site_set_permissions( site, PERMS_ALL );
	}
	
	/* symbolic link handling */
	toggle = glade_xml_get_widget( xml, "s_ignore" );
	if(gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON(toggle) )) {
		screem_site_set_symlinks( site, SYM_IGNORE );
	}
	toggle = glade_xml_get_widget( xml, "s_follow" );
	if(gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON(toggle) )) {
		screem_site_set_symlinks( site, SYM_FOLLOW );
	}
	toggle = glade_xml_get_widget( xml, "s_maintain" );
	if(gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON(toggle) )) {
		screem_site_set_symlinks( site, SYM_MAINTAIN );
	}
	
	/* remote path */
	entry = glade_xml_get_widget( xml, "remote_path_entry" );
	entry = gnome_entry_gtk_entry( GNOME_ENTRY( entry ) );
	screem_site_set_remote_path( site, gtk_entry_get_text( GTK_ENTRY( entry ) ) );
	
	/* remote user */
	entry = glade_xml_get_widget( xml, "remote_user_entry" );
	entry = gnome_entry_gtk_entry( GNOME_ENTRY( entry ) );
	screem_site_set_remote_user( site, gtk_entry_get_text( GTK_ENTRY( entry ) ) );
	
	/* remote password */
	entry = glade_xml_get_widget( xml, 
				      "remote_password_entry");
	screem_site_set_remote_pass( site, gtk_entry_get_text( GTK_ENTRY( entry ) ) );

	screem_site_save( site );
}

static void screem_site_gather_stats( ScreemSite *site, GladeXML *xml )
{
	GtkWidget *last_upload;
	GtkWidget *files_list;
	GtkWidget *total_size;

	GtkCellRenderer *prenderer;
	GtkCellRenderer *renderer;
	GtkTreeViewColumn *column;

	GtkTreeModel *model;

	gchar *num;

	gint total;
	gchar *sitecopy_file;
	const gchar *name;

	struct stat st;

	last_upload = glade_xml_get_widget( xml, "last_upload" );
	files_list = glade_xml_get_widget( xml, "files_list" );
	total_size = glade_xml_get_widget( xml, "total_size" );

	name = screem_site_get_name( site );

	sitecopy_file = g_strconcat( g_get_home_dir(), G_DIR_SEPARATOR_S,
				     ".sitecopy", G_DIR_SEPARATOR_S, name,
				     NULL );
	if( ! stat( sitecopy_file, &st ) ) {
		gchar *tmp;

		tmp = ctime( &st.st_mtime );
		tmp[ strlen( tmp ) - 1 ] = '\0';
		gtk_label_set_text( GTK_LABEL( last_upload ), tmp );
	} else {
		gtk_label_set_text( GTK_LABEL( last_upload ), _("Unknown") );
	}
	g_free( sitecopy_file );

	model = screem_site_get_statistics( site );
	total = GPOINTER_TO_INT( g_object_get_data( G_OBJECT( model ), 
						    "total" ) );
	num = g_strdup_printf( _("%i bytes"), total );
	gtk_label_set_text( GTK_LABEL( total_size ), num );
	g_free( num );

	prenderer = gtk_cell_renderer_pixbuf_new();
	renderer = gtk_cell_renderer_text_new();
	column = gtk_tree_view_column_new();

	gtk_tree_view_column_set_title( column, _( "Mime Type" ) );
	gtk_tree_view_column_pack_start( column, prenderer, FALSE );
	gtk_tree_view_column_pack_start( column, renderer, TRUE );
	gtk_tree_view_column_set_resizable( column, TRUE );
	
	gtk_tree_view_append_column( GTK_TREE_VIEW( files_list ), column );
	gtk_tree_view_column_set_attributes( column, renderer,
					     "text", 0,
					     NULL );
	gtk_tree_view_column_set_attributes( column, prenderer,
					     "pixbuf", 1,
					     NULL );

	renderer = gtk_cell_renderer_text_new();
	column = gtk_tree_view_column_new();
	
	gtk_tree_view_column_set_title( column, _( "Number" ) );
	gtk_tree_view_column_pack_start( column, renderer, TRUE );
	gtk_tree_view_append_column( GTK_TREE_VIEW( files_list ), column );
	gtk_tree_view_column_set_attributes( column, renderer,
					     "text", 2, NULL );

	renderer = gtk_cell_renderer_text_new();
	column = gtk_tree_view_column_new();
	
	gtk_tree_view_column_set_title( column, _( "Total Size" ) );
	gtk_tree_view_column_pack_start( column, renderer, TRUE );
	gtk_tree_view_append_column( GTK_TREE_VIEW( files_list ), column );
	gtk_tree_view_column_set_attributes( column, renderer,
					     "text", 3, NULL );

	renderer = gtk_cell_renderer_text_new();
	column = gtk_tree_view_column_new();
	
	gtk_tree_view_column_set_title( column, _( "Average Size" ) );
	gtk_tree_view_column_pack_start( column, renderer, TRUE );
	gtk_tree_view_append_column( GTK_TREE_VIEW( files_list ), column );
	gtk_tree_view_column_set_attributes( column, renderer,
					     "text", 4, NULL );

	gtk_tree_view_set_model( GTK_TREE_VIEW( files_list ), model );
	g_object_unref( model );
}


void screem_site_create_template( ScreemWindow *window )
{
	gchar *uri;

	uri = tempnam( NULL, NULL );
	if( uri ) {
		if( ! mkdir_recursive( uri,
				       GNOME_VFS_PERM_USER_ALL ) ) {
			g_free( uri );
			uri = NULL;
		}
	}

	if( uri ) {
		GConfClient *client;
		gchar *glade_path;
		GladeXML *xml;
		ScreemFileBrowser *browser;
		GtkWidget *widget;
		GtkTreeModel *model;
		GtkCellRenderer *prenderer;
		GtkCellRenderer *renderer;
		GtkTreeViewColumn *column;
		GtkTreeSelection *selection;
		xmlDocPtr doc;

		doc = xmlNewDoc( XML_DEFAULT_VERSION );
		xmlDocSetRootElement( doc,
				      xmlNewDocNode( doc, NULL,
						     "screem_template",
						     NULL ) );

		client = gconf_client_get_default();
		
		glade_path = gconf_client_get_string(client,
						     "/apps/screem/general/glade_path",
						     NULL );

		xml = glade_xml_new( glade_path, "site_template_window",
				     NULL );

		widget = glade_xml_get_widget( xml, "view" );
		
		browser = screem_file_browser_new();
		
		screem_file_browser_scan_directory( browser, uri,
						    FILE_BROWSE_ROOT |
						    FILE_BROWSE_RECURSE );

		model = screem_file_browser_get_model( browser );

		selection = gtk_tree_view_get_selection( GTK_TREE_VIEW(widget) );
		gtk_tree_selection_set_mode( GTK_TREE_SELECTION( selection ),
					     GTK_SELECTION_SINGLE );
		
		prenderer = gtk_cell_renderer_pixbuf_new();
		renderer = gtk_cell_renderer_text_new();
		column = gtk_tree_view_column_new();
		gtk_tree_view_column_set_title( column, "Template" );
		gtk_tree_view_column_pack_start( column, prenderer, FALSE );
		gtk_tree_view_column_pack_start( column, renderer, TRUE );
		gtk_tree_view_column_set_resizable( column, TRUE );
		gtk_tree_view_append_column( GTK_TREE_VIEW( widget ), column );
		gtk_tree_view_column_set_attributes( column, renderer,
						     "text", 
						     FILE_BROWSER_NAME_COL,
						     NULL );
		gtk_tree_view_column_set_attributes( column, prenderer,
						     "pixbuf",
						     FILE_BROWSER_ICON_COL,
						     NULL );
		
		gtk_tree_view_set_model( GTK_TREE_VIEW( widget ), model );
		
		widget = glade_xml_get_widget( xml, "site_template_window" );

		/* connect to selection change signal */
		g_signal_connect( G_OBJECT( selection ), "changed",
				  G_CALLBACK(screem_site_create_template_select),
				  widget );

		g_object_set_data( G_OBJECT( widget ), "doc", doc );
		g_object_set_data( G_OBJECT( browser ), "uri", uri );
		g_object_set_data( G_OBJECT( widget ), "browser", browser );
		g_signal_connect( G_OBJECT( widget ),
				  "destroy",
				  G_CALLBACK(screem_site_create_template_finish),
				  browser );
		g_signal_connect( G_OBJECT( browser ), "added",
				  G_CALLBACK(screem_site_create_template_added),
				  widget );
		g_signal_connect(G_OBJECT( browser ), "removed",
				 G_CALLBACK(screem_site_create_template_removed),
				 widget );

		gtk_window_set_transient_for( GTK_WINDOW( widget ),
					      GTK_WINDOW( window ) );
		gtk_widget_show_all( widget );
		glade_xml_signal_autoconnect( xml );

		g_free( glade_path );
		g_object_unref( client );
	} else {
		/* FIXME: inform user failed to create temp directory */

	}
}

void screem_site_create_template_response( GtkWidget *dialog,
					   gint response )
{
	xmlDocPtr doc;
	gchar *filename;
	const gchar *uri;
	ScreemFileBrowser *browser;

	doc = (xmlDocPtr)g_object_get_data( G_OBJECT( dialog ), "doc" );
	browser = SCREEM_FILE_BROWSER( g_object_get_data( G_OBJECT( dialog ),
							  "browser" ) );
	uri = (const gchar*)g_object_get_data( G_OBJECT( browser ), "uri" );

	switch( response ) {
	case 0:
		/* load */
		filename = bonobo_file_selector_open( NULL, FALSE,
						      _("Select Site Template"),
						      NULL, NULL );
		if( filename ) {
			xmlDocPtr newdoc;
			ScreemSite *site;
			gchar *projectfile;
			gchar *url;

			url = tempnam( NULL, NULL );
			if( url ) {
				if(!mkdir_recursive(url,
						    GNOME_VFS_PERM_USER_ALL)) {
					g_free( url );
					url = NULL;
				}
			}
			if( url ) {
				g_object_set_data( G_OBJECT( browser ), "uri",
						   url );

				newdoc = xmlParseFile( filename );
				g_object_set_data( G_OBJECT( dialog ), "doc",
						   NULL );
				
				site = screem_site_new();
				screem_site_set_site_template( site, filename );
				screem_site_set_pathname( site, url );
				screem_site_create( site );
				g_object_unref( site );
				projectfile = g_strconcat( url, 
							   G_DIR_SEPARATOR_S,
							   ".project.screem",
							   NULL );
				delete_file( projectfile );
				g_free( projectfile );
				
				g_object_set_data( G_OBJECT( dialog ), "doc",
						   newdoc );
				
				xmlFreeDoc( doc );
				
				g_free( filename );

				screem_file_browser_scan_directory( browser,
								    url,
								    FILE_BROWSE_ROOT | FILE_BROWSE_RECURSE );
				delete_dir( uri );
				g_free( (gchar*)uri );
			}
		}
		break;
	case 1:
		filename = bonobo_file_selector_save( NULL, FALSE,
						      _("Select directory to save site as"),
						      NULL, NULL, NULL);
		if( filename ) {
			xmlSaveFile( filename, doc );
			g_free( filename );
		}
		break;
	default:
		gtk_widget_destroy( dialog );
		break;
	}
}

static void screem_site_create_template_finish( GtkWidget *widget,
						gpointer data )
{
	ScreemFileBrowser *browser;
	xmlDocPtr doc;
	gchar *uri;

	browser = SCREEM_FILE_BROWSER( data );
	uri = (gchar*)g_object_get_data( G_OBJECT( browser ), "uri" );

	doc = (xmlDocPtr)g_object_get_data( G_OBJECT( widget ), "doc" );
	if( doc ) {
		xmlFreeDoc( doc );
	}
	delete_dir( uri );
	g_free( uri );
	g_object_unref( browser );
}

static void screem_site_create_template_select( GtkTreeSelection *selection,
						gpointer data )
{
	GtkWidget *widget;
	GladeXML *xml;
	GtkTreeModel *model;
	GtkTreeIter it;
	GtkTreeIter sit;

	xmlDocPtr doc;
	ScreemFileBrowser *browser;
	const gchar *base;

	doc = (xmlDocPtr)g_object_get_data( G_OBJECT( data ), "doc" );
	browser = SCREEM_FILE_BROWSER( g_object_get_data( G_OBJECT( data ),
							  "browser" ) );
	base = (const gchar*)g_object_get_data( G_OBJECT( browser ), "uri" );


	widget = GTK_WIDGET( gtk_tree_selection_get_tree_view( selection ) );
	xml = glade_get_widget_tree( widget );

	model = gtk_tree_view_get_model( GTK_TREE_VIEW( widget ) );
	gtk_tree_model_get_iter_first( model, &it );

	widget = glade_xml_get_widget( xml, "maintable" );

	if( gtk_tree_selection_get_selected( selection, &model,
					     &sit ) ) {
		gtk_widget_set_sensitive( widget, TRUE );

		if( ! memcmp( &it, &sit, sizeof( GtkTreeIter ) ) ) {
			/* clear ui */
			widget = glade_xml_get_widget( xml, "name" );
			gtk_entry_set_text( GTK_ENTRY( widget ), "" );
			widget = glade_xml_get_widget( xml, "template" );
			gtk_entry_set_text( GTK_ENTRY( widget ), "" );
		} else {
			/* fill in ui */
			GList *list;
			GList *tmp;
			const gchar *url;
			const gchar *uri;
			xmlNodePtr node;
			gchar *name;
			gchar *template;

			list = g_list_append( NULL, &sit );
			tmp = screem_file_browser_get_pathnames( browser,
								 list, FALSE );
			g_list_free( list );
			uri = url = (const gchar*)tmp->data;
			g_list_free( tmp );
			if( ! strncmp( "file://", url, strlen( "file://" ) ) ) {
				url += strlen( "file://" );
			}
			url += strlen( base );

			node = xmlDocGetRootElement( doc );
			node = screem_site_template_locate_uri( node, url );
			if( node ) {
				name = xmlGetProp( node, "name" );
				template = xmlGetProp( node, "template" );
			} else {
				name = NULL;
				template = NULL;
			}

			if( ! name ) {
 				name = g_strdup( "" );
			}
			if( ! template ) {
				template = g_strdup( "" );
			}

			widget = glade_xml_get_widget( xml, "name" );
			gtk_entry_set_text( GTK_ENTRY( widget ), name );
			widget = glade_xml_get_widget( xml, "template" );
			gtk_entry_set_text( GTK_ENTRY( widget ), template );

			g_free( name );
			g_free( template );

			if( screem_uri_is_dir( uri ) ) {
				widget = glade_xml_get_widget( xml, "directory");
			} else {
				widget = glade_xml_get_widget( xml, "file" );
			}
			gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON( widget ),
						      TRUE );
		}
	} else {
		gtk_tree_selection_select_iter( selection, &it );
	}
}

static void screem_site_create_template_added( ScreemFileBrowser *browser,
					       const gchar *uri,
					       gpointer data )
{
	GtkWidget *widget;
	GladeXML *xml;
	xmlDocPtr doc;
	const gchar *base;
	const gchar *url;

	gchar **names;
	gint i;
	xmlNodePtr node;
	gboolean isdir;

	widget = GTK_WIDGET( data );
	xml = glade_get_widget_tree( widget );
	doc = (xmlDocPtr)g_object_get_data( G_OBJECT( widget ), "doc" );
	base = (const gchar*)g_object_get_data( G_OBJECT( browser ), "uri" );

	url = uri;
	if( ! strncmp( "file://", url, strlen( "file://" ) ) &&
	    strncmp( "file://", base, strlen( "file://" ) ) ) {
		uri += strlen( "file://" );
	}
	uri += strlen( base );

	/* traverse doc creating nodes as necessary to locate
	   a node with name == uri, if we can't locate it, create it */

	if( doc ) {
		node = xmlDocGetRootElement( doc );
	} else {
		node = NULL;
	}
      
	names = g_strsplit( uri, G_DIR_SEPARATOR_S, -1 );
	for( i = 0; node && names[ i ]; ++ i ) {
		xmlNodePtr child;

		if( *names[ i ] == '\0' ) {
			continue;
		}

		child = node->xmlChildrenNode;
		while( child ) {
			gchar *name;
			gboolean match;

			name = xmlGetProp( child, "name" );
			match = ( name && ! strcmp( names[ i ], name ) );
			g_free( name );
			if( match ) {
				break;
			} else {
				child = child->next;
			}
		}
		if( ! child ) {
			/* add node */
			if( names[ i + 1 ] ) {
				isdir = TRUE;
			} else {
				isdir = screem_uri_is_dir( url );
			}
			if( isdir ) {
				child = xmlNewChild( node, NULL, "directory",
						     NULL );
			} else {
				child = xmlNewChild( node, NULL, "file",
						     NULL );
			}
			xmlSetProp( child, "name", names[ i ] );
		}
		node = child;
	}

	g_strfreev( names );
}

static void screem_site_create_template_removed( ScreemFileBrowser *browser,
						 const gchar *uri,
						 gpointer data )
{
	GtkWidget *widget;
	GladeXML *xml;
	xmlDocPtr doc;
	const gchar *base;
	const gchar *url;

	xmlNodePtr node;

	widget = GTK_WIDGET( data );
	xml = glade_get_widget_tree( widget );
	doc = (xmlDocPtr)g_object_get_data( G_OBJECT( widget ), "doc" );
	base = (const gchar*)g_object_get_data( G_OBJECT( browser ), "uri" );

	url = uri;
	uri += strlen( "file://" ) + strlen( base );

	/* traverse doc and delete relevant node */

	if( doc ) {
		node = xmlDocGetRootElement( doc );
	} else {
		node = NULL;
	}
      
	node = screem_site_template_locate_uri( node, uri );
	if( node ) {
		xmlUnlinkNode( node );
		xmlFreeNode( node );
	}
}

static xmlNodePtr screem_site_template_locate_uri( xmlNodePtr node, const gchar *uri )
{
	gint i;
	gchar **names;

	names = g_strsplit( uri, G_DIR_SEPARATOR_S, -1 );
	for( i = 0; node && names[ i ]; ++ i ) {
		xmlNodePtr child;

		if( *names[ i ] == '\0' ) {
			continue;
		}

		child = node->xmlChildrenNode;
		while( child ) {
			gchar *name;
			gboolean match;

			name = xmlGetProp( child, "name" );
			match = ( name && ! strcmp( names[ i ], name ) );
			g_free( name );
			if( match ) {
				break;
			} else {
				child = child->next;
			}
		}
		node = child;

		if( node && ! names[ i + 1 ] ) {
			break;
		}
			
	}

	g_strfreev( names );

	return node;
}

void screem_site_create_template_add( GtkWidget *widget )
{
	GladeXML *xml;
	ScreemFileBrowser *browser;
	GtkTreeSelection *selection;
	GtkTreeModel *model;
	GtkTreeIter it;
	xmlDocPtr doc;
	
	xml = glade_get_widget_tree( widget );

	widget = glade_xml_get_widget( xml, "site_template_window" );
	browser = SCREEM_FILE_BROWSER( g_object_get_data( G_OBJECT( widget ),
							  "browser" ) );
	doc = (xmlDocPtr)g_object_get_data( G_OBJECT( widget ), "doc" );

	widget = glade_xml_get_widget( xml, "view" );
	selection = gtk_tree_view_get_selection( GTK_TREE_VIEW( widget ) );
	model = gtk_tree_view_get_model( GTK_TREE_VIEW( widget ) );

	if( gtk_tree_selection_get_selected( selection, &model, &it ) ) {
		GList *list;
		GList *tmp;
		const gchar *url;
		gchar *uri;
		gchar *temp;
		const gchar *name;
		gboolean isdir;
		gboolean created;

		widget = glade_xml_get_widget( xml, "name" );
		name = gtk_entry_get_text( GTK_ENTRY( widget ) );
		widget = glade_xml_get_widget( xml, "directory");
		isdir = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget));

		list = g_list_append( NULL, &it );
		tmp = screem_file_browser_get_pathnames( browser,
							 list, FALSE );
		g_list_free( list );
		url = (const gchar*)tmp->data;
		g_list_free( tmp );

		/* if url isn't a directory, we want its parent */
		if( ! screem_uri_is_dir( url ) ) {
			temp = g_dirname( url );
		} else {
			temp = g_strdup( url );
		}
		uri = g_strconcat( temp, G_DIR_SEPARATOR_S, name, NULL );
		g_free( temp );

		if( isdir ) {
			created = mkdir_recursive( uri,  
						   GNOME_VFS_PERM_USER_ALL );
		} else {
			created = save_file( uri, "", 
					     GNOME_VFS_PERM_USER_ALL );
		}
		if( created ) {
			xmlNodePtr node;
			const gchar *template;

			widget = glade_xml_get_widget( xml,
						       "site_template_window" );

			screem_site_create_template_added( browser,
							   uri,
							   widget );
			node = xmlDocGetRootElement( doc );
			node = screem_site_template_locate_uri( node,
								uri );
			widget = glade_xml_get_widget( xml, "template" );
			template = gtk_entry_get_text( GTK_ENTRY( widget ) );
			if( *template ) {
				xmlSetProp( node, "template", template );
			}
		}
		g_free( uri );
	}
}


void screem_site_create_template_remove( GtkWidget *widget )
{
	GladeXML *xml;
	ScreemFileBrowser *browser;
	GtkTreeSelection *selection;
	GtkTreeModel *model;
	GtkTreeIter it;
	
	xml = glade_get_widget_tree( widget );

	widget = glade_xml_get_widget( xml, "site_template_window" );
	browser = SCREEM_FILE_BROWSER( g_object_get_data( G_OBJECT( widget ),
							  "browser" ) );

	widget = glade_xml_get_widget( xml, "view" );
	selection = gtk_tree_view_get_selection( GTK_TREE_VIEW( widget ) );
	model = gtk_tree_view_get_model( GTK_TREE_VIEW( widget ) );

	if( gtk_tree_selection_get_selected( selection, &model, &it ) ) {
		GList *list;
		GList *tmp;
		const gchar *url;

		list = g_list_append( NULL, &it );
		tmp = screem_file_browser_get_pathnames( browser,
							 list, FALSE );
		g_list_free( list );
		url = (const gchar*)tmp->data;
		g_list_free( tmp );
		delete_dir( url );
	}
}
