/*  Screem:  screem-tagtree.c
 *
 *  the tag tree object
 *
 *  Copyright (C) 2001 David A Knight
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  For contact information with the author of this source code please see
 *  the AUTHORS file.  If there is no AUTHORS file present then check the
 *  about box under the help menu for a contact address
 */
#include <config.h>

#include <dirent.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>

#include <gtk/gtktreeview.h>
#include <gtk/gtkscrolledwindow.h>
#include <gtk/gtkcellrenderertext.h>
#include <gtk/gtkhbox.h>
#include <gtk/gtkvbox.h>
#include <gtk/gtksizegroup.h>
#include <gtk/gtkdialog.h>

#include <libgnome/gnome-i18n.h>

#include <libgnomevfs/gnome-vfs-mime-handlers.h>
#include <libgnomevfs/gnome-vfs-mime-utils.h>
#include <libgnomevfs/gnome-vfs-uri.h>

#include <glib/gmarkup.h>

#include "screem-tagtree.h"

#include "screem-window.h"
#include "screem-window-private.h"
#include "screem-editor.h"

#include "screem-file-browser.h"

#include "fileops.h"

struct ScreemTagTreePrivate {
	GtkWidget *box;
	
	GtkWidget *view;

	GtkTreeStore *store;

	GHashTable *paths;

	ScreemWindow *window;
	ScreemFileBrowser *browser;
};

typedef struct {
	ScreemFileBrowser *browser;
	GtkTreeModel *model;
	GHashTable *paths;
} ScreemTagTreeBuildData;

typedef struct {
	ScreemTagTreeBuildData *data;
	GSList *states;
	GSList *iters;
} ScreemTagTreeParser;

static gboolean screem_tag_tree_parse_file( ScreemTagTreeBuildData *bdata, 
						const gchar *filename );
static gboolean screem_tag_tree_parse_string( ScreemTagTreeBuildData *bdata, 
						const gchar *str,
						guint len );

static void screem_tag_tree_start_element( GMarkupParseContext *context,
					const gchar *element,
					const gchar **attributes,
					const gchar **values,
					gpointer data,
					GError **error );
static void screem_tag_tree_end_element( GMarkupParseContext *context,
					const gchar *element,
					gpointer data,
					GError **error );
static void screem_tag_tree_text( GMarkupParseContext *context,
				const gchar *text,
				gsize len,
				gpointer data,
				GError **error );
static void screem_tag_tree_error( GMarkupParseContext *context,
				GError *error,
				gpointer data );
static void screem_tag_tree_parse_node( ScreemTagTreeParser *parser,
					const gchar **attributes,
					const gchar **values,
					gboolean group );
static void screem_tag_tree_parse_insert( ScreemTagTreeParser *parser,
					  const gchar **attributes,
					  const gchar **values );

static GMarkupParser parser_cbs = {
	screem_tag_tree_start_element,
	screem_tag_tree_end_element,
	screem_tag_tree_text,
	NULL,
	screem_tag_tree_error
};

typedef struct NodeInfo {
	gchar *name;
	
	gchar *open;
	gchar *close;
	
	gchar *file;
	gchar *text;
	
	gchar *description;
	
	GtkWidget *dialog;
} NodeInfo;

typedef enum {
	PARSER_STATE_NONE = 0,
	PARSER_STATE_REF,
	PARSER_STATE_GROUP,
	PARSER_STATE_ENTRY,
	PARSER_STATE_FUNCTION,
	PARSER_STATE_INSERT,
	PARSER_STATE_DESCRIPTION
} ScreemTagTreeParserState;

static const gchar *parser_tags[] = {
	"",
	"ref",
	"group",
	"entry",
	"function",
	"insert",
	"description"	
};


static gboolean screem_tag_tree_parse_file( ScreemTagTreeBuildData *bdata, 
						const gchar *filename )
{
	GString *file;
	gboolean ret;
	
	ret = FALSE;
	file = load_file( filename );
	if( file ) {
		ret = screem_tag_tree_parse_string( bdata, file->str, file->len );
	}
	return ret;	
}

static gboolean screem_tag_tree_parse_string( ScreemTagTreeBuildData *bdata, 
						const gchar *str,
						guint len )
{
	GMarkupParseContext *ctx;
	ScreemTagTreeParser *data;
	gboolean ret;

	g_return_val_if_fail( bdata != NULL, FALSE );
	g_return_val_if_fail( str != NULL, FALSE );
	g_return_val_if_fail( len > 0, FALSE );

	ret = FALSE;

	data = g_new0( ScreemTagTreeParser, 1 );
	data->data = bdata;
	data->states = g_slist_prepend( data->states,
					GINT_TO_POINTER( PARSER_STATE_NONE ) );
	data->iters = g_slist_prepend( data->iters, NULL );
	ctx = g_markup_parse_context_new( &parser_cbs, 0, 
					  data, NULL );
	if( ctx ) {
		ret = g_markup_parse_context_parse( ctx, str, len, NULL );
		g_markup_parse_context_free( ctx );
	}
	if( data->iters->next ) {
		g_slist_foreach( data->iters,
				 (GFunc)gtk_tree_iter_free, NULL );
	}
	g_slist_free( data->iters );
	g_slist_free( data->states );
	g_free( data );

	return ret;	
}

static void screem_tag_tree_start_element( GMarkupParseContext *context,
					const gchar *element,
					const gchar **attributes,
					const gchar **values,
					gpointer data,
					GError **error )
{
	ScreemTagTreeParser *parser;
	ScreemTagTreeParserState state;

	parser = (ScreemTagTreeParser*)data;
	state = GPOINTER_TO_INT( parser->states->data );
	
	switch( state ) {
		case PARSER_STATE_NONE:
			if( ! strcmp( "ref", element ) ) {
				screem_tag_tree_parse_node( parser,
							    attributes,
							    values, TRUE );
				state = PARSER_STATE_REF;
			} else {
				/* unknown root element, ignore */
			}
			break;
		case PARSER_STATE_REF:
		/*ref can't have entry / function according to our dtd,
			but it simplifies adding the current doctype branch,
			so we allow it here */
		case PARSER_STATE_GROUP:
			if( ! strcmp( "group", element ) ) {
				screem_tag_tree_parse_node( parser,
							    attributes,
							    values, TRUE );
				state = PARSER_STATE_GROUP;
			} else if( ! strcmp( "entry", element ) ) {
				screem_tag_tree_parse_node( parser,
							    attributes,
							    values, FALSE );
				state = PARSER_STATE_ENTRY;
			} else if( ! strcmp( "function", element ) ) {
				screem_tag_tree_parse_node( parser,
							    attributes,
							    values, FALSE );
				state = PARSER_STATE_FUNCTION;
			}
			break;
		case PARSER_STATE_ENTRY:
			if( ! strcmp( "insert", element ) ) {
				screem_tag_tree_parse_insert( parser,
								attributes,
								values );
				state = PARSER_STATE_INSERT;
			} else if( ! strcmp( "description", element ) ) {
				state = PARSER_STATE_DESCRIPTION;
			}
			break;
		case PARSER_STATE_FUNCTION:
			if( ! strcmp( "insert", element ) ) {
				screem_tag_tree_parse_insert( parser,
								attributes,
								values );
				state = PARSER_STATE_INSERT;
			} else if( ! strcmp( "description", element ) ) {
				state = PARSER_STATE_DESCRIPTION;
			} else {
				state = PARSER_STATE_NONE;
			}
			break;
		case PARSER_STATE_INSERT:
		case PARSER_STATE_DESCRIPTION:
			/* states should not occur */
			state = PARSER_STATE_NONE;
			break;
		default:
			/* unknown state, shouldn't happen */
			state = PARSER_STATE_NONE;
			break;	
	}

	if( state != PARSER_STATE_NONE ) {
		parser->states = g_slist_prepend( parser->states,
						  GINT_TO_POINTER( state ) );
	}
}

static void screem_tag_tree_end_element( GMarkupParseContext *context,
					const gchar *element,
					gpointer data,
					GError **error )
{
	ScreemTagTreeParser *parser;
	ScreemTagTreeParserState state;
	GSList *top;

	parser = (ScreemTagTreeParser*)data;
	state = GPOINTER_TO_INT( parser->states->data );
	top = parser->states;

	switch( state ) {
		default:
			if( strcmp( parser_tags[ state ], element ) ) {
				state = PARSER_STATE_NONE;
			}
			break;
		case PARSER_STATE_NONE:
			/* shouldn't happen */
			state = PARSER_STATE_NONE;
			break;
	}

	if( state != PARSER_STATE_NONE ) {
		parser->states = top->next;
		top->next = NULL;
		g_slist_free( top );
		if( state == PARSER_STATE_REF ||
		    state == PARSER_STATE_GROUP ||
		    state == PARSER_STATE_ENTRY ||
		    state == PARSER_STATE_FUNCTION ) {
			top = parser->iters;
			parser->iters = top->next;
			top->next = NULL;
			if( top->data ) {
				gtk_tree_iter_free( (GtkTreeIter*)top->data );
			}
			g_slist_free( top );
		}
	}
}

static void screem_tag_tree_text( GMarkupParseContext *context,
				const gchar *text,
				gsize len,
				gpointer data,
				GError **error )
{
	ScreemTagTreeParser *parser;
	ScreemTagTreeParserState state;
	ScreemTagTreeBuildData *bdata;
	GSList *top;
	
	GtkTreeIter *it;
	GValue value = { 0 };
	NodeInfo *info;
	
	parser = (ScreemTagTreeParser*)data;
	state = GPOINTER_TO_INT( parser->states->data );
	bdata = parser->data;
		
	top = parser->iters;
	
	it = top->data;
      	gdk_threads_enter();
	if( it ) {
		gtk_tree_model_get_value( GTK_TREE_MODEL( bdata->model ),
					  it, FILE_BROWSER_DATA_COL,
					  &value );
		info = g_value_get_pointer( &value );
		g_value_unset( &value );
	} else {
		info = NULL;
	}
	
	switch( state ) {
		case PARSER_STATE_INSERT:
			if( info ) {
				info->text = g_strdup( text );
			}
			break;
		case PARSER_STATE_DESCRIPTION:
			if( info ) {
				info->description = g_strdup( text );
			}
			break;
		default:
			/* shouldn't happen */
			break;
	}
     	gdk_threads_leave();
}

static void screem_tag_tree_error( GMarkupParseContext *context,
				GError *error,
				gpointer data )
{
	
}

static void screem_tag_tree_parse_node( ScreemTagTreeParser *parser,
					const gchar **attributes,
					const gchar **values,
					gboolean group )
{
	GSList *top;
	GtkTreeIter it;
	GtkTreeIter *iter;
	const gchar *name;
	const gchar *description;
	gint i;
	GtkTreePath *path;
	NodeInfo *info;

	ScreemTagTreeBuildData *bdata;
	
	bdata = parser->data;
	
	top = parser->iters;
	
	info = NULL;	
	name = description = NULL;
	for( i = 0; attributes[ i ]; ++ i ) {
		if( ! strcmp( "name", attributes[ i ] ) ) {
			name = values[ i ];
		} else if( ! strcmp( "description", attributes[ i ] ) ) {
			description = values[ i ];
		}
	}
	
	/* name="" is required, but we will create one if it doesn't exist */
	if( ! name ) {
		name = _( "Unknown Resource" );
	}
	
	/* see if we already have a node matching name at this level,
	   if so we just push that iter instead of creating a new one */

      	gdk_threads_enter();
	
	path = g_hash_table_lookup( bdata->paths, name );

	if( path && ! gtk_tree_model_get_iter( GTK_TREE_MODEL( bdata->model ),
						&it, path ) ) {
		path = NULL;
	} else if( path && ! strcmp( _( "Current Doctype" ), name ) ) {
		gtk_tree_store_remove( GTK_TREE_STORE( bdata->model ), &it );
		gtk_tree_path_free( path );
		path = NULL;
		g_hash_table_remove( bdata->paths, name );
	}
	
	if( ! path ) {
		GdkPixbuf *pixbuf;
		
		gtk_tree_store_append( GTK_TREE_STORE( bdata->model ),
					&it, (GtkTreeIter*)top->data );						 
		
		pixbuf = NULL;
		if( group ) {
			pixbuf = screem_file_browser_get_icon( bdata->browser,
								"/", 16, 16,
								FALSE );
		}
		if( pixbuf ) {
			gtk_tree_store_set( GTK_TREE_STORE( bdata->model ),
					   &it, FILE_BROWSER_ICON_COL,
					   pixbuf, -1 );
			g_object_unref( pixbuf );
		}
		gtk_tree_store_set( GTK_TREE_STORE( bdata->model ), 
					&it, FILE_BROWSER_NAME_COL,
					name, -1 );
		info = g_new0( NodeInfo, 1 );
		info->name = g_strdup( name );
		if( description && group ) {
			info->description = g_strdup( description );
		}
		gtk_tree_store_set( GTK_TREE_STORE( bdata->model ),
					&it, FILE_BROWSER_DATA_COL,
					info, -1 );

		path = NULL;
	 }
	
	iter = gtk_tree_iter_copy( &it );
	parser->iters = g_slist_prepend( parser->iters, iter );
	
	if( ! path ) {
		path = gtk_tree_model_get_path( GTK_TREE_MODEL( bdata->model ),
						&it );
		g_hash_table_insert( bdata->paths, info->name,
				     path );
	}
	
     	gdk_threads_leave();
}


static void screem_tag_tree_parse_insert( ScreemTagTreeParser *parser,
					const gchar **attributes,
					const gchar **values )
{
	ScreemTagTreeParserState state;
	GSList *top;
	
	GtkTreeIter *it;
	GValue value = { 0 };
	NodeInfo *info;
	gint i;
	ScreemTagTreeBuildData *bdata;
	
	bdata = parser->data;
	state = GPOINTER_TO_INT( parser->states->data );
	top = parser->iters;

      	gdk_threads_enter();
	
	it = top->data;
	if( it ) {
		gtk_tree_model_get_value( GTK_TREE_MODEL( bdata->model ),
					  it, FILE_BROWSER_DATA_COL,
					  &value );
		info = g_value_get_pointer( &value );
		g_value_unset( &value );
	} else {
		info = NULL;
	}
	
	for( i = 0; attributes[ i ]; ++ i ) {
		const gchar *attr;
		const gchar *val;
		
		attr = attributes[ i ];
		val = values[ i ];
		
		if( ! strcmp( "href", attr ) ) {
			GdkPixbuf *pixbuf;	
			gchar *path;
			path = relative_to_full( val, 
						 RESOURCE_PATH );
			info->file = path;
			pixbuf = screem_file_browser_get_icon( bdata->browser,
								info->file,
								16, 16, FALSE );
			gtk_tree_store_set( GTK_TREE_STORE( bdata->model ), 
					    it, FILE_BROWSER_ICON_COL,
					    pixbuf, -1 );
			g_object_unref( pixbuf );
		} else if( ! strcmp( "open", attr ) ) {
			info->open = g_strdup( val );
		} else if( ! strcmp( "close", attr ) ) {
			info->close = g_strdup( val );
		}
	}
      	gdk_threads_leave();
}

enum {
	ARG_0,
	ARG_WINDOW
};
static gint screem_tag_tree_compare_func( GtkTreeModel *model, 
				    	  GtkTreeIter *a, GtkTreeIter *b,
				    	  gpointer data );
static void screem_tagtree_sel_change( GtkTreeSelection *selection,
					ScreemTagTree *tree );
static ScreemFileBrowser *screem_tag_tree_get_browser( void );
static void screem_tag_tree_set_model( ScreemTagTree *tree, 
					GtkTreeModel *model );
static gpointer screem_tag_tree_create( gpointer data );

static void screem_tag_tree_class_init( ScreemTagTreeClass *klass );
static void screem_tag_tree_init( ScreemTagTree *tag_tree );
static void screem_tag_tree_finalize( GObject *object );
static void screem_tag_tree_set_prop( GObject *object, guint property_id,
				    const GValue *value, GParamSpec *pspec );
static void screem_tag_tree_get_prop( GObject *object, guint property_id,
				      GValue *value, GParamSpec *pspec);
static void screem_tag_tree_size_request( GtkWidget *widget,
                           	  	  GtkRequisition *requisition );
static void screem_tag_tree_size_allocate( GtkWidget *widget,
                              		   GtkAllocation *allocation );

ScreemTagTree *screem_tag_tree_new()
{
	ScreemTagTree *tree;
	GType type;

	type = screem_tag_tree_get_type();
	
	tree = SCREEM_TAG_TREE( g_object_new( type, NULL ) );

	return tree;
}

gboolean screem_tag_tree_load_trees( void )
{
	ScreemFileBrowser *browser;
	static GThread *thread = NULL;
	
	browser = screem_tag_tree_get_browser();
	
	/* spawn load thread, so the ui will still be interactive
	   as tag tree can take a while to load */
	if( ! thread ) {
		thread = g_thread_create( screem_tag_tree_create, browser,
					  FALSE, NULL );
	} else {
		g_warning( "screem_tag_tree_load_trees() called more than once\n" );
	}
	
	return FALSE;
}

void screem_tag_tree_add_dtd( ScreemTagTree *tag_tree, ScreemDTD *dtd )
{
	GString *str;
	const GSList *elements;
	ScreemTagTreeBuildData *bdata;
	
	if( ! dtd ) {
		return;
	}
		
	elements = screem_dtd_get_elements( dtd );
	
	str = g_string_new( "<?xml version=\"1.0\"?>" );
	g_string_append_printf( str, "<ref name=\"%s\" description=\"%s\">",
				_( "Current Doctype" ),
				_( "Elements for the current doctype" ) );
	while( elements ) {
		const gchar *name;
		ScreemDTDClose close;

		name = screem_dtd_element_get_name( (ScreemDTDElement*)elements->data );
		close = screem_dtd_element_close_state( dtd, name );

		g_string_append_printf(str, 
				"<entry name=\"%s\"><insert open=\"&lt;%s&gt;\"",
				name, name );
		if( close != MUSTNT ) {
			g_string_append_printf( str, " close=\"&lt;/%s&gt;\"",
						name );
		}
		g_string_append( str, "/></entry>" );
	
		elements = elements->next;
	}

	g_string_append( str, "</ref>" );

	bdata = g_new0( ScreemTagTreeBuildData, 1 );
	bdata->browser = tag_tree->private->browser;
	bdata->paths = tag_tree->private->paths;
	bdata->model = GTK_TREE_MODEL( tag_tree->private->store );
	
	screem_tag_tree_parse_string( bdata, str->str, str->len );
	g_string_free( str, TRUE );
	g_free( bdata );
}


/* static stuff */
static gboolean screem_tag_tree_click( GtkTreeView *tree_view,
				       GtkTreePath *path,
				       GtkTreeViewColumn *column,
				       gpointer data )
{
	GtkTreeIter iter;
	GtkTreeModel *model;
	GtkTreeSelection *sel;

	ScreemTagTree *tag_tree;

	tag_tree = SCREEM_TAG_TREE( data );


	model = GTK_TREE_MODEL( tag_tree->private->store );
	sel = gtk_tree_view_get_selection( GTK_TREE_VIEW( tree_view ) );

	if( sel ) {
		GValue value = { 0 };
		NodeInfo *info;
		ScreemWindow *window;
		ScreemEditor *editor;

		window = SCREEM_WINDOW( tag_tree->private->window );

		editor = SCREEM_EDITOR( window->details->editor );

		gtk_tree_selection_get_selected( sel, &model, &iter );
		gtk_tree_model_get_value( model, &iter, FILE_BROWSER_DATA_COL,
					 &value );

		info = g_value_get_pointer( &value );

		if( info && info->dialog ) {
			gtk_widget_show_all( info->dialog );
		}
		
		if( info && info->file ) {
			GnomeVFSURI *vuri;
			
			vuri = gnome_vfs_uri_new( info->file );
			if( ! vuri ) {
				screem_window_show_error( window, _( "Invalid URI for requested resource" ) );
			} else if( gnome_vfs_uri_is_local( vuri ) ||
				  ! window->details->offline ) {
				GString *data;
				
				data = load_file( info->file );
				/* now set info->open to be the file contents,
			   	and set info->file to NULL, this way
			   	we don't refetch files we already have */
				g_free( info->file );
				info->file = NULL;
				info->text = data->str;
				g_string_free( data, FALSE );
			} else {
				screem_window_show_message( window, _( "Remote resources are not available in offline mode" ) );	
			}

			if( vuri ) {	
				gnome_vfs_uri_unref( vuri );
			}
		} else if( info && info->open ) {
			screem_editor_insert_markup( editor, info->open,
						     info->close );
		}
		/* we do info->text here so you can have something like
		 * <insert open="&lt;title&gt;" close="&lt;title&gt;">
		 *      page title
		 * </insert>
		 */
		if( info && info->text ) {
			screem_editor_insert( editor, -1, info->text );
		}
		g_value_unset( &value );
		
	} else {
		/* hmm, no iterator */
		
	}

	return (sel != NULL);
}


static gboolean screem_tag_tree_press( GtkWidget *widget, 
				       GdkEventButton *event,
				       ScreemTagTree *tag_tree )
{
	return FALSE;
}

static gint screem_tag_tree_compare_func( GtkTreeModel *model, 
				    	  GtkTreeIter *a, GtkTreeIter *b,
				    	  gpointer data )
{
	GValue avalue = {0,};
	GValue bvalue = {0,};
        NodeInfo *ainfo;
        NodeInfo *binfo;

	int ret = 0;

	gtk_tree_model_get_value( model, a, FILE_BROWSER_DATA_COL, &avalue );
	gtk_tree_model_get_value( model, b, FILE_BROWSER_DATA_COL, &bvalue );

	ainfo = g_value_get_pointer( &avalue );
	binfo = g_value_get_pointer( &bvalue );
	
	if( ainfo && ! binfo ) {
		ret = -1;
	} else if( binfo && !ainfo ) {
		ret = 1;
	} else if( ! ainfo && ! binfo ) {
		ret = 0;
	} else {
		ret = strcmp( ainfo->name, binfo->name );
	}
	
	g_value_unset( &avalue );
	g_value_unset( &bvalue );

	return ret;
}

static void screem_tagtree_sel_change( GtkTreeSelection *selection,
					ScreemTagTree *tree )
{
	ScreemWindow *window;
	GtkTreeModel *model;
	GtkTreeIter iter;
	GValue value = { 0 };
	NodeInfo *info;
	
	window = SCREEM_WINDOW( tree->private->window );
	if( gtk_tree_selection_count_selected_rows( selection ) > 0 ) {
		gtk_tree_selection_get_selected( selection, &model, &iter );
		gtk_tree_model_get_value( model, &iter, FILE_BROWSER_DATA_COL,
					 &value );
		info = g_value_get_pointer( &value );
		if( info && info->description ) {
			gchar *msg;
			
			msg = g_strconcat( _( "Resource: "), info->name, " : ",
					   info->description, NULL );
			
			screem_window_show_message( window, msg );
			
			g_free( msg );
		}
		g_value_unset( &value );
	}
}

static ScreemFileBrowser *screem_tag_tree_get_browser( void )
{
	static ScreemFileBrowser *browser = NULL;
	
	if( ! browser ) {
		GHashTable *table;
		
		browser = screem_file_browser_new();
		screem_file_browser_set_mode( browser,
						FILE_BROWSE_RECURSE );
		screem_file_browser_set_sort_func( browser,
					   screem_tag_tree_compare_func,
					   NULL );
		
		table = g_hash_table_new( g_str_hash, g_str_equal );
		g_object_set_data( G_OBJECT( browser ), "paths", table );
	}
	
	return browser;
}

static void screem_tag_tree_set_model( ScreemTagTree *tree, GtkTreeModel *model )
{
	ScreemTagTreePrivate *priv;
	
	g_return_if_fail( SCREEM_IS_TAG_TREE( tree ) );
	g_return_if_fail( model != NULL );
	
	priv = tree->private;

	gtk_tree_view_set_model( GTK_TREE_VIEW( tree->private->view ), model );
	
	if( GTK_IS_TREE_SORTABLE( model ) ) {
		GtkTreeModelSort *sortable;
		
		sortable = GTK_TREE_MODEL_SORT( model );
		model = gtk_tree_model_sort_get_model( sortable );
	}
	priv->store = GTK_TREE_STORE( model );	
}

static gpointer screem_tag_tree_create( gpointer data )
{
	const gchar *home;
	const gchar *mime_type;
	gchar *file;
	gint num;
	gint pos;
	struct dirent **namelist;

	gint i;
	const gchar *treepaths[] = {
		DATADIR"/screem/tagtrees",
		NULL,
		NULL
	};
	ScreemTagTreeBuildData *bdata;
		
	bdata = g_new0( ScreemTagTreeBuildData, 1 );
	bdata->browser = SCREEM_FILE_BROWSER( data );
	bdata->paths = g_object_get_data( G_OBJECT( data ), "paths" );
	bdata->model = screem_file_browser_get_model(SCREEM_FILE_BROWSER(data));
	bdata->model = gtk_tree_model_sort_get_model( GTK_TREE_MODEL_SORT( bdata->model ) );
	
	/* load tag trees */
	home = g_get_home_dir();
	treepaths[ 1 ] = g_strconcat( home, G_DIR_SEPARATOR_S, ".screem", NULL );

	for( i = 0; treepaths[ i ]; ++ i ) {
		num = scandir( treepaths[ i ], &namelist, 0, alphasort );
		
		for( pos = 0; pos < num; pos ++ ) {
			gchar *path;
			
			file = namelist[ pos ]->d_name;
			path = g_strconcat( treepaths[ i ],
						G_DIR_SEPARATOR_S,
					    	file, NULL );
			mime_type = gnome_vfs_get_mime_type( path );
			if( mime_type &&
		    	! strcmp( "application/x-screem-tag-tree", mime_type ) ) {
				screem_tag_tree_parse_file( bdata, path );
			}
			g_free( path );
			free( namelist[ pos ] );
		}
		if( num > 0 ) {
			free( namelist );
		}
	}
	g_free( (gchar*)treepaths[ 1 ] );

	return NULL;
}

/* G Object stuff */
#define PARENT_TYPE GTK_TYPE_BIN

static gpointer parent_class;

static void screem_tag_tree_class_init( ScreemTagTreeClass *klass )
{
	GObjectClass *object_class;
	GtkWidgetClass *widget_class;

	GParamSpec *pspec;

	object_class = G_OBJECT_CLASS( klass );
	widget_class = (GtkWidgetClass *)klass;

	object_class->finalize = screem_tag_tree_finalize;
	parent_class = g_type_class_peek_parent( klass );

	object_class->get_property = screem_tag_tree_get_prop;
	object_class->set_property = screem_tag_tree_set_prop;

	pspec = g_param_spec_pointer( "window", "window",
				      "window object",
				      G_PARAM_READABLE |
				      G_PARAM_WRITABLE );
	g_object_class_install_property( object_class,
					 ARG_WINDOW, pspec );
	
	widget_class->size_request = screem_tag_tree_size_request;
	widget_class->size_allocate = screem_tag_tree_size_allocate;
}

static void screem_tag_tree_init( ScreemTagTree *tag_tree )
{
	GtkWidget *sw;
	GtkCellRenderer *rend;
	GtkCellRenderer *prend;
	GtkTreeViewColumn *col;
	GtkTreeSelection *selection;
	
	ScreemTagTreePrivate *priv;
	
	priv = tag_tree->private = g_new0( ScreemTagTreePrivate, 1 );

	priv->browser = screem_tag_tree_get_browser();
		
	priv->box = gtk_vbox_new( FALSE, 0 );
	gtk_container_add( GTK_CONTAINER( tag_tree ), priv->box );
		
	sw = gtk_scrolled_window_new( NULL, NULL );
	gtk_scrolled_window_set_policy( GTK_SCROLLED_WINDOW( sw ),
					GTK_POLICY_AUTOMATIC,
                                        GTK_POLICY_AUTOMATIC );
	gtk_scrolled_window_set_shadow_type( GTK_SCROLLED_WINDOW( sw ),
						GTK_SHADOW_IN );
	priv->view = gtk_tree_view_new();
	screem_tag_tree_set_model( tag_tree, screem_file_browser_get_model( priv->browser ) );
	
        gtk_tree_view_set_rules_hint( GTK_TREE_VIEW( priv->view ), TRUE );

	selection = gtk_tree_view_get_selection( GTK_TREE_VIEW( priv->view ) );
	g_signal_connect( G_OBJECT( selection ), "changed",
			  G_CALLBACK( screem_tagtree_sel_change ), tag_tree );

	priv->paths = g_object_get_data( G_OBJECT( priv->browser ), "paths" );
	
	prend = gtk_cell_renderer_pixbuf_new();
	rend = gtk_cell_renderer_text_new();
	col = gtk_tree_view_column_new();
	
	gtk_tree_view_column_pack_start( col, prend, FALSE );
	gtk_tree_view_column_pack_start( col, rend, TRUE );
	gtk_tree_view_column_set_resizable( col, TRUE );
	gtk_tree_view_append_column( GTK_TREE_VIEW( priv->view ), col );
	
	gtk_tree_view_column_set_title( col, _( "Resources" ) );
	gtk_tree_view_column_set_attributes( col, rend,
					     "text", 
					     FILE_BROWSER_NAME_COL,
					     NULL );
	gtk_tree_view_column_set_attributes( col, prend,
					     "pixbuf",
					     FILE_BROWSER_ICON_COL,
					     NULL );

	gtk_tree_view_set_headers_visible( GTK_TREE_VIEW( priv->view ), TRUE );

	gtk_widget_show( priv->view );

	g_signal_connect( G_OBJECT( priv->view ), "row_activated",
			  G_CALLBACK( screem_tag_tree_click ),
			  tag_tree );

	g_signal_connect( G_OBJECT( priv->view ), "button_press_event",
			  G_CALLBACK( screem_tag_tree_press ),
			  tag_tree );

	gtk_container_add( GTK_CONTAINER( sw ), priv->view );

	gtk_box_pack_start( GTK_BOX( priv->box ), sw, TRUE, TRUE, 0 );

	gtk_widget_show_all( GTK_WIDGET( tag_tree ) );
}


static void screem_tag_tree_finalize( GObject *object )
{
	G_OBJECT_CLASS( parent_class )->finalize( object );
}

static void screem_tag_tree_set_prop( GObject *object, guint property_id,
				    const GValue *value, GParamSpec *pspec )
{
	ScreemTagTree *tree;

	tree = SCREEM_TAG_TREE( object );

	switch( property_id ) {
	case ARG_WINDOW:
		tree->private->window = g_value_get_pointer( value );

		break;
	}
}

static void screem_tag_tree_get_prop( GObject *object, guint property_id,
				    GValue *value, GParamSpec *pspec)
{
	switch( property_id ) {
	case ARG_WINDOW:
		g_value_set_pointer( value,
				     (gpointer)
				     SCREEM_TAG_TREE( object )->private->window );
		break;
	}
}

static void screem_tag_tree_size_request( GtkWidget *widget,
                           	  	  GtkRequisition *requisition )
{
	GtkBin *bin;

	bin = GTK_BIN (widget);

	requisition->width = GTK_CONTAINER( widget )->border_width * 2;
	requisition->height = GTK_CONTAINER( widget )->border_width * 2;

	if( bin->child && GTK_WIDGET_VISIBLE( bin->child ) ) {
		GtkRequisition child_requisition;

		gtk_widget_size_request( bin->child, &child_requisition );
		requisition->width += child_requisition.width;
		requisition->height += child_requisition.height;
	}
}

static void screem_tag_tree_size_allocate( GtkWidget *widget,
                              		   GtkAllocation *allocation )
{
	GtkBin *bin;
	GtkAllocation child_allocation;

	bin = GTK_BIN( widget );
	widget->allocation = *allocation;
	
	if( bin->child ) {
	child_allocation.x = allocation->x + 
				GTK_CONTAINER( widget )->border_width; 
	child_allocation.y = allocation->y + 
				GTK_CONTAINER (widget)->border_width;
	child_allocation.width = MAX( allocation->width - 
				      GTK_CONTAINER( widget )->border_width * 2,
					0);
	child_allocation.height = MAX( allocation->height - 
				       GTK_CONTAINER (widget)->border_width * 2,
					0);
	gtk_widget_size_allocate( bin->child, &child_allocation );
	}
}

GType screem_tag_tree_get_type()
{
	static GType type = 0;
	
	if( ! type ) {
		static const GTypeInfo info = {
			sizeof( ScreemTagTreeClass ),
			NULL, /* base init */
			NULL, /* base finalise */
			(GClassInitFunc)screem_tag_tree_class_init,
			NULL, /* class finalise */
			NULL, /* class data */
			sizeof( ScreemTagTree ),
			0, /* n_preallocs */
			(GInstanceInitFunc)screem_tag_tree_init
		};

		type = g_type_register_static( PARENT_TYPE,
					       "ScreemTagTree",
					       &info, 0 );
	}

	return type;
}
